package applet.fileviewer;

import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.border.TitledBorder;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.table.TableCellRenderer;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;
import applet.JSIDPlay2;
import applet.SidTuneConverter;
import applet.TuneTab;
import applet.events.favorites.IAddFavoritesTab;
import applet.events.favorites.IFavoriteTabNames;
import applet.events.favorites.IGetFavorites;
import applet.favorites.IFavorites;
import applet.fileviewer.search.ISearchListener;
import applet.fileviewer.search.SearchInIndexThread;
import applet.fileviewer.search.SearchIndexCreator;
import applet.fileviewer.search.SearchIndexerThread;
import applet.fileviewer.search.SearchThread;
import applet.sidtune.SidTuneInfoCache;
import applet.stil.STILView;
import applet.ui.JNiceButton;

/**
 * Common view base for HVSC and CGSC collections.
 * Loosely based on Rhythmbox, which is probably based on iTunes.
 * Display is divided to 2 vertical panels of identical widths
 * 
 * - 1st shows file meta info in table, and composer's photo, scaled
 *   to 100% of width. These take the whole vertical space.
 *   
 * - 2nd column show search bar, which is used to match song and artist name

 * - 2nd column displays list of artists.

 * - 2nd column displays list of songs matching search criteria and selected artist.
 *   - currently playing symbol
 *   - artist name
 *   - song name
 *   - total song length (?)
 * 
 * @author Ken Hndel
 * @author Antti Lankila
 */
public abstract class FileViewer extends TuneTab implements TreeSelectionListener, ISearchListener {
	private static final Properties sidAuthors = new Properties();
	
	static {
		InputStream is = FileViewer.class.getResourceAsStream("pictures.properties");
		try {
			sidAuthors.load(is);
		} catch (IOException e) {
			throw new ExceptionInInitializerError(e);
		}
	}

	protected JSIDPlay2 fSidplay2;

	protected JTree fFileTree;

	protected TreeModel fTreemodel;

	protected JTable fInfoTable;

	protected TuneInfoTableModel fInfoModel;

	private JComponent photograph;

	private Image composerImage;
	
	protected JTextField fSearchFor = new JTextField("", 20);

	private JButton fStartButton;
	private JButton fStopButton;

	protected JTextField fRootDir = new JTextField();

	private final String fTitle;

	private final String fURL;

	private final String fMessage;

	private Connection fConnection;

	protected SearchThread fSearchThread;
	protected Object fSavedState;
	protected String fLastSearchFor;
	protected int fLastCriteriaIndex;
	protected boolean fChangedOptions;

	protected JComboBox fScopeBox, fCriteriaBox;

	private JComboBox fResultBox;

	private JButton fRecreateButton;

	protected File fLastDir;

	private JButton fResetButton;

	protected IFavorites fSearchResultFavorites;

	private int progCount;

	public FileViewer(final JSIDPlay2 sidplay, final String title, final String sourceURL, final String message) {
		this.fSidplay2 = sidplay;
		this.fTitle = title;
		this.fURL = sourceURL;
		this.fMessage = message;
		fFileTree = new JTree(new DefaultMutableTreeNode(fMessage));
		fFileTree.setExpandsSelectedPaths(true);
		sidplay.support.addPropertyChangeListener(this);
	}

	@Override
	public void createContents() {
		setLayout(new GridBagLayout());
		final GridBagConstraints c = new GridBagConstraints();

		fInfoTable = new JTable(fInfoModel = new TuneInfoTableModel(fSidplay2)) {
			@Override
			public TableCellRenderer getCellRenderer(final int row, final int column) {
				return new TuneInfoTableRenderer();
			}
		};
		fInfoTable.setTableHeader(null);

		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		JComponent infoTableContainer = new JScrollPane(fInfoTable);
		infoTableContainer.setBorder(new TitledBorder("Tune Info"));
		add(infoTableContainer, c);

		c.gridy ++;
		c.weighty = 1;
		photograph = new JPanel();
		photograph.setLayout(new GridLayout(1, 1));
		photograph.add(new JComponent() {
			{
				setOpaque(true);
				setPreferredSize(new Dimension(200, 200));
			}
			
			@Override
			public void paint(Graphics g) {
				g.setColor(getBackground());
				g.fillRect(0, 0, getWidth(), getHeight());
				if (composerImage != null) {
					int picWidth = composerImage.getWidth(null);
					int picHeight = composerImage.getHeight(null);
					
					int availableWidth = getWidth();
					int availableHeight = getHeight();
					
					/* figure out which dimension limits scaling first */
					float scaleW = (float) availableWidth / picWidth;
					float scaleH = (float) availableHeight / picHeight;

					float safeScale = scaleW > scaleH ? scaleH : scaleW;
					picWidth *= safeScale;
					picHeight *= safeScale;
					
					((Graphics2D) g).setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
					g.drawImage(composerImage,
							(availableWidth - picWidth) / 2,
							(availableHeight - picHeight) / 2,
							picWidth,
							picHeight,
							null);
				}
			}
			
			@Override
			public void update(Graphics g) {
				paint(g);
			}
		});
		photograph.setBorder(new TitledBorder("Composer's Photo"));
		add(photograph, c);

		fFileTree.setCellRenderer(new TuneTreeRenderer(this, fSidplay2));
		fFileTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		fFileTree.setScrollsOnExpand(true);
		fFileTree.setToolTipText("Open context menu on files and folders to add tunes to your favorites!");
		
		// Listen for when the selection changes.
		fFileTree.addTreeSelectionListener(this);
		fFileTree.addKeyListener(new KeyAdapter() {

			@Override
			public void keyTyped(final KeyEvent e) {
				/* handle return presses */
				if (e.getKeyChar() != '\n') {
					return;
				}
				
				/* find current selection */
				final TreePath path = fFileTree.getSelectionPath();
				if (path == null) {
					return;
				}
				
				TuneFile f = (TuneFile) path.getLastPathComponent();
				if (f.isFile()) {
					fSidplay2.playFile((File) path.getLastPathComponent(), FileViewer.this);
					fFileTree.repaint();
				}
			}

		});
		fFileTree.addMouseListener(new MouseAdapter() {

			private JPopupMenu fContextPopup;

			@Override
			public void mouseClicked(final MouseEvent mouseEvent) {
				final TreePath selectionPath = fFileTree.getSelectionPath();
				if (selectionPath != null && mouseEvent.getButton() == MouseEvent.BUTTON1 && mouseEvent.getClickCount() > 1) {
					if (selectionPath.getLastPathComponent() instanceof TuneFile && ((TuneFile) selectionPath.getLastPathComponent()).isFile()) {
						fSidplay2.playFile((File) selectionPath.getLastPathComponent(), FileViewer.this);
						fFileTree.repaint();
					}
				} else if (selectionPath != null && mouseEvent.getButton() == MouseEvent.BUTTON3 && mouseEvent.getClickCount() == 1) {
					if (selectionPath.getLastPathComponent() instanceof TuneFile
							/*
							 * && ((TuneFile) path.getLastPathComponent()) .isFile()
							 */) {
						fContextPopup = new JPopupMenu("Tune Actions");
						final JMenu addToFavorites = new JMenu("Add to Favorites");
						fSidplay2.support.firePropertyChange(IFavoriteTabNames.class.getName(), null, new IFavoriteTabNames() {

							public void setFavoriteTabNames(final String[] names, final String selected) {
								for (int i = 0; i < names.length; i++) {
									final int index = i;
									final String title = names[i];
									final JMenuItem tabItem = new JMenuItem(title);
									tabItem.addActionListener(new ActionListener() {

										public void actionPerformed(final ActionEvent e) {
											final TreePath path = fFileTree.getSelectionPath();
											if (path.getLastPathComponent() instanceof TuneFile
													/*
													 * && ((TuneFile) path .getLastPathComponent ()) .isFile()
													 */) {
												fSidplay2.support.firePropertyChange(IGetFavorites.class.getName(), null, new IGetFavorites() {

													public int getIndex() {
														return index;
													}

													public void setFavorites(final IFavorites favorites) {
														final TreePath newPath = fFileTree.getSelectionPath();
														favorites.addToFavorites(new File[] { new File(((File) newPath.getLastPathComponent()).getAbsolutePath()) });
													}

												});
											}
										}

									});
									addToFavorites.add(tabItem);
								}
							}

						});
						fContextPopup.add(addToFavorites);
						final JMenu convertItem = new JMenu("Convert to...");

						final JMenuItem psid64 = new JMenuItem("PSID64 (.prg)");
						psid64.addActionListener(new ActionListener() {

							public void actionPerformed(final ActionEvent e) {
								final TreePath path = fFileTree.getSelectionPath();
								if (path.getLastPathComponent() instanceof TuneFile
										/*
										 * && ((TuneFile) path.getLastPathComponent()) .isFile()
										 */) {
									final TuneFile tuneFile = (TuneFile) path.getLastPathComponent();
									final JFileChooser fc = new JFileChooser(fLastDir);
									fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
									final int result = fc.showOpenDialog(fSidplay2);
									if (result == JFileChooser.APPROVE_OPTION && fc.getSelectedFile() != null) {
										fLastDir = fc.getSelectedFile();

										convertFile(tuneFile);
									}
								}
							}

						});
						convertItem.add(psid64);

						fContextPopup.add(convertItem);

						fContextPopup.show((Component) mouseEvent.getSource(), mouseEvent.getX(), mouseEvent.getY());
					}
				}
			}

		});

		c.gridy --;
		c.gridx = 1;
		c.gridheight = 2;
		c.weighty = 1;
		add(new JScrollPane(fFileTree), c);

		c.gridy += 2;

		c.gridx = 0;
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridheight = 1;
		c.weighty = 0;
		
		final JPanel search = new JPanel();
		search.setBorder(new TitledBorder("Search"));
		search.setLayout(new BoxLayout(search, BoxLayout.Y_AXIS));
		add(search, c);

		final JPanel searchOptions = new JPanel();
		searchOptions.setLayout(new FlowLayout(FlowLayout.LEFT));
		search.add(searchOptions);

		final JLabel searchFor = new JLabel("Search for:") {
			@Override
			public Dimension getPreferredSize() {
				return new Dimension(60, super.getPreferredSize().height);
			}
		};
		searchOptions.add(searchFor);

		final String[] criterias = new String[2 + SidTuneInfoCache.SIDTUNE_INFOS.length + STILView.INFOS.length];
		for (int i = 0; i < criterias.length; i++) {
			if (i < 2) {
				criterias[i] = i == 0 ? "File Name" : "Full Path";
			} else if (i - 2 < SidTuneInfoCache.SIDTUNE_INFOS.length) {
				criterias[i] = SidTuneInfoCache.SIDTUNE_INFOS[i - 2];
			} else {
				criterias[i] = STILView.INFOS[i - 2 - SidTuneInfoCache.SIDTUNE_INFOS.length];
			}
		}

		fCriteriaBox = new JComboBox(criterias);
		fCriteriaBox.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				if (fCriteriaBox.getSelectedIndex() != fLastCriteriaIndex) {
					fChangedOptions = true;
					fLastCriteriaIndex = fCriteriaBox.getSelectedIndex();
				}
			}
		});
		searchOptions.add(fCriteriaBox);

		fSearchFor.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				if (!fSearchFor.getText().equals(fLastSearchFor)) {
					fChangedOptions = true;
					fLastSearchFor = fSearchFor.getText();
				}
				startSearch(false);
			}
		});
		searchOptions.add(fSearchFor);

		fStartButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/search.png")));
		fStartButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				startSearch(false);
			}

		});
		fStartButton.setToolTipText("Search Next");
		searchOptions.add(fStartButton);

		fStopButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/player_stop.png")));
		fStopButton.setToolTipText("Search Stop");
		fStopButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				if (fSearchThread != null && fSearchThread.isAlive()) {
					fSearchThread.setAborted(true);
				}
			}

		});
		fStopButton.setEnabled(false);
		searchOptions.add(fStopButton);

		fResetButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/control_start_blue.png")));
		fResetButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				resetSearch();
			}

		});
		fResetButton.setToolTipText("Search From Start Again");
		searchOptions.add(fResetButton);

		fRecreateButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/database_edit.png")));
		fRecreateButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				final int response = JOptionPane.showConfirmDialog(fSidplay2, String.format("Re-Create Database %s ?: ", fRootDir.getText() + "/" + SearchIndexCreator.JSIDPLAY2_IDX), "Create Search Database", JOptionPane.YES_NO_OPTION);
				if (response == JOptionPane.YES_OPTION) {
					startSearch(true);
				}
			}

		});
		fRecreateButton.setToolTipText("Re-Create Search Database");
		searchOptions.add(fRecreateButton);

		fScopeBox = new JComboBox(new String[] { "Forward", "Backward", });
		searchOptions.add(fScopeBox);

		final JPanel searchActions = new JPanel();
		searchActions.setLayout(new FlowLayout(FlowLayout.LEFT));
		search.add(searchActions);

		final JLabel optionsLbl = new JLabel("Options:");
		searchActions.add(optionsLbl);

		searchActions.add(new JLabel());
		searchActions.add(new JLabel());

		final JLabel actionLbl = new JLabel("Result Action:");
		searchActions.add(actionLbl);

		fResultBox = new JComboBox(new String[] { "Show Next Match", "Add to a new Playlist" });
		searchActions.add(fResultBox);

		c.gridy ++;

		c.gridx = 0;
		c.gridwidth = GridBagConstraints.REMAINDER;
		
		final JPanel options = new JPanel();
		options.setBorder(new TitledBorder("Configuration"));
		options.setLayout(new BoxLayout(options, BoxLayout.Y_AXIS));
		add(options, c);

		final JPanel baseDirOptions = new JPanel();
		baseDirOptions.setLayout(new FlowLayout(FlowLayout.LEFT));
		options.add(baseDirOptions);

		final JLabel baseDir = new JLabel("Base dir:");
		baseDirOptions.add(baseDir);
		fRootDir.setEnabled(false);
		baseDirOptions.add(fRootDir);

		final JButton browsebutton = new JButton("Browse...", new ImageIcon(JSIDPlay2.class.getResource("icons/_active__open.png")));
		JSIDPlay2.makeDisabledIcon(browsebutton);
		baseDirOptions.add(browsebutton);
		browsebutton.addActionListener(new ActionListener() {

			private File fLastDir2;

			public void actionPerformed(final ActionEvent e) {
				final JFileChooser fc = new JFileChooser(fLastDir2);
				fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				final int result = fc.showOpenDialog(fSidplay2);
				if (result == JFileChooser.APPROVE_OPTION && fc.getSelectedFile() != null) {
					fLastDir2 = fc.getSelectedFile().getParentFile();
					final String hvsc = fc.getSelectedFile().getAbsolutePath();
					setRoot(hvsc);
				}
			}

		});
		final JLinkButton linkButton;
		try {
			linkButton = new JLinkButton("Download " + fTitle, new URL(fURL));
			linkButton.addActionListener(new ActionListener() {

				public void actionPerformed(final ActionEvent e) {
					try {
						// Try to use applet context to open the URL.
						// if we aren't an applet, then we'll get a NPE.
						fSidplay2.getAppletContext().showDocument(linkButton.getLinkURL(), "downloadTunes");
					}
					catch (NullPointerException npe) {
						// an application opens the default browser
						Desktop desktop = null;
						if (Desktop.isDesktopSupported()) {
							desktop = Desktop.getDesktop();
							if (desktop.isSupported(Desktop.Action.BROWSE)) {
								try {
									desktop.browse(linkButton.getLinkURL().toURI());
								} catch (final IOException ioe) {
									ioe.printStackTrace();
								} catch (final URISyntaxException use) {
									use.printStackTrace();
								}
							}
						}
					}
				}

			});
			baseDirOptions.add(linkButton);
		} catch (final MalformedURLException e) {
			e.printStackTrace();
		}

	}

	/**
	 * Creates an ImageIcon if the path is valid.
	 * 
	 * @param String
	 *            - resource path
	 * @param String
	 *            - description of the file
	 */
	protected ImageIcon createImageIcon(final String path, final String description) {
		final java.net.URL imgURL = getClass().getResource(path);
		if (imgURL != null) {
			return new ImageIcon(imgURL, description);
		} else {
			System.err.println("Couldn't find file: " + path);
			return null;
		}
	}

	protected void setRoot(final String hvsc) {
		fRootDir.setText(hvsc);
		if (fFileTree.getModel() != fTreemodel || !new File(hvsc).equals(fTreemodel.getRoot())) {
			fFileTree.setModel(fTreemodel = new TuneTreeModel(fRootDir.getText()));
		}
		if (((TuneFile) fTreemodel.getRoot()).exists()) {
			fFileTree.setSelectionPath(new TreePath(fTreemodel.getRoot()));
		}
		setRootDir(hvsc);
		repaint();
		resetSearch();
		if (fConnection != null) {
			try {
				fConnection.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
			fConnection = null;
		}
	}

	protected abstract void setRootDir(String hvsc);

	/**
	 * On tree selection display tune info in the table
	 */
	public void valueChanged(final TreeSelectionEvent treeselectionevent) {
		final TreePath treePath = treeselectionevent.getNewLeadSelectionPath();
		if (treePath == null) {
			return;
		}
		final Object pathComponent = treePath.getLastPathComponent();
		if (pathComponent instanceof TuneFile) {
			final TuneFile file = (TuneFile) pathComponent;
			if (file.isFile()) {
				fInfoModel.setFile(file);
				showPhoto();
				fInfoTable.updateUI();
			}
		}
	}

	protected void showPhoto() {
		Object authorInfo = fInfoModel.getValueAt(SidTuneInfoCache.INFO_AUTHOR, 1);
		if (authorInfo == null) {
			// author not available
			return;
		}
		String photoRes = sidAuthors.getProperty(authorInfo.toString());
		composerImage = null;
		
		if (photoRes != null) {
			photoRes = "Photos/" + photoRes;
			final URL resource = JSIDPlay2.class.getResource(photoRes);
			composerImage = new ImageIcon(resource).getImage();
		}

		photograph.repaint();
	}

	public void searchStart() {
		fStartButton.setEnabled(false);
		fStopButton.setEnabled(true);
		fResetButton.setEnabled(false);
		fRecreateButton.setEnabled(false);
	}

	public void searchStop(final boolean canceled) {
		fStartButton.setEnabled(true);
		fStopButton.setEnabled(false);
		fResetButton.setEnabled(true);
		fRecreateButton.setEnabled(true);

		// remember search state
		fSavedState = fSearchThread.getSearchState();
		fSidplay2.setProgressBar(100);
	}

	public void searchHit(final File current) {
		if (fSearchThread instanceof SearchIndexerThread) {
			// if search index is created, do not show the next result
			fSidplay2.setProgressBar(++ progCount % 100);
			return;
		}
		switch (fResultBox.getSelectedIndex()) {
		case 0:
			fSearchThread.setAborted(true);
			searchStop(true);

			showNextHit(current);
			break;

		case 1:
			fSearchResultFavorites.addToFavorites(new File[] { current });
			break;

		default:
			break;
		}

	}

	private void showNextHit(final File matchFile) {
		final File rootFile = new File(fRootDir.getText());
		try {
			if (matchFile.getCanonicalPath().startsWith(rootFile.getCanonicalPath())) {
				final String name = matchFile.getCanonicalPath().substring(rootFile.getCanonicalPath().length()).replace('\\', '/');
				if (name.startsWith("/")) {
					final StringTokenizer selectFile = new StringTokenizer(name, "/");
					if (fFileTree.getModel() == fTreemodel) {
						final ArrayList<TuneFile> pathSegs = new ArrayList<TuneFile>();
						TuneFile lastFile = new TuneFile(new File(fRootDir.getText()).getCanonicalPath());
						pathSegs.add(lastFile);
						while (selectFile.hasMoreTokens()) {
							final String nextSegment = selectFile.nextToken();
							final TuneFile curFile = new TuneFile(lastFile, nextSegment);
							pathSegs.add(curFile);
							lastFile = curFile;
						}
						final TreePath treePath = new TreePath(pathSegs.toArray(new TuneFile[pathSegs.size()]));
						fFileTree.setSelectionPath(treePath);
						SwingUtilities.invokeLater(new Runnable() {

							public void run() {
								fFileTree.scrollPathToVisible(fFileTree.getSelectionPath());
							}

						});
					}
					// System.err.println("select: " + treePath);
				}
			}
		} catch (final Exception e) {
			e.printStackTrace();
		}
	}

	protected void startSearch(boolean forceRecreate) {
		if (fSearchThread != null && fSearchThread.isAlive()) {
			return;
		}

		if (!new File(fRootDir.getText()).exists()) {
			return;
		}

		fSidplay2.setProgressBar(0);
		progCount = 0;
		
		if (forceRecreate && fConnection != null) {
			try {
				// shutdown previous database
				fConnection.close();
				// trigger a new connection to drop previous tables
				fConnection = null;
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		if (fConnection == null) {
			fConnection = SearchIndexCreator.getConnection(fRootDir.getText(), forceRecreate);
		}
		
		/* validate database: version is inserted only after successful create completes. */
		if (! forceRecreate) {
			try {
				Statement s = fConnection.createStatement();
				ResultSet rs = s.executeQuery("SELECT count(*) FROM version");
				if (!rs.next() || rs.getInt(1) == 0) {
					throw new SQLException("Database is not complete, recreating");
				}
			} catch (SQLException e) {
				System.err.print(e.getMessage());
				forceRecreate = true;
				fConnection = SearchIndexCreator.getConnection(fRootDir.getText(), forceRecreate);
			}
		}

		if (forceRecreate) {
			final File root = new File(fRootDir.getText());
			fSearchThread = new SearchIndexerThread(root);
			fSearchThread.addSearchListener(this);
			fSearchThread.addSearchListener(new SearchIndexCreator(root, fSidplay2, fConnection));

			fSearchThread.start();
		} else {
			switch (fResultBox.getSelectedIndex()) {
			case 1:
				// Add result to favorites?
				// Create new favorites tab
				fSidplay2.support.firePropertyChange(IAddFavoritesTab.class.getName(), null, new IAddFavoritesTab() {

					public String getTitle() {
						return "New Tab";
					}

					public void setFavorites(final IFavorites favorites) {
						fSearchResultFavorites = favorites;
					}
				});
				break;

			default:
				break;
			}

			final SearchInIndexThread t = new SearchInIndexThread(fRootDir.getText(), fConnection, fScopeBox.getSelectedIndex() != 1);
			t.addSearchListener(this);
			t.setFieldName((String) fCriteriaBox.getSelectedItem());
			t.setFieldValue(fSearchFor.getText());
			t.setCaseSensitive(false);
			if (fChangedOptions) {
				resetSearch();
				fChangedOptions = false;
			}
			fSearchThread = t;
			fSearchThread.restoreSearchState(fSavedState);
			fSearchThread.start();
		}

	}

	protected void resetSearch() {
		fSavedState = null;
	}

	protected void convertFile(final File tuneFile) {
		SidTuneConverter c = new SidTuneConverter(fSidplay2);
		c.convertFiles(new File[] { tuneFile }, fLastDir);
	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
		if (m_tune == null) {
			return;
		}
		// auto-expand current selected tune
		showNextHit(new File(m_tune.getInfo().filename));
	}
}
