/**
 * 
 */
package applet.fileviewer.search;

import java.io.File;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;

import libsidutils.STIL.Info;
import libsidutils.STIL.STILEntry;
import libsidutils.STIL.TuneEntry;
import applet.JSIDPlay2;
import applet.sidtune.SidTuneInfoCache;

public final class SearchIndexCreator implements ISearchListener {

	static {
		try {
			Class.forName("org.hsqldb.jdbcDriver");
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			throw new ExceptionInInitializerError(e);
		}
	}
	
	/* bump this each time incompatible change in database occurs. */
	private static final int correctVersion = 3;

	public static final String JSIDPLAY2_IDX = "jsidplay2.idx";

	private final SidTuneInfoCache infoCache;

	private final Connection conn;

	private final File fRoot;

	private PreparedStatement collectionInsertStatement, stilInsertStatement;

	private final JSIDPlay2 fSidplay2;

	private int count;

	public SearchIndexCreator(final File rootDir, final JSIDPlay2 sidplay2, final Connection conn) {
		fRoot = rootDir;
		infoCache = new SidTuneInfoCache(sidplay2);
		fSidplay2 = sidplay2;
		this.conn = conn;

		prepareStatements();
	}

	private void prepareStatements() {
		try {
			String insert = "INSERT INTO collection VALUES (?, ?"; // Full Path,
			// File Name
			for (@SuppressWarnings("unused") final
					String field : SidTuneInfoCache.SIDTUNE_INFOS) {
				insert += ", ?";
			}
			insert += ", ?"; // STIL glb. Comment
			insert += ")";
			collectionInsertStatement = conn.prepareStatement(insert);

			insert = "INSERT INTO stil VALUES (?";
			for (int i = 0; i < 5; i++) {
				insert += ", ?";
			}
			insert += ")";
			stilInsertStatement = conn.prepareStatement(insert);

		} catch (final SQLException e) {
			throw new RuntimeException(e);
		}
	}

	public static final Connection getConnection(final String root, final boolean create) {
		final File idxFile = new File(root, JSIDPLAY2_IDX);
		// System.err.println("Connect to: " + idxFile.getAbsolutePath());
		try {
			final Connection connection = DriverManager.getConnection("jdbc:hsqldb:file:" + idxFile.getAbsolutePath() + ";shutdown=true");
			connection.setAutoCommit(false);
			initDatabase(connection, create);
			return connection;
		} catch (final SQLException e) {
			throw new RuntimeException(e);
		}
	}

	private static void initDatabase(final Connection conn, final boolean forceDrop) throws SQLException {
		if (!forceDrop) {
			try {
				final ResultSet versionSet = conn.prepareStatement("SELECT number FROM version").executeQuery();
				if (versionSet.next() && versionSet.getInt(1) == correctVersion) {
					return;
				}
			} catch (final SQLException e) {
				System.err.println(e.getMessage());
			}
			/* invalid version. We probably have to make a new database */
		}

		final ResultSet tables = conn.getMetaData().getTables(null, null, null, new String[] { "TABLE", "VIEW" });
		while (tables.next()) {
			try {
				String schema = tables.getString("TABLE_SCHEM");
				if (schema == null) {
					schema = "";
				}
				final String name = tables.getString("TABLE_NAME");
				conn.prepareStatement("DROP TABLE " + ("".equals(schema) ? "" : schema + ".") + name).execute();
			} catch (final SQLException e) {
				throw new RuntimeException(e);
			}
		}
		
		conn.prepareStatement("SET PROPERTY \"hsqldb.cache_scale\" 9").execute();
		
		/* store current version */
		conn.prepareStatement("CREATE CACHED TABLE version (number INTEGER)").execute();

		String create = "CREATE CACHED TABLE collection (\"Full Path\" varchar(256), \"File Name\" varchar(256)";
		for (final String s : SidTuneInfoCache.SIDTUNE_INFOS) {
			create += ",  \"" + s + "\" varchar(256)";
		}
		create += ", \"STIL glb. Comment\" varchar(4096)";
		create += ", PRIMARY KEY (\"Full Path\"))";
		conn.prepareStatement(create).execute();

		create = "CREATE CACHED TABLE stil (\"Full Path\" varchar(256)";
		for (final String s : new String[] { "STIL Name", "STIL Author", "STIL Title", "STIL Artist", "STIL Comment" }) {
			create += ", \"" + s + "\" varchar(2048)";
		}
		create += ")";
		conn.prepareStatement(create).execute();
		
		conn.commit();
	}

	public void searchStart() {
	}

	public void searchHit(final File matchFile) {
		count ++;
		try {
			final String relativePath = makeRelative(matchFile);
			collectionInsertStatement.setString(1, relativePath);
			collectionInsertStatement.setString(2, matchFile.getName());
			final Object[] infos = matchFile.isFile() ? infoCache.getInfo(matchFile.getAbsolutePath()) : null;
			for (int i = 0; i < SidTuneInfoCache.SIDTUNE_INFOS.length; i++) {
				collectionInsertStatement.setString(i + 3, infos != null && infos[i] != null ? infos[i].toString() : null);
			}

			/* put STIL Global Comment into the title, if available */
			final STILEntry stilEntry = fSidplay2.getSTIL(matchFile.getAbsolutePath());
			collectionInsertStatement.setString(3 + SidTuneInfoCache.SIDTUNE_INFOS.length, stilEntry != null ? stilEntry.globalComment : null);
			if (stilEntry != null) {
				/* add tune infos */
				addSTILInfo(relativePath, stilEntry.infos);

				/* go through subsongs & add them as well */
				for (final TuneEntry entry : stilEntry.subtunes) {
					addSTILInfo(relativePath, entry.infos);
				}
			}

			collectionInsertStatement.execute();
		
			if ((count % 2000) == 0) {
				conn.commit();
				conn.prepareStatement("CHECKPOINT").execute();
			}
		} catch (final Exception e) {
			System.err.println("Indexing failure on: " + matchFile.getAbsolutePath()
					+ ": " + e.getMessage());
		}
	}

	private void addSTILInfo(final String relativePath, final ArrayList<Info> infos)
	throws SQLException {
		stilInsertStatement.setString(1, relativePath);
		for (final Info info : infos) {
			stilInsertStatement.setString(2, info.name);
			stilInsertStatement.setString(3, info.author);
			stilInsertStatement.setString(4, info.title);
			stilInsertStatement.setString(5, info.artist);
			stilInsertStatement.setString(6, info.comment);
		}
		stilInsertStatement.execute();
	}

	private String makeRelative(final File matchFile) throws Exception {
		final String canonicalPath = matchFile.getCanonicalPath();
		final String rootCanonicalPath = fRoot.getCanonicalPath();
		if (canonicalPath.startsWith(rootCanonicalPath)) {
			final String name = canonicalPath.substring(rootCanonicalPath.length()).replace(File.separatorChar, '/');
			if (name.startsWith("/")) {
				return name.substring(1);
			}
		}
		return matchFile.getCanonicalPath();
	}

	public void searchStop(final boolean canceled) {
		try {
			if (!canceled) {
				conn.prepareStatement("INSERT INTO version VALUES (" + correctVersion + ")").execute();
				conn.commit();
				conn.prepareStatement("CHECKPOINT DEFRAG").execute();
			} else {
				conn.rollback();
			}
		} catch (final SQLException e) {
			throw new RuntimeException(e);
		}
	}
}