package applet.sidtune;

import java.io.File;
import java.text.DateFormat;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Map.Entry;

import libsidplay.sidtune.SidTune;
import libsidplay.sidtune.SidTuneInfo;
import applet.JSIDPlay2;

@SuppressWarnings("serial")
public class SidTuneInfoCache {

	public static final String[] SIDTUNE_INFOS = new String[] { "Title",
		"Author", "Released", "Format", "Player ID", "No. of Songs",
		"Start song", "Clock Freq.", "Speed", "SID", "Compatibility",
		"Tune length", "Audio", "SID chip base 1", "SID chip base 2",
		"Driver Address", "Load Address", "Init Address", "Player Address",
		"File Date", "File size (kb)", "Tune size (b)",
		"Reloc. start page", "Reloc. no. pages" };

	public static Class<?> SIDTUNE_TYPES[] = new Class[] { String.class,
		String.class, String.class, String.class, String.class,
		Integer.class, Integer.class, String.class, String.class,
		String.class, String.class, String.class, String.class,
		String.class, String.class, String.class, String.class,
		String.class, String.class, String.class, Date.class,
		Integer.class, Integer.class, String.class, String.class, };

	public static int INFO_TITLE = 0;

	public static int INFO_AUTHOR = 1;

	public static int INFO_RELEASED = 2;

	public static int INFO_FORMAT = 3;

	public static int INFO_PLAYER_ID = 4;

	public static int INFO_NO_OF_SONGS = 5;

	public static int INFO_START_SONG = 6;

	public static int INFO_CLOCK_FREQ = 7;

	public static int INFO_SPEED = 8;

	public static int INFO_SID = 9;

	public static int INFO_COMPATIBILITY = 10;

	public static int INFO_TUNE_LENGTH = 11;

	public static int INFO_AUDIO = 12;

	public static int INFO_SID_CHIP_BASE_1 = 13;

	public static int INFO_SID_CHIP_BASE_2 = 14;

	public static int INFO_DRIVER_ADDRESS = 15;

	public static int INFO_LOAD_ADDRESS = 16;

	public static int INFO_INIT_ADDRESS = 17;

	public static int INFO_PLAYER_ADDRESS = 18;

	public static int INFO_FILE_DATE = 19;

	public static int INFO_FILE_SIZE = 20;

	public static int INFO_TUNE_SIZE = 21;

	public static int INFO_RELOC_START_PAGE = 22;

	public static int INFO_RELOC_NO_PAGES = 23;

	private final JSIDPlay2 fSidplay2;

	private static final LinkedHashMap<String, Object[]> pathNamesToTune = new LinkedHashMap<String, Object[]>() {
		private static final int MAX_ENTRIES = 128;

		@Override
		protected boolean removeEldestEntry(final Entry<String, Object[]> eldest) {
			return size() > MAX_ENTRIES;
		}
	};

	public SidTuneInfoCache(final JSIDPlay2 sidplay2) {
		this.fSidplay2 = sidplay2;
	}

	public Object[] getInfo(final String pathName) {
		Object[] infos = pathNamesToTune.get(pathName);
		if (infos == null) {
			infos = new Object[SIDTUNE_INFOS.length];
			pathNamesToTune.put(pathName, infos);
			try {
				final SidTune tune = SidTune.load(pathName);
				if (tune != null) {
					// select first sub-tune to gather sub-song specific infos
					// like song speed
					tune.selectSong(1);

					for (int i = 0; i < SIDTUNE_INFOS.length; i++) {
						infos[i] = getInfo(tune, pathName, i);
					}
				}
			}
			catch (final Exception e) {
				e.printStackTrace();
			}
		}

		return infos;
	}

	private Object getInfo(final SidTune sidTune, final String pathName, final int infoType) {
		final SidTuneInfo tuneInfo = sidTune.getInfo();
		switch (infoType) {
		case 0:
		case 1:
		case 2:
			return getInfos(sidTune)[infoType];

		case 3:
			return sidTune.getClass().getSimpleName();

		case 4:
			return getPlayer(sidTune);

		case 5:
			return tuneInfo.songs;

		case 6:
			return tuneInfo.startSong;

		case 7:
			return tuneInfo.clockSpeed;

		case 8:
			return tuneInfo.songSpeed;

		case 9:
			return tuneInfo.sidModel;

		case 10:
			return tuneInfo.compatibility;

		case 11:
			return getFullSongLength(fSidplay2.getFullSongLength(sidTune));

		case 12:
			return getStereo(sidTune.isStereo());

		case 13:
			return String.format("$%04x", tuneInfo.sidChipBase1);

		case 14:
			return String.format("$%04x", tuneInfo.sidChipBase2);

		case 15:
			return getDriverAddress(tuneInfo);

		case 16:
			return getLoadAddress(tuneInfo);

		case 17:
			return getInitAddress(tuneInfo);

		case 18:
			return getPlayerAddress(tuneInfo);

		case 19: {
			final File tuneFile = new File(pathName);
			return getDate(tuneFile.lastModified());
		}
		case 20: {
			final File tuneFile = new File(pathName);
			return getKilos(tuneFile.length());
		}
		case 21:
			return tuneInfo.dataFileLen;

		case 22:
			return String.format("$%02x", tuneInfo.relocStartPage);

		case 23:
			return String.format("$%02x", tuneInfo.relocPages);

		default:
			return "";
		}
	}

	private String[] getInfos(final SidTune currentTune) {
		final SidTuneInfo tuneInfo = currentTune.getInfo();
		final String infos[] = new String[] { "?", "?", "?" };
		if (tuneInfo.numberOfInfoStrings > 0) {
			if (tuneInfo.numberOfInfoStrings == 3) {
				for (int i = 0; i < 3; i++) {
					if (tuneInfo.infoString[i] != null) {
						infos[i] = tuneInfo.infoString[i];
					}
				}
			} else {
				for (int i = 0; i < Math.min(tuneInfo.numberOfInfoStrings, 3); i++) {
					if (tuneInfo.infoString[i] != null) {
						infos[i] = tuneInfo.infoString[i];
					}
				}
			}
		}

		return infos;
	}

	private String getDriverAddress(final SidTuneInfo tuneInfo) {
		final StringBuffer line = new StringBuffer();
		if (tuneInfo.determinedDriverAddr == 0) {
			line.append("not present");
		} else {
			line.append(String.format("$%04x", tuneInfo.determinedDriverAddr));
			line.append(String.format("-$%04x", tuneInfo.determinedDriverAddr
					+ tuneInfo.determinedDriverLength - 1));
		}
		return line.toString();
	}

	private String getLoadAddress(final SidTuneInfo tuneInfo) {
		final StringBuffer line = new StringBuffer();
		line.append(String.format("$%04x", tuneInfo.loadAddr));
		line.append(String.format("-$%04x", tuneInfo.loadAddr
				+ tuneInfo.c64dataLen - 1));
		return line.toString();
	}

	private String getInitAddress(final SidTuneInfo tuneInfo) {
		final StringBuffer line = new StringBuffer();
		if (tuneInfo.playAddr == 0xffff) {
			line.append(String.format("SYS $%04x", tuneInfo.initAddr));
		} else {
			line.append(String.format("$%04x", tuneInfo.initAddr));
		}
		return line.toString();
	}

	private String getPlayerAddress(final SidTuneInfo tuneInfo) {
		final StringBuffer line = new StringBuffer();
		if (tuneInfo.playAddr != 0xffff) {
			line.append(String.format("$%04x", tuneInfo.playAddr));
		}
		return line.toString();
	}

	private String getStereo(final boolean stereo) {
		return stereo ? "Stereo" : "Mono";
	}

	private String getPlayer(SidTune tune) {
		StringBuilder ids = new StringBuilder();
		for (String s : tune.identify()) {
			if (ids.length() > 0) {
				ids.append(", ");
			}
			ids.append(s);
		}
		return ids.toString();
	}

	private String getFullSongLength(final int songLength) {
		if (songLength < 1) {
			return "Unknown";
		} else {
			return String.format("%02d:%02d", (songLength / 60 % 100),
					(songLength % 60));
		}
	}

	private Date getDate(final long lastModified) {
		return new Date(lastModified) {
			@Override
			public String toString() {
				return DateFormat.getDateInstance(DateFormat.SHORT)
				.format(this);
			}
		};
	}

	private Integer getKilos(final long length) {
		return Math.round(length / 1024);
	}
}
