/*
 * DKnob2.java
 * (c) 2005 by Joakim Eriksson
 *
 * DKnob is a component similar to JSlider but with
 * round "user interface", a knob.
 */
package applet.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Arc2D;

import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

public class DKnob2 extends JComponent {
	private final static float START = 225;
	private final static float LENGTH = 270;
	private final static float PI = (float) 3.1415;
	private final static float START_ANG = START / 360 * PI * 2;
	private final static float LENGTH_ANG = LENGTH / 360 * PI * 2;
	//  private final static float DRAG_RES = (float) 0.01;
	private final static float MULTIP = 180 / PI;
	private final static Color DEFAULT_FOCUS_COLOR = new Color(0x8080ff);

	public static final Color DARK = new Color(0x40, 0x40, 0x40, 0xe0);
	public static final Color DARK_T = new Color(0x40, 0x40, 0x40, 0x80);
	public static final Color DARK_L = new Color(0x80, 0x80, 0x80, 0xa0);
	public static final Color LIGHT_D = new Color(0xa0, 0xa0, 0xa0, 0xa0);
	public static final Color LIGHT = new Color(0xc0, 0xc0, 0xc0, 0xa0);
	public static final Color LIGHT_T = new Color(0xc0, 0xc0, 0xc0, 0x80);

	//  private int SHADOWX = 1;
	//  private int SHADOWY = 1;
	private final float CLICK_SPEED;
	private int size;
	protected int middle;

	private String label = "";
	private String type = "";

	private final Polygon knob = new Polygon();

	private final static Dimension MIN_SIZE = new Dimension(30, 40);
	private final static Dimension PREF_SIZE = new Dimension(40, 60);

	// Set the antialiasing to get the right look!
	private final static RenderingHints AALIAS = new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

	private ChangeEvent changeEvent = null;
	private final EventListenerList listenerList = new EventListenerList();

	protected Arc2D hitArc = new Arc2D.Float(Arc2D.PIE);

	protected float ang = START_ANG;
	protected float val;

	private int lowInt = 0;
	private int highInt = 100;
	private float divisor = 0;

	private final Color focusColor;

	public DKnob2(final String label) {
		this(label, "");
	}

	public DKnob2(final String label, final String type) {
		CLICK_SPEED = 0.01F;
		//    SHADOWX = 1;
		//    SHADOWY = 1;

		this.label = label;
		this.type = type;

		for (int i = 0, n = 120; i < n; i++) {
			final int y = 20 + (int) (0.5 + (Math.sin(i * 3.141592 / 4) * 1.5 + 20) * Math.sin(i * 3.141592 * 2.0 / n));
			final int x = 20 + (int) (0.5 + (Math.sin(i * 3.141592 / 4) * 1.5 + 20) * Math.cos(i * 3.141592 * 2.0 / n));
			knob.addPoint(x, y);
		}

		focusColor = DEFAULT_FOCUS_COLOR;

		setPreferredSize(PREF_SIZE);
		setFocusable(true);

		hitArc.setAngleStart(235); // Degrees ??? Radians???
		addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(final MouseEvent me) {
				requestFocus();
			}

			@Override
			public void mouseClicked(final MouseEvent me) {
				if (!isEnabled()) {
					return;
				}
				hitArc.setAngleExtent(-(LENGTH + 20));
				if (hitArc.contains(me.getX(), me.getY())) {
					hitArc.setAngleExtent(MULTIP * (ang - START_ANG) - 10);
					if (hitArc.contains(me.getX(), me.getY())) {
						decValue();
					} else {
						incValue();
					}
				}
			}
		});

		// Let the user control the knob with the mouse
		addMouseMotionListener(new MouseMotionAdapter() {
			@Override
			public void mouseDragged(final MouseEvent me) {
				if (!isEnabled()) {
					return;
				}
				// Measure relative the middle of the button!
				final int xpos = middle - me.getX();
				final int ypos = middle - me.getY();
				double newAng = Math.atan2(ypos, xpos) + START_ANG - Math.PI;
				if (newAng < -START_ANG + Math.PI) {
					newAng += 2 * Math.PI;
				}
				newAng /= LENGTH_ANG;
				setFractionalValue((float) newAng);
			}

			@Override
			public void mouseMoved(final MouseEvent me) {
				if (!isEnabled()) {
					return;
				}
			}
		});

		// Let the user control the knob with the keyboard
		addKeyListener(new KeyListener() {

			public void keyTyped(final KeyEvent e) {
			}

			public void keyReleased(final KeyEvent e) {
			}

			public void keyPressed(final KeyEvent e) {
				if (!isEnabled()) {
					return;
				}
				final int k = e.getKeyCode();
				if (k == KeyEvent.VK_RIGHT) {
					incValue();
				} else if (k == KeyEvent.VK_LEFT) {
					decValue();
				}
			}
		});

		// Handle focus so that the knob gets the correct focus highlighting.
		addFocusListener(new FocusListener() {
			public void focusGained(final FocusEvent e) {
				repaint();
			}

			public void focusLost(final FocusEvent e) {
				repaint();
			}
		});
	}

	protected void incValue() {
		setFractionalValue(val + CLICK_SPEED);
	}

	protected void decValue() {
		setFractionalValue(val - CLICK_SPEED);
	}

	public int getIntValue() {
		return lowInt + Math.round(val * (highInt - lowInt));
	}

	public void setIntValue(final int val) {
		setFractionalValue((float) (val - lowInt) / (highInt - lowInt));
	}

	public float getValue() {
		return getIntValue() / getDivisor();
	}

	public void setValue(final float val) {
		setIntValue(Math.round(val * getDivisor()));
	}

	private void setFractionalValue(float val) {
		if (val < 0) {
			val = 0;
		}
		if (val > 1) {
			val = 1;
		}
		this.val = val;
		ang = START_ANG - LENGTH_ANG * val;
		repaint();
		fireChangeEvent();
	}

	public void addChangeListener(final ChangeListener cl) {
		listenerList.add(ChangeListener.class, cl);
	}

	public void removeChangeListener(final ChangeListener cl) {
		listenerList.remove(ChangeListener.class, cl);
	}

	@Override
	public Dimension getMinimumSize() {
		return MIN_SIZE;
	}

	protected void fireChangeEvent() {
		// Guaranteed to return a non-null array
		final Object[] listeners = listenerList.getListenerList();
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == ChangeListener.class) {
				// Lazily create the event:
				if (changeEvent == null) {
					changeEvent = new ChangeEvent(this);
				}
				((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);
			}
		}
	}

	// Paint the DKnob
	@Override
	public void paint(final Graphics g) {
		if (g instanceof Graphics2D) {
			final Graphics2D g2d = (Graphics2D) g;

			final int width = getWidth();
			final int height = getHeight() - 10;
			size = Math.min(width, height) - 5;
			middle = size / 2;

			g2d.setBackground(getParent().getBackground());
			g2d.addRenderingHints(AALIAS);

			// For the size of the "mouse click" area
			hitArc.setFrame(1, 1, size + 12, size + 12);

			g2d.scale(size / 30.0, size / 30.0);
			drawKnob(g2d);
		}
	}

	private void drawKnob(final Graphics2D g2d) {
		g2d.scale(0.5, 0.5);
		g2d.drawString(label, 2, 10);
		g2d.scale(2, 2);
		//    g2d.translate(0,6);
		final int yoff = 6;

		if (hasFocus()) {
			g2d.setColor(focusColor);
			g2d.fill(new Arc2D.Double(0.0, 0.0 + yoff, 27.0 - yoff, 27.0 - yoff, 90.0, 360.0, Arc2D.OPEN));
		}
		if (isEnabled()) {
			g2d.setColor(DARK);
		} else {
			g2d.setColor(LIGHT);
		}
		g2d.fill(new Arc2D.Double(1.0, 1.0 + yoff, 25.0 - yoff, 25.0 - yoff, 90.0, 360.0, Arc2D.OPEN));
		g2d.setColor(DARK_L);
		g2d.fill(new Arc2D.Double(2.0, 2.0 + yoff, 22.0 - yoff, 22.0 - yoff, 90.0, 360.0, Arc2D.OPEN));
		g2d.setColor(LIGHT_D);
		g2d.draw(new Arc2D.Double(3.0, 3.0 + yoff, 21.0 - yoff, 21.0 - yoff, 75.0, 120.0, Arc2D.OPEN));
		g2d.setColor(LIGHT);
		g2d.draw(new Arc2D.Double(4.0, 4.0 + yoff, 20.0 - yoff, 20.0 - yoff, 122.5, 25.0, Arc2D.OPEN));

		g2d.setColor(DARK);
		g2d.fill(new Arc2D.Double(5.0, 5.0 + yoff, 18.0 - yoff, 18.0 - yoff, 90.0, 360.0, Arc2D.OPEN));

		g2d.setColor(DARK_T);
		g2d.fill(new Arc2D.Double(9.0, 9.0 + yoff, 18.0 - yoff, 18.0 - yoff, 90.0, 360.0, Arc2D.OPEN));
		g2d.setColor(DARK_L);
		g2d.fill(new Arc2D.Double(6.0, 6.0 + yoff, 16.0 - yoff, 16.0 - yoff, 90.0, 360.0, Arc2D.OPEN));
		g2d.setColor(DARK_T);
		g2d.draw(new Arc2D.Double(7.0, 7.0 + yoff, 14.0 - yoff, 14.0 - yoff, 270.0, 90.0, Arc2D.OPEN));
		g2d.setColor(LIGHT_D);
		g2d.draw(new Arc2D.Double(7.0, 7.0 + yoff, 15.0 - yoff, 15.0 - yoff, 90.0, 90.0, Arc2D.OPEN));

		g2d.setColor(LIGHT_D);
		g2d.fill(new Arc2D.Double(7.0, 7.0 + yoff, 14.0 - yoff, 14.0 - yoff, 180 * ang / 3.141592 - 60, 120.0, Arc2D.OPEN));
		g2d.setColor(DARK);
		g2d.draw(new Arc2D.Double(6.0, 6.0 + yoff, 16.0 - yoff, 16.0 - yoff, 180 * ang / 3.141592 - 1, 2.0, Arc2D.PIE));

		g2d.setColor(DARK);
		g2d.draw(new Arc2D.Double(5.0, 5.0 + yoff, 18.0 - yoff, 18.0 - yoff, 180 * ang / 3.141592 - 8, 16.0, Arc2D.OPEN));
		g2d.draw(new Arc2D.Double(4.0, 4.0 + yoff, 20.0 - yoff, 20.0 - yoff, 180 * ang / 3.141592 - 4, 8.0, Arc2D.OPEN));

		g2d.scale(0.5, 0.5);
		if (getDivisor() != 0) {
			g2d.drawString(getIntValue() / getDivisor() + " " + type, 10, 64);

		} else {
			g2d.drawString("" + getIntValue() + " " + type, 10, 64);
		}
	}

	public void setInterval(final int low, final int high) {
		this.lowInt = low;
		this.highInt = high;
	}

	// Just used for printing the label...
	public void setDivisor(final float divisor) {
		this.divisor = divisor;
	}

	public float getDivisor() {
		return divisor;
	}

}
