/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation. For the full
 * license text, see http://www.gnu.org/licenses/gpl.html.
 */
package applet.video;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.DirectColorModel;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.SinglePixelPackedSampleModel;
import java.awt.image.WritableRaster;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import libsidplay.components.mos656x.MOS656X;

/**
 * The actual Swing canvas that shows the C64 Screen.
 * 
 * @author Joerg Jahnke (joergjahnke@users.sourceforge.net)
 */
@SuppressWarnings("serial")
class C64Canvas extends Canvas implements PropertyChangeListener {

	// image with C64 screen to display
	private Image screenImage;
	// border color
	private final Color borderColor = new Color(0);

	private MOS656X vic;

	protected C64Canvas() {
		setIgnoreRepaint(true);
	}

	@Override
	public Dimension getPreferredSize() {
		if (vic == null) {
			return new Dimension(0, 0);
		} else {
			return new Dimension(vic.getBorderWidth(), vic.getBorderHeight());
		}
	}

	/**
	 * Set the instance being displayed
	 */
	public void setC64(final MOS656X vic) {
		this.vic = vic;
		vic.addPropertyChangeListener(this);

		final SampleModel sm = new SinglePixelPackedSampleModel(DataBuffer.TYPE_INT,
				vic.getBorderWidth(), vic.getBorderHeight(),
				new int[] { 0x00ff0000, 0x0000ff00, 0x000000ff });
		final DataBufferInt dbi = new DataBufferInt(vic.getPixels(), vic.getBorderHeight() * vic.getBorderWidth());
		final WritableRaster wr = Raster.createWritableRaster(sm, dbi, null);
		final ColorModel cm = new DirectColorModel(24, 0xff0000, 0x00ff00, 0x0000ff);
		screenImage = new BufferedImage(cm, wr, false, null);
		setSize(getPreferredSize());
		repaint();
	}

	@Override
	public void update(final Graphics g) {
		paint(g);
	}

	/**
	 * Repaint the screen. First the background is cleared. Then status messages
	 * are displayed and lastly the contents of the C64 screen are painted.
	 * 
	 * @param g
	 *            graphics context to use
	 */
	@Override
	public void paint(final Graphics g) {
		// show border
		if (screenImage == null) {
			g.setColor(borderColor);
			g.fillRect(0, 0, getWidth(), getHeight());
			return;
		}
		// show emulator screen without the first two pixels
		// (lost by PAL emulation)
		g.drawImage(screenImage, -2, 0, this);
	}

	// implementation of the PropertyChangeListener interface
	public void propertyChange(final PropertyChangeEvent evt) {
		if ("pixels".equals(evt.getPropertyName())) {
			repaint();
		}
	}
}
