package libsidplay;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;

import libsidplay.common.Event;
import libsidplay.common.EventScheduler;
import libsidplay.common.IReSIDExtension;
import libsidplay.common.ISID2Types.Clock;
import libsidplay.common.SIDEmu;
import libsidplay.components.c1530.DatasetteEnvironment;
import libsidplay.components.c1541.C1541Environment;
import libsidplay.components.cart.Cartridge;
import libsidplay.components.cart.supported.REU;
import libsidplay.components.mos6510.IMOS6510Extension;
import libsidplay.components.mos6510.MOS6510;
import libsidplay.components.mos6526.CIAEnvironment;
import libsidplay.components.mos6526.MOS6526;
import libsidplay.components.mos656x.MOS656X;
import libsidplay.components.pla.Bank;
import libsidplay.components.pla.PLA;
import libsidplay.components.ram.SystemRAMBank;

public abstract class C64 implements CIAEnvironment, DatasetteEnvironment, C1541Environment {
	public static final long BOOT_TIME = 2141000;

	/** System clock */
	private Clock clock;

	/** MMU chip */
	private final PLA pla;
	
	/** CPU */
	private final MOS6510 cpu;

	/** CIA1 */
	protected final MOS6526 cia1;

	/** CIA2 */
	protected final MOS6526 cia2;

	/** VIC */
	private final MOS656X vic;

	/** System memory array */
	protected final SystemRAMBank ramBank = new SystemRAMBank();
	
	/** System event context */
	private final EventScheduler context;

	/** Number of entrances to play routine */
	private int callsToPlayRoutine = 0;

	/** Playroutine address */
	private int playAddr;

	/** The interested party for playroutine entrances. */
	private IMOS6510Extension playRoutineObserver;

	protected class SIDBank extends Bank {
		private final static int SID2_MAPPER_SIZE = 32;

		private final SIDEmu[] sidemu = new SIDEmu[Player.SID2_MAX_SIDS];
		
		/**
		 * Mapping table in d400-d7ff
		 */
		private final int sidmapper[] = new int[32];

		/**
		 * SID listener for detecting SID writes
		 */
		private final IReSIDExtension[] sidWriteListener = new IReSIDExtension[Player.SID2_MAX_SIDS];

		public void reset() {
			for (SIDEmu s : sidemu) {
				if (s != null) {
					s.reset((byte) 0xf);
				}
			}
			Arrays.fill(sidWriteListener, null);
		}
	
		public void setSIDMapping(int address, int chipNum) {
			sidmapper[address >> 5 & SID2_MAPPER_SIZE - 1] = chipNum;
		}

		public void setSidWriteListener(int i, IReSIDExtension listener) {
			sidWriteListener[i] = listener;
		}
		
		@Override
		public byte read(int address) {
			final int i = sidmapper[address >> 5 & SID2_MAPPER_SIZE - 1];
			if (sidemu[i] != null) {
				return sidemu[i].read(address & 0x1f);
			} else {
				return (byte) 0xff;
			}
		}
		
		@Override
		public void write(int address, byte value) {
			final int i = sidmapper[address >> 5 & SID2_MAPPER_SIZE - 1];
			if (sidemu[i] != null) {
				sidemu[i].write(address & 0x1f, value);
			}
			if (sidWriteListener[i] != null) {
				final long time = context.getTime(Event.Phase.PHI2);
				sidWriteListener[i].write(time, 0, address & 0x1f, value);
			}
		}

		public SIDEmu getSID(int i) {
			return sidemu[i];
		}

		public void setSID(int i, SIDEmu s) {
			sidemu[i] = s;
		}
	}
	private final SIDBank sidBank = new SIDBank();
	
	
	/**
	 * Area backed by RAM, including cpu port addresses 0 and 1.
	 * 
	 * This is bit of a fake. We know that the CPU port is an internal
	 * detail of the CPU, and therefore CPU should simply pay the price
	 * for reading/writing to 0/1.
	 * 
	 * However, that would slow down all accesses, which is suboptimal. Therefore
	 * we install this little hook to the 4k 0 region to deal with this.
	 * 
	 * @author Antti Lankila
	 */
	protected class ZeroRAMBank extends Bank {
		/* Value written to processor port.  */
		private byte dir;
		private byte data;

	    /* Value read from processor port.  */
	    private byte data_read;

	    /* State of processor port pins.  */
	    private byte data_out;

	    // TODO some weird stuff with top bits 6 and 7 which aren't connected on C64

		/* Tape motor status.  */
		private byte old_port_data_out;
		/* Tape write line status.  */
		private byte old_port_write_bit;

		public void reset() {
			ramBank.reset();
			old_port_data_out = (byte) 0xff;
			old_port_write_bit = (byte) 0xff;
			data = 0x3f;
		    data_out = 0x3f;
		    data_read = 0x3f;
		    dir = 0;
		}

		private void updateCpuPort() {
			data_out = (byte) ((data_out & ~dir) | (data & dir));
			data_read = (byte) ((data | ~dir) & (data_out | 0x17));
			pla.setCpuPort(data_read);

			if (0 == (dir & 0x20)) {
				data_read &= 0xdf;
			}
			if (0 == (dir & 0x10) && C64.this.getTapeSense()) {
				data_read &= 0xef;
			}

			if (((dir & data) & 0x20) != old_port_data_out) {
				old_port_data_out = (byte) ((dir & data) & 0x20);
				C64.this.setMotor(0 == old_port_data_out);
			}

			if (((~dir | data) & 0x8) != old_port_write_bit) {
				old_port_write_bit = (byte) ((~dir | data) & 0x8);
				C64.this.toggleWriteBit(((~dir | data) & 0x8) != 0);
			}
		}

		@Override
		public byte read(int address) {
			if (address == 0) {
				return dir;
			} else if (address == 1) {
				return data_read;
			} else {
				return ramBank.read(address);
			}
		}
		
		@Override
		public void write(int address, byte value) {
			if (address == 0) {
				dir = value;
				updateCpuPort();
				value = vic.getLastReadByte();
			} else if (address == 1) {
				data = value;
				updateCpuPort();
				value = vic.getLastReadByte();
			}
			ramBank.write(address, value);
		}
	}
	private final ZeroRAMBank zeroRAMBank = new ZeroRAMBank();

	/**
	 * Set playroutine address to watch by CPU emulation.
	 * 
	 * @param playAddr
	 */
	public void setPlayAddr(int playAddr) {
		this.playAddr = playAddr;
	}
	
	public final int callsToPlayRoutineSinceLastTime() {
		final int val = callsToPlayRoutine;
		callsToPlayRoutine = 0;
		return val;
	}

	public C64() {
		context = new EventScheduler();

		pla = new PLA(context, sidBank, zeroRAMBank, ramBank);

		cpu = new MOS6510(context) {
			@Override
			public byte cpuRead(int address) {
				return pla.cpuRead(address);
			}

			@Override
			public void cpuWrite(int address, byte value) {
				pla.cpuWrite(address, value);
			}
			
			@Override
			protected void doJSR() {
				super.doJSR();
				if (Register_ProgramCounter == playAddr) {
					if (playRoutineObserver != null) {
						final long time = context.getTime(Event.Phase.PHI2);
						playRoutineObserver.fetch(time);
					}

					callsToPlayRoutine ++;
				}
			}
		};
		pla.setCpu(cpu);

		vic = new MOS656X(context) {
			@Override
			protected void interrupt(boolean b) {
				pla.setIRQ(b);
			}
			
			@Override
			protected void setBA(boolean b) {
				pla.setBA(b);
			}

			@Override
			protected byte vicReadColorMemoryPHI2(int address) {
				return pla.vicReadColorMemoryPHI2(address);
			}
			
			@Override
			protected byte vicReadMemoryPHI1(int address) {
				return pla.vicReadMemoryPHI1(address);
			}
			
			@Override
			protected byte vicReadMemoryPHI2(int address) {
				return pla.vicReadMemoryPHI2(address);
			}
		};
		pla.setVic(vic);
		
		cia1 = new MOS6526(context, MOS6526.Model.MOS6526) {
			@Override
			public void interrupt(boolean state) {
				pla.setIRQ(state);
			}
			
			@Override
			public void write(int addr, final byte data) {
				addr &= 0xf;
				super.write(addr, data);

				if (addr == PRB || addr == DDRB) {
					final byte btnAorLP = (byte) ((regs[PRB] | ~regs[DDRB]) & 0x10);
					if (btnAorLP != 0x10) {
						vic.triggerLightpen();
					} else {
						vic.clearLightpen();
					}
				}
			}

			@Override
			public byte read(int addr) {
				addr &= 0xf;
				byte read = super.read(addr);

				if (addr == PRA || addr == PRB) {
					final byte joy1 = C64.this.readJoystick(0);
					final byte joy2 = C64.this.readJoystick(1);
					if (addr == PRA) {
						final byte prb = super.read(PRB);
						read &= C64.this.readKeyboardPRA(prb & joy1) & joy2;
					} else if (addr == PRB) {
						final byte pra = super.read(PRA);
						read &= C64.this.readKeyboardPRB(pra & joy2) & joy1;
					}
				}

				return read;
			}
		};
		pla.setCia1(cia1);
		
		cia2 = new MOS6526(context, MOS6526.Model.MOS6526) {
			private byte old_pa;
			
			@Override
			public void reset() {
				old_pa = (byte) 0xff;
				super.reset();
			}
			
			@Override
			public void interrupt(boolean state) {
				pla.setNMI(state);
			}
			
			@Override
			public void write(int addr, byte data) {
				addr &= 0xf;
				super.write(addr, data);

				if (addr == PRA || addr == DDRA) {
					pla.setVicMemBase((~regs[PRA] & regs[DDRA] & 3) << 14);
				}

				// Floppy support
				data = (byte)(regs[PRA] | ~(regs[DDRA]));
		        if (data != old_pa) {
					C64.this.writeToIECBus((byte) ~data);
		            old_pa = data;
		        }
			}
			
			@Override
			public byte read(int addr) {
				addr &= 0xf;
				if (addr == PRA) {
					return (byte) ((super.read(addr) & 0x3f) | C64.this.readFromIECBus());
				} else {
					return super.read(addr);
				}
			}
			
		};
		pla.setCia2(cia2);
	}

	/**
	 * Perform the equivalent of full power-on reset of C64,
	 * re-initializing everything.
	 */
	public void reset() {
		context.reset();

		pla.reset();
		cpu.triggerRST();
		cia1.reset();
		cia2.reset();
		vic.reset();
		zeroRAMBank.reset();
		
		playAddr = -1;
	}

	/**
	 * Get C64's event scheduler
	 * 
	 * @return the scheduler
	 */
	public EventScheduler getEventScheduler() {
		return context;
	}

	/**
	 * Return the requested SID
	 * 
	 * @param i
	 * @return the SID
	 */
	public SIDEmu getSID(int i) {
		return sidBank.getSID(i);
	}

	/**
	 * Set the requested SID
	 * 
	 * @param i sid number to set
	 * @param sidemu the sid to set
	 */
	public void setSid(int i, SIDEmu sidemu) {
		sidBank.setSID(i, sidemu);
	}

	/**
	 * Return the array backing C64 RAM
	 * 
	 * @return the RAM
	 */
	public byte[] getRAM() {
		return ramBank.array();
	}

	/**
	 * Return CPU emulator
	 * 
	 * @return the cpu
	 */
	public MOS6510 getCPU() {
		return cpu;
	}

	public void setPlayRoutineObserver(IMOS6510Extension observer) {
		playRoutineObserver = observer;
	}

	public void setSidWriteListener(int i, IReSIDExtension listener) {
		sidBank.setSidWriteListener(i, listener);
	}

	public MOS656X getVIC() {
		return vic;
	}

	public void setClock(Clock clock) {
		this.clock = clock;
		
		context.setCyclesPerSecond(clock.getCpuFrequency());
		cia1.setDayOfTimeRate(clock.getCyclesPerFrame());
		cia2.setDayOfTimeRate(clock.getCyclesPerFrame());
		
		if (clock == Clock.PAL) {
			vic.chip(MOS656X.Model.MOS6569R5);
		} else {
			vic.chip(MOS656X.Model.MOS6567R8);
		}
	}

	public Clock getClock() {
		return clock;
	}

	public void setSecondSIDAddress(int sidChipBase2) {
		sidBank.setSIDMapping(0xd400, 0);
		sidBank.setSIDMapping(sidChipBase2, 1);
	}
	
	public void insertCartridge(final String filename)
	throws IOException {
		pla.setCartridge(null);
		if (filename == null) {
			return;
		}

		final InputStream is = new FileInputStream(filename);
		final Cartridge cart;
		if (filename.toLowerCase().endsWith(".reu")) {
			cart = REU.readImage(pla, is);
		} else {
			cart = Cartridge.readImage(pla, is);
		}
		pla.setCartridge(cart);
		is.close();
	}

	public Cartridge getCartridge() {
		return pla.getCartridge();
	}
	
	protected PLA getPLA() {
		return pla;
	}
}
