/**
 *                                 Main Library Code
 *                                SIDs Mixer Routines
 *                             Library Configuration Code
 *                    xa65 - 6502 cross assembler and utility suite
 *                          reloc65 - relocates 'o65' files
 *        Copyright (C) 1997 Andr Fachat (a.fachat@physik.tu-chemnitz.de)
 *        ----------------------------------------------------------------
 *  begin                : Fri Jun 9 2000
 *  copyright            : (C) 2000 by Simon White
 *  email                : s_a_white@email.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 * @author Ken Hndel
 *
 */
package libsidplay;

import hardsid_builder.HardSIDBuilder;
import libsidplay.common.Event;
import libsidplay.common.Event.Phase;
import libsidplay.common.EventScheduler;
import libsidplay.common.ISID2Types.Clock;
import libsidplay.common.ISID2Types.PlayerConfig;
import libsidplay.common.SIDBuilder;
import libsidplay.common.SIDEmu;
import libsidplay.components.c1530.Datasette;
import libsidplay.components.c1541.C1541;
import libsidplay.components.c1541.IECBus;
import libsidplay.components.joystick.IJoystick;
import libsidplay.components.keyboard.Keyboard;
import libsidplay.components.mos6510.IMOS6510Disassembler;
import libsidplay.components.mos6510.MOS6510;
import libsidplay.components.mos6526.MOS6526;
import libsidplay.components.mos656x.MOS656X;
import libsidplay.sidtune.SidTune;
import libsidplay.sidtune.SidTuneInfo;
import resid_builder.ReSID;

public class Player  {
	public final static int SID2_MAX_SIDS = 2;

	private final C64 c64;
	
	private final IECBus iecBus;

	private final C1541Runner c1541Runner;

	private final Keyboard keyboard;

	private final C1541 c1541;

	private final Datasette datasette;
	
	private final IJoystick[] joystickPort = new IJoystick[2];
	
	protected class MixerEvent extends Event {
		/* Scheduling time for next sample mixing event. 2500 is roughly 2.5 ms,
		 * and this rate is needed to support 96 kHz stereo tunes. */
		protected static final int MIXER_EVENT_RATE = 2500;

		private int sampleIndex;
		private float[] sampleBuffer;
		
		protected MixerEvent() {
			super("Mixer");
		}
		
		protected void begin(float[] buffer) {
			sampleIndex = 0;
			sampleBuffer = buffer;
		}
		
		protected boolean notFinished() {
			return sampleIndex != sampleBuffer.length;
		}
		
		protected int samplesGenerated() {
			return sampleIndex;
		}
		
		@Override
		public void event() {
			final SIDEmu chip1 = c64.getSID(0);
			final SIDEmu chip2 = c64.getSID(1);
			
			/* this clocks the SID to the present moment, if it isn't already. */
			chip1.clock();
			final float[] buf1 = chip1.buffer();
			float[] buf2 = null;
			if (chip2 != null) {
				chip2.clock();
				buf2 = chip2.buffer();
			}
			
			/* extract buffer info now that the SID is updated.
			 * clock() may update bufferpos. */
			final int samples = chip1.bufferpos();
			/* If chip2 exists, its bufferpos is expected to be identical to chip1's. */
			
			int i = 0;
			final float lvol = config.leftVolume;
			final float rvol = config.rightVolume;
			while (i < samples) {
				/* Handle whatever output the sid has generated so far */
				if (sampleIndex == sampleBuffer.length) {
					break;
				}
			
				float sample1 = 0;
				if (buf1 != null) {
					sample1 += buf1[i];
				}
				float sample2 = 0;
				if (buf2 != null) {
					sample2 += buf2[i];
				}
				sample1 *= lvol;
				sample2 *= rvol;
				
				/* mono mix. */
				if (buf2 != null && config.channels == 1) {
					sample1 = (sample1 + sample2) * 0.5f;
				}
				/* stereo clone, for people who keep stereo on permanently. */
				if (buf2 == null && config.channels == 2) {
					sample2 = sample1;
				}
			
				sampleBuffer[sampleIndex ++] = sample1;
				if (config.channels == 2) {
					sampleBuffer[sampleIndex ++] = sample2;
				}
				
				i ++;
			}

			/* move unconsumed part of samples from end to start */
			if (buf1 != null) {
				System.arraycopy(buf1, i, buf1, 0, samples - i);
			}
			if (buf2 != null) {
				System.arraycopy(buf2, i, buf2, 0, samples - i);
			}

			chip1.bufferpos(samples - i);
			if (chip2 != null) {
				chip2.bufferpos(samples - i);
			}
			
			/* Post a callback to ourselves. */
			c64.getEventScheduler().schedule(this, MIXER_EVENT_RATE);
		}
	}
	
	private final MixerEvent mixerEvent = new MixerEvent();

	private SidTune tune;

	private PlayerConfig config = new PlayerConfig();

	public void reset() {
		c64.reset();
		keyboard.reset();
		datasette.reset();

		// Reset C1541
		c1541.reset();
		turnDriveOnOff(driveEnabled);

		/* forward C64 2.5 seconds to boot it up */
		{
			SIDEmu s1 = c64.getSID(0);
			SIDEmu s2 = c64.getSID(1);

			c64.setSid(0, null);
			c64.setSid(1, null);

			final boolean[] bootWait = new boolean[] { true };
			final Event bootEvent = new Event("Wait for boot") {
				@Override
				public void event() {
					bootWait[0] = false;
				}
			};
			c64.getEventScheduler().scheduleAbsolute(bootEvent, C64.BOOT_TIME, Event.Phase.PHI1);
			while (bootWait[0]) {
				c64.getEventScheduler().clock();
			}

			/* avoid 2-3 second catchup on next event */
			if (s1 != null) {
				s1.clocksSinceLastAccess();
			}
			if (s2 != null) {
				s2.clocksSinceLastAccess();
			}
			
			c64.setSid(0, s1);
			c64.setSid(1, s2);
		}
		
		c64.getEventScheduler().schedule(mixerEvent, MixerEvent.MIXER_EVENT_RATE, Event.Phase.PHI2);
		
		/* Autostart program, if we have one. */
		if (tune != null) {
			c64.setPlayAddr(tune.getInfo().playAddr);
			
			final int address = tune.placeProgramInMemory(c64.getRAM());
			if (address != -1) {
				/* Set initial volume for psid.
				 * Ideally the driver would do this for us, but whatever...
				 */
				for (int i = 0; i < SID2_MAX_SIDS; i ++) {
					SIDEmu s = c64.getSID(i);
					if (s != null) {
						s.write(0x18, (byte) 0xf);
					}
				}
				c64.getCPU().forcedJump(address);
			} else {
				byte[] ram = c64.getRAM();
				ram[0x30c] = (byte) (tune.getInfo().currentSong - 1);
				ram[0x277] = (byte) 'R';
				ram[0x278] = (byte) 'U';
				ram[0x279] = (byte) 'N';
				ram[0x27a] = (byte) ':';
				ram[0x27b] = 13;
				ram[0xc6] = 5;
			}
		}
	}

	public void mute(final int sidNum, final int voice, final boolean enable) {
		SIDEmu s = c64.getSID(sidNum);
		if (s != null) {
			s.voice(voice, enable);
		}
	}

	public void close() {
		for (int i = 0; i < SID2_MAX_SIDS; i ++) {
			SIDEmu s = c64.getSID(i);
			if (s != null) {
				SIDBuilder b = s.builder();
				b.unlock(s);
				c64.setSid(i, null);
			}
		}
	}

	/**
	 * Set the ICs environment variable to point to this player
	 */
	public Player() {
		c64 = new C64() {
			public byte readFromIECBus() {
				if (driveEnabled) {
					c1541Runner.synchronize();
					return iecBus.readFromIECBus();
				}
				return (byte) 0x80;
			}

			public void writeToIECBus(final byte data) {
				if (driveEnabled) {
					c1541Runner.synchronize();
					iecBus.writeToIECBus(data);
				}
			}
			
			public byte readKeyboardPRA(final int i) {
				return keyboard.readColumnMatrix(i);
			}

			public byte readKeyboardPRB(final int i) {
				return keyboard.readRowMatrix(i);
			}

			public byte readJoystick(final int port) {
				if (joystickPort[port] == null) {
					return (byte) 0xff;
				}
				return joystickPort[port].getValue();
			}
			
			public boolean getTapeSense() {
				return datasette.getTapeSense();
			}
			
			public void setMotor(boolean state) {
				datasette.setMotor(state);
			}
			
			public void toggleWriteBit(boolean state) {
				datasette.toggleWriteBit(state);
			}
			
			public void setFlag() {
				cia1.setFlag();
			}
		};

		keyboard = new Keyboard() {
			@Override
			public void restore() {
				/* in reality, the PLA chip has a RESTORE line, and it pulls the NMI
				 * low. We're modeling the overall NMI and IRQ state in the PLA, so
				 * we can't really do that. We just pulse NMI instead. */
				c64.getPLA().setNMI(true);
				c64.getPLA().setNMI(false);
			}
		};
		
		datasette = new Datasette(c64.getEventScheduler()) {
			@Override
			public void setFlag() {
				c64.setFlag();
			}
		};
		
		iecBus = new IECBus();
		c1541 = new C1541(iecBus, 0);
		iecBus.setFloppies(new C1541[] { c1541 });
		c1541Runner = new SameThreadC1541Runner(c64.getEventScheduler(), c1541.getEventScheduler());
	}
		
	public String getCredits() {
		StringBuffer credit = new StringBuffer();
		// Get component credits
		credit.append("Java Version and User Interface:\n"
				+ "\tCopyright (C) 2007-2010 Ken Hndel\n"
				+ "\thttp://sourceforge.net/projects/jsidplay2/\n");
		credit
		.append("Distortion Simulation and development: Antti S. Lankila\n"
				+ "\thttp://bel.fi/~alankila/c64-sw/\n");
		credit.append("Network SID Device:\n"
				+ "\tSupported by Wilfred Bos, The Netherlands\n"
				+ "\thttp://www.acid64.com\n");
		credit.append("Testing and Feedback: Nata of proNoise\n"
				+ "\thttp://www.nata.netau.net/\n");
		credit.append("graphical output:\n"
				+ "\t(C) 2007 Joakim Eriksson\n"
				+ "\t(C) 2009, 2010 Antti S. Lankila\n");
		credit.append("Drag and Drop support: Robert Harder\n"
				+ "\thttp://iharder.sourceforge.net\n");
		credit
		.append("MP3 downloads from Stone Oakvalley's Authentic SID Collection (SOASC=):\n"
				+ "\thttp://www.6581-8580.com/\n");
		credit.append("PSID to PRG converter (PSID64 v0.8):\n"
				+ "\tCopyright (C) 2001-2007  Roland Hermans\n"
				+ "\thttp://sourceforge.net/projects/psid64/\n");
		credit.append("SID dump file (SIDDump V1.04):\n"
				+ "\tCopyright (C) 2007 Lasse rni\n");
		credit.append("HVSC playroutine identity scanner (SIDId V1.07):\n"
				+ "\tCopyright (C) 2007 Lasse rni\n");
		credit.append("C1541 Floppy Disk Drive Emulation:\n"
				+ "\tCopyright (C) 2010 VICE (the Versatile Commodore Emulator)\n"
				+ "\thttp://www.viceteam.org/\n");
		credit.append("Based on libsidplay v2.1.1 engine:\n"
				+ "\tCopyright (C) 2000 Simon White sidplay2@yahoo.com\n"
				+ "\thttp://sidplay2.sourceforge.net\n");
		credit.append(MOS6510.credits());
		credit.append(MOS6526.credits());
		credit.append(MOS656X.credits());
		credit.append(HardSIDBuilder.credits());
		credit.append(ReSID.credits());
		return credit.toString();
	}

	public final PlayerConfig getConfig() {
		return config;
	}

	/**
	 * Build C64 system capable of playing the loaded tune.
	 * Call this method after load(), because settings requested
	 * by SidTune may affect the system that is being built.
	 * 
	 * @param cfg
	 * @return
	 */
	public void setConfig(final PlayerConfig cfg) {
		/* Determine PAL/NTSC mode */
		Clock cpuFreq = cfg.userClock;
		if (cpuFreq == null) {
			cpuFreq = cfg.defaultClock;
			if (tune != null) {
				SidTuneInfo tuneInfo = tune.getInfo();
				switch (tuneInfo.clockSpeed) {
				case UNKNOWN:
				case ANY:
					cpuFreq = cfg.defaultClock;
					break;
				case PAL:
				case NTSC:
					cpuFreq = Clock.valueOf(tuneInfo.clockSpeed.toString());
					break;
				}
			}
		}
		c64.setClock(cpuFreq);
		c1541Runner.setMainCpuSpeed(cpuFreq);

		/* Determine number of SIDs */
		{
			cfg.channels = 1;
			if (tune != null) {
				final SidTuneInfo m_tuneInfo = tune.getInfo();

				if (m_tuneInfo.sidChipBase2 != 0) {
					cfg.channels = 2;
					c64.setSecondSIDAddress(m_tuneInfo.sidChipBase2);
				}
			}
			if (cfg.forceStereoTune) {
				cfg.channels = 2;
				c64.setSecondSIDAddress(cfg.dualSidBase);
			}
		}
		
		config = cfg;
	}

	/**
	 * Load a program to play.
	 * 
	 * @param tune
	 * @return 0 on success
	 */
	public void setTune(final SidTune tune) {
		this.tune = tune;
	}

	public int play(final float[] buffer) {
		mixerEvent.begin(buffer);

		// Start the player loop
		while (mixerEvent.notFinished()) {
			c64.getEventScheduler().clock();
		}

		return mixerEvent.samplesGenerated();
	}

	public int time() {
		EventScheduler c = c64.getEventScheduler();
		return (int) ((c.getTime(Phase.PHI2) - C64.BOOT_TIME) / c.getCyclesPerSecond());
	}

	public final void setDebug(final IMOS6510Disassembler disassembler) {
		c64.getCPU().setDebug(disassembler);
	}


	/**
	 * Is the floppy disk drive enabled?
	 */
	private boolean driveEnabled;

	public void turnDriveOnOff(final boolean on) {
		driveEnabled = on;
		if (on) {
			c64.getEventScheduler().scheduleThreadSafe(new Event("Begin C64-C1541 sync") {
				@Override
				public void event() {
					c1541Runner.reset();
				}
			});
		} else {
			c64.getEventScheduler().scheduleThreadSafe(new Event("End C64-C1541 sync") {
				@Override
				public void event() {
					c1541Runner.cancel();
				}
			});
		}
	}

	public SIDEmu getSid(final int num) {
		return c64.getSID(num);
	}

	public Keyboard getKeyboard() {
		return keyboard;
	}

	public void setJoystick(final int i, final IJoystick joystickReader) {
		joystickPort[i] = joystickReader;
	}

	public C64 getC64() {
		return c64;
	}

	public C1541 getC1541() {
		return c1541;
	}

	public Datasette getDatasette() {
		return datasette;
	}
}

abstract class C1541Runner extends Event {
	protected final EventScheduler context, c1541Context;
	private int conversionFactor;
	private int accum;
	private long c64Last;
	
	public C1541Runner(EventScheduler context, EventScheduler c1541Context) {
		super("C64 permits C1541 to continue");
		this.context = context;
		this.c1541Context = c1541Context;
	}

	/**
	 * Return the number of clock ticks that 1541 should advance.
	 * 
	 * @return
	 */
	protected int updateSlaveTicks() {
		long oldC64Last = c64Last;
		c64Last = context.getTime(Phase.PHI2);
		int elapsedC64 = (int) (c64Last - oldC64Last);

		int wholeClocks = 0;
		while (elapsedC64 != 0) {
			int chunk = elapsedC64 > 4096 ? 4096 : elapsedC64;
			elapsedC64 -= chunk;
			
			accum += conversionFactor * chunk;
			wholeClocks += accum >> 16;
			accum &= 0xffff;
		}
		
		return wholeClocks;
	}
	
	protected void setMainCpuSpeed(Clock clock) {
		conversionFactor = (int) (1000000.0 / clock.getCpuFrequency() * 65536.0 + 0.5);
	}
	
	protected void reset() {
		c64Last = context.getTime(Phase.PHI2);
	}
	
	abstract protected void cancel();
	abstract protected void synchronize();
}

class SameThreadC1541Runner extends C1541Runner {
	protected boolean notTerminated;
	
	private final Event terminationEvent = new Event("Pause C1541") {
		@Override
		public void event() {
			notTerminated = false;
		}
	};
	
	protected SameThreadC1541Runner(EventScheduler context, EventScheduler c1541Context) {
		super(context, c1541Context);
	}
	
	private void clockC1541Context() {
		int targetTime = updateSlaveTicks();
		if (targetTime == 0) {
			return;
		}
		
		c1541Context.schedule(terminationEvent, targetTime, Event.Phase.PHI2);
		notTerminated = true;
		while (notTerminated) {
			c1541Context.clock();
		}
	}

	@Override
	protected void reset() {
		super.reset();
		context.schedule(this, 0, Event.Phase.PHI2);
	}

	@Override
	protected void cancel() {
		context.cancel(this);
	}

	@Override
	protected void synchronize() {
		clockC1541Context();
	}

	@Override
	public void event() {
		synchronize();
		context.schedule(this, 2000);
	}
}