/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation. For the full
 * license text, see http://www.gnu.org/licenses/gpl.html.
 */
package libsidplay.components.c1541;

import static libsidplay.mem.IC1541_1.C1541_C000;
import static libsidplay.mem.IC1541_2.C1541_E000;

import java.util.Arrays;

import libsidplay.common.Event;
import libsidplay.common.Event.Phase;
import libsidplay.common.EventScheduler;
import libsidplay.components.mos6510.MOS6510;

/**
 * C1541 instance which manages the Floppy CPU and the two VIAs.<br>
 * <br>
 * A good German documentation on the 1541 floppy can be found at <a href='http://www.trikaliotis.net/download/DieFloppy1541-v4.pdf'>http://www.trikaliotis.net/download/DieFloppy1541-v4.pdf</a
 * > or <a href='http://www.softwolves.pp.se/idoc/alternative/vc1541_de/'>http://www.softwolves.pp.se/idoc/alternative/vc1541_de/</a
 * > .<br>
 * Good English ROM listings can be found at <a href='http://www.ffd2.com/fridge/docs/1541dis.html'>http://www.ffd2.com/fridge/docs/1541dis.html</a
 * > or <a href=
 * 'http://www.the-dreams.de/aay1541.txt'>http://www.the-dreams.de/aay1541.txt</a
 * > .
 * 
 * @author Jrg Jahnke (joergjahnke@users.sourceforge.net)
 */
public final class C1541 extends Floppy {
	/**
	 * Disk ID (0-3)
	 */
	private final int id;
	/**
	 * the CPU of this device
	 */
	private final MOS6510 cpu;
	/**
	 * VIA Bus controller
	 */
	protected final VIA6522BC viaBc;
	/**
	 * VIA Disk Controller
	 */
	protected final VIA6522DC viaDc;
	/**
	 * RAM
	 */
	protected final byte[] ram;
	/**
	 * Event context.
	 */
	protected final EventScheduler context;
	/**
	 * Disk image support.
	 */
	private final DiskImage img;
	/**
	 * Disk rotation.
	 */
	protected final Rotation rot;

	/** Number of VIA chips asserting IRQ */
	private int irqCount;

	/** Simplified, joined ROM image. */
	protected static final byte[] rom = new byte[16384];

	static {
		System.arraycopy(C1541_C000, 0, rom, 0,    8192);
		System.arraycopy(C1541_E000, 0, rom, 8192, 8192);
	}
	
	/**
	 * Create a new C1541 instance
	 * 
	 * @param iecBus
	 *            IEC bus
	 * @param id
	 *            floppy ID (0-3)
	 */
	public C1541(final IECBus iecBus, final int id) {
		this.id = id;
		context = new EventScheduler();

		// Create disk image support
		img = new DiskImage(this, getID() + 8);

		// Create disk rotation support.
		rot = new Rotation(this);

		// Setup RAM area
		ram = new byte[0x800];

		// Create a CPU for the floppy disk drive
		cpu = new MOS6510(context) {
			/**
			 * Read from address
			 * 
			 * Implementing chips: 74LS42 (binary-to-decimal decoder, with 4 input
			 * bins of A10, A11, A12, A15 and 10 output values, where:
			 * 0, 1 -> RAM, 6 -> BC, 7 -> DC, giving the mapping 0, 0x400, 0x1800 and 0x1c00
			 * for the A0-A12 lines, and excluding the 74LS42 for A15 lines (ROM).
			 * 
			 * Address lines such as 0x8000 - 0xffff result in select of ROM, with
			 * A13 line controlling which ROM image. CPU line A14 is not connected.
			 * 
			 * @param address
			 *            memory address
			 */
			@Override
			public byte cpuRead(final int address) {
				if (address < 0x8000) {
					final int chip = address & 0x1c00;
					if (chip < 0x800) {
						return ram[address & 0x7ff];
					}
					if (chip == 0x1800) {
						return viaBc.read(address & 0xf);
					}
					if (chip == 0x1c00) {
						return viaDc.read(address & 0xf);
					}
					/* Unconnected bus. */
					return (byte) 0xff;
				} else {
					return rom[address & 0x3fff];
				}
			}

			/**
			 * Write to address
			 * 
			 * @param address
			 *            memory address
			 */
			@Override
			public void cpuWrite(final int address, final byte data) {
				if (address < 0x8000) {
					final int chip = address & 0x1c00;
					if (chip < 0x800) {
						ram[address & 0x7ff] = data;
					}
					if (chip == 0x1800) {
						context.schedule(new Event("W"){
							@Override
							public void event() {
								viaBc.write(address & 0xf, data);
							}
						}, 1);
					}
					if (chip == 0x1c00) {
						context.schedule(new Event("W"){
							@Override
							public void event() {
								viaDc.write(address & 0xf, data);
							}
						}, 1);
					}
				}
			}

			/**
			 * The V flag is connected to the disk controller's byte ready -line.
			 * The controller sets the overflow flag during disk rotation whenever
			 * 8 bits have passed under the R/W head and the control line isn't masked.
			 */
			@Override
			public boolean getFlagV() {
				/* drive motor on? rotate disk */
				if ((byteReadyActive & 4) != 0) {
					rot.rotateDisk();
				}
				return super.getFlagV();
			}

			@Override
			public void setFlagV(final boolean state) {
				/* drive motor on? rotate disk */
				if ((byteReadyActive & 4) != 0) {
					rot.rotateDisk();
				}
				super.setFlagV(state);
			}
		};

		// Create the Bus Controller
		viaBc = new VIA6522BC(this, iecBus, context);
		// Create the Disk Controller
		viaDc = new VIA6522DC(this, rot, context);
	}

	/**
	 * Return the drive ID
	 * 
	 * @return drive ID, ranging from 0-3
	 */
	public final int getID() {
		return id;
	}

	/**
	 * Get Bus controller of this floppy.
	 * 
	 * @return bus controller of this floppy
	 */
	public VIA6522BC getBusController() {
		return viaBc;
	}

	/**
	 * Get drive controller of the floppy.
	 * 
	 * @return drive controller of this floppy
	 */
	public VIACore getDriveController() {
		return viaDc;
	}

	/**
	 * Reset normally.
	 */
	@Override
	public void reset() {
		context.reset();
		super.reset();
		cpu.triggerRST();
		rot.reset();
		viaBc.reset();
		viaDc.reset();
		
		// clear RAM
		Arrays.fill(ram, (byte) 0);
		irqCount = 0;
	}

	/**
	 * Insert Disk.<BR>
	 * A previously inserted disk will be ejected first.
	 * 
	 * @param fn
	 *            Filename
	 */
	public void insertDisk(final String fn) {
		img.attach(fn);
	}

	@Override
	protected long cpuClk() {
		return context.getTime(Phase.PHI2);
	}

	public boolean getMotor() {
		return (viaDc.oldpb & 0x4) != 0;
	}

	public boolean getLED() {
		return (viaDc.oldpb & 0x8) != 0;
	}

	public int getTrack() {
		return currentHalfTrack >> 1;
	}

	public void setIRQ(final boolean state) {
		if (state) {
			if (irqCount == 0) {
				cpu.triggerIRQ();
			}
			irqCount ++;
		} else {
			irqCount --;
			if (irqCount == 0) {
				cpu.clearIRQ();
			}
		}
	}

	protected MOS6510 getCPU() {
		return cpu;
	}

	public EventScheduler getEventScheduler() {
		return context;
	}
}
