/*
 * driveimage.c
 *
 * Written by
 *  Andreas Boose <viceteam@t-online.de>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */
package libsidplay.components.c1541;

import static libsidplay.components.c1541.Floppy.MAX_TRACKS_1541;
import static libsidplay.components.c1541.Floppy.MIN_TRACKS_1541;
import static libsidplay.components.c1541.GCR.NUM_MAX_BYTES_TRACK;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;

/**
 * Support of D64 disk images.
 * 
 * @author Ken Hndel
 *
 */
public class DiskImage {
	/**
	 * GCR byte size of a whole sector
	 */
	private static final int GCR_SECTOR_SIZE_WITH_HEADER = 354;
		
	public static final int  DISK_IMAGE_TYPE_G64 = 100;
	public static final int  DISK_IMAGE_TYPE_D64 = 1541;

	/* CBM DOS Input Processor Error Codes.  */
	private static final int  CBMDOS_IPE_OK                      =0;
	private static final int  CBMDOS_IPE_READ_ERROR_BNF          =20;
	private static final int  CBMDOS_IPE_READ_ERROR_SYNC         =21;
	private static final int  CBMDOS_IPE_READ_ERROR_DATA         =22;
	private static final int  CBMDOS_IPE_READ_ERROR_CHK          =23;
	private static final int  CBMDOS_IPE_READ_ERROR_GCR          =24;
	private static final int  CBMDOS_IPE_WRITE_ERROR_VER         =25;
	private static final int  CBMDOS_IPE_WRITE_PROTECT_ON        =26;
	private static final int  CBMDOS_IPE_READ_ERROR_BCHK         =27;
	private static final int  CBMDOS_IPE_WRITE_ERROR_BIG         =28;
	private static final int  CBMDOS_IPE_DISK_ID_MISMATCH        =29;
	private static final int  CBMDOS_IPE_NOT_READY               =74;

	private static final int D64_FILE_SIZE_35 = 174848; /* D64 image, 35 tracks */

	/**
	 * Recently attached file handle.
	 */
	RandomAccessFile fd;
	/**
	 * Recently attached file name.
	 */
	String filename;
	/**
	 * Error in each sector appended to the D64 disk image.
	 */
	private byte[] errorInfo;
	/**
	 * Attach mode read-only.
	 */
	boolean readOnly;
	/**
	 * Track count of this image.
	 */
	int tracks;

	/**
	 * Last attached image.
	 */
	private DiskImage oldImage;

	/**
	 * Floppy disk drive where this disk is attached to.
	 */
	private final Floppy drive;
	/**
	 * Device address of the floppy disk drive where this disk is attached to.
	 */
	private final int unit;
	
	public int type;

	/**
	 * Create a new image for the floppy disk drive with device ID.
	 * 
	 * @param drive
	 *            floppy disk drive
	 * @param unit
	 *            device ID
	 */
	public DiskImage(final Floppy drive, final int unit) {
		this.drive = drive;
		this.unit = unit;
	}
	
	/**
	 * Attach a D64 image file (.d64).
	 * 
	 * @param filename
	 *            file name of the image
	 * @return true - success
	 */
	protected boolean attach(final String filename) {
		DiskImage newimage = new DiskImage(this.drive, this.unit);

		if (filename == null) {
			System.err.println("No name, cannot attach floppy image.");
			return false;
		}

		newimage.filename = filename;

		if (!open(newimage)) {
			newimage.fd = null;
			newimage.filename = null;
			return false;
		}

		if (oldImage != null) {
			detach(oldImage);
		}

		if (!attachInternal(newimage)) {
		    close(newimage);
			newimage.fd = null;
			newimage.filename = null;
			oldImage = null;
			return false;
		}
		oldImage = newimage;
		return true;
	}

	private boolean open(DiskImage image) {
		try {
			if (image.readOnly) {
				image.fd = new RandomAccessFile(image.filename, "r");
			} else {
				image.fd = new RandomAccessFile(image.filename, "rw");
			}
		} catch (IOException e) {
			try {
				/*
				 * If we cannot open the image read/write, try to open it read
				 * only.
				 */
				image.fd = new RandomAccessFile(image.filename, "r");
				image.readOnly = true;
			} catch (FileNotFoundException e1) {
				System.err.println(String.format("Cannot open file `%s'.",
						image.filename));
				return false;
			}
		}

		if (!checkForD64(image)) {
			if (!checkForGCR(image)) {
				try {
					image.fd.close();
				} catch (IOException e) {
					System.err.println(String.format("Cannot close file `%s'.",
							image.filename));
					return false;
				}
				System.err.println(String.format("Unknown disk image `%s'.",
						image.filename));
				return false;
			}
		}
		return true;
	}

	private boolean checkForGCR(DiskImage image) {
	    int trackfield;
	    byte header[]=new byte[32];

	    try {
		    image.fd.seek(0);
		    image.fd.readFully(header, 0, header.length);
	    } catch (IOException e) {
	    	System.err.println("Cannot read image header.");
	    	return false;
	    }

	    if (!new String(header).startsWith("GCR-1541")) {
	    	return false;
	    }

	    if (header[8] != 0) {
	        System.err.println(String.format("Import GCR: Unknown GCR image version %d.",
	                  (int)header[8]));
	        return false;
	    }

	    if ((header[9]&0xff) < MIN_TRACKS_1541 * 2 || (header[9]&0xff) > MAX_TRACKS_1541 * 2) {
	        System.err.println(String.format("Import GCR: Invalid number of tracks (%d).",
	                  (int)header[9]));
	        return false;
	    }

	    trackfield = (header[10]&0xff) + ((header[11]&0xff) << 8);
	    if (trackfield != 7928) {
	        System.err.println(String.format("Import GCR: Invalid track field number %d.",
	                  trackfield));
	        return false;
	    }

	    image.type = DISK_IMAGE_TYPE_G64;
	    image.tracks = (header[9]&0xff) / 2;
	    image.errorInfo = null;

        return true;
	}

	/**
	 * Detect 35..42 track d64 image, determine image parameters. Walk from 35
	 * to 42, calculate expected image file size for each track, and compare
	 * this with the size of the given image.
	 * 
	 * @param image
	 *            image to check
	 * @return true - success
	 */
	private boolean checkForD64(DiskImage image) {
		int checkimageTracks;
		int checkimageBlocks, checkimageRealsize;
		try {
			checkimageRealsize = (int) image.fd.length();
		} catch (IOException e1) {
			e1.printStackTrace();
			checkimageRealsize = 0;
		}
		checkimageTracks = MIN_TRACKS_1541; /* start at track 35 */
		checkimageBlocks = D64_FILE_SIZE_35 >> 8;

		boolean checkimageErrorinfo = false;
		while (true) {
			/* check if image matches "checkimage_tracks" */
			if (checkimageRealsize == checkimageBlocks << 8) {
				/* image file matches size-with-no-error-info */
				checkimageErrorinfo = false;
				break;

			} else if (checkimageRealsize == (checkimageBlocks << 8) + checkimageBlocks) {
				/* image file matches size-with-error-info */
				checkimageErrorinfo = true;
				break;
			}

			/* try next track (all tracks from 35 to 42 have 17 blocks) */
			checkimageTracks++;
			checkimageBlocks += 17;

			/* we tried them all up to 42, none worked, image must be corrupt */
			if (checkimageTracks > MAX_TRACKS_1541)
				return false;
		}

		/*** set parameters in image structure, read error info */
	    image.type = DISK_IMAGE_TYPE_D64;
		image.tracks = checkimageTracks;

		// Initialize
		errorInfo = null;
		if (checkimageErrorinfo) {
			errorInfo = new byte[checkimageBlocks];
			try {
				image.fd.seek((256 * checkimageBlocks));
				image.fd.readFully(errorInfo, 0, checkimageBlocks);
			} catch (IOException e) {
				return false;
			}
		}

		/*** log and return successfully */
		System.out.println(String.format(
				getType(image) + " disk image recognised: %s, %d tracks%s",
				image.filename, image.tracks, image.readOnly ? " (read only)."
						: "."));
		return true;
	}

	protected void detach(DiskImage image) {
		detachInternal(image);
	    close(image);
		image.fd = null;
		image.filename = null;
	}

	/* Detach a disk image from the true drive emulation. */
	private boolean detachInternal(DiskImage image)
	{
	    if (this.drive.image != null) {
			System.out.printf("Unit %d: %s disk image detached: %s.\n", this.unit,
					getType(image), image.filename);
	    }
	    this.drive.gcrDataWriteback();
	    Arrays.fill(this.drive.gcr.data, (byte) 0);
	    this.drive.detachClk = this.drive.cpuClk();
	    this.drive.gcrImageLoaded = false;
	    this.drive.image = null;
	    return true;
	}

	/* Attach a disk image to the true drive emulation. */
	private boolean attachInternal(DiskImage image) {
		this.drive.attachClk = this.drive.cpuClk();
		if (this.drive.detachClk > 0)
			this.drive.attachDetachClk = this.drive.cpuClk();
		this.drive.askExtendDiskImage = true;

		System.out.println(String.format(
				"Unit %d: %s disk image attached: %s.", this.unit,
				getType(image), image.filename));
		this.drive.image = image;

	    if (drive.image.type == DISK_IMAGE_TYPE_G64) {
	        try {
				if (!drive.gcr.readG64Image(image)) {
					return false;
				}
			} catch (IOException e) {
		    	System.err.println("Could not read GCR disk image.");
				return false;
			}
	    } else {
	    	// drive.image.type == DISK_IMAGE_TYPE_D64
			if (setID()) {
				readD64();
			} else {
				return false;
			}
	    }
		this.drive.gcrImageLoaded = true;
		return true;
	}

	private boolean setID() {
		if (null == this.drive.image)
			return false;

		byte buffer[] = new byte[256];
		int rc = readSector(this.drive.image, buffer, 0, 18, 0);
		if (rc >= 0) {
			this.drive.diskID1 = buffer[0xa2];
			this.drive.diskID2 = buffer[0xa3];
		}

		return true;
	}

	private int readSector(DiskImage image, byte[] buf, int bufPos,
			int track, int sector) {
		if (image.fd == null) {
			System.err.println("Attempt to read without disk image.");
			return 74;
		}

		int sectors = checkSector(image, track, sector);

		if (sectors < 0) {
			System.err.println(String.format(
					"Track %d, Sector %d out of bounds.", track, sector));
			return 66;
		}

		long offset = sectors << 8;

		try {
			image.fd.seek(offset);
			image.fd.readFully(buf, bufPos, 256);
		} catch (IOException e) {
			System.err.println(String.format(
					"Error reading T:%d S:%d from disk image.", track,
					sector));
			return -1;
		}
        if (errorInfo != null) {
            switch (errorInfo[sectors]) {
              case 0x0:
              case 0x1:
                return CBMDOS_IPE_OK;               /* 0 */
              case 0x2:
                return CBMDOS_IPE_READ_ERROR_BNF;   /* 20 */
              case 0x3:
                return CBMDOS_IPE_READ_ERROR_SYNC;  /* 21 */
              case 0x4:
                return CBMDOS_IPE_READ_ERROR_DATA;  /* 22 */
              case 0x5:
                return CBMDOS_IPE_READ_ERROR_CHK;   /* 23 */ 
              case 0x7:
                return CBMDOS_IPE_WRITE_ERROR_VER;  /* 25 */
              case 0x8:
                return CBMDOS_IPE_WRITE_PROTECT_ON; /* 26 */
              case 0x9:
                return CBMDOS_IPE_READ_ERROR_BCHK;  /* 27 */
              case 0xA:
                return CBMDOS_IPE_WRITE_ERROR_BIG;  /* 28 */
              case 0xB:
                return CBMDOS_IPE_DISK_ID_MISMATCH; /* 29 */
              case 0xF:
                return CBMDOS_IPE_NOT_READY;        /* 74 */
              case 0x10:
                return CBMDOS_IPE_READ_ERROR_GCR;   /* 24 */
              default:
                return 0;
            }
        }
		return 0;
	}

	private int checkSector(DiskImage image, int track, int sector) {
		if (track < 1)
			return -1;

		int sectors = 0;
		if (image.type == DISK_IMAGE_TYPE_D64) {
			if (track > MAX_TRACKS_1541
					|| sector >= drive.imageSectorPerTrack(
							track))
				return -1;

			for (int i = 1; i < track; i++)
				sectors += drive.imageSectorPerTrack(i);
			sectors += sector;
		} else {
			// image.type == DISK_IMAGE_TYPE_G64
	        if (track > image.tracks || track > MAX_TRACKS_1541 || sector
	                >= drive.imageSectorPerTrack(track))
	                return -1;
	            for (int i = 1; i < track; i++)
	                sectors += drive.imageSectorPerTrack(i);
	            sectors += sector;
		}

		return (sectors);
	}

	private void readD64() {
		if (this.drive.image == null) {
			return;
		}
		this.drive.imageInitTrackSizeD64();
		this.drive.setHalfTrack(this.drive.currentHalfTrack);

		byte buffer[] = new byte[260];
		buffer[258] = buffer[259] = 0;

		for (int track = 1; track <= this.drive.image.tracks; track++) {
			int ptr = (track - 1) * NUM_MAX_BYTES_TRACK;
			int maxSector = drive.imageSectorPerTrack(track);
			/* Clear track to avoid read errors. */
			Arrays.fill(this.drive.gcr.data, ptr, ptr + NUM_MAX_BYTES_TRACK,
					(byte) 0x55);

			for (int sector = 0; sector < maxSector; sector++) {
				ptr = (track - 1) * NUM_MAX_BYTES_TRACK
					+ (this.drive.gapsBetweenSectors(track) + GCR_SECTOR_SIZE_WITH_HEADER) * sector;

				int rc = readSector(this.drive.image, buffer, 1, track,
						sector);
				if (rc < 0) {
					System.err.println(String.format(
							"Cannot read T:%d S:%d from disk image.", track,
							sector));
					continue;
				}

				if (rc == 21) {
					ptr = (track - 1) * NUM_MAX_BYTES_TRACK;
					Arrays.fill(this.drive.gcr.data, ptr, ptr
							+ NUM_MAX_BYTES_TRACK, (byte) 0x00);
					break;
				}

				buffer[0] = (byte) ((rc == 22) ? 0xff : 0x07);

				int chksum = buffer[1];
				for (int i = 2; i < 257; i++)
					chksum ^= buffer[i];
				buffer[257] = (byte) ((rc == 23) ? chksum ^ 0xff : chksum);
				this.drive.gcr.convertSectorToGCR(buffer, this.drive.gcr.data,
						ptr, track, sector, this.drive.diskID1, this.drive.diskID2,
						(byte) (rc));
			}
		}
	}

	private String getType(DiskImage image) {
		if (image.type==DISK_IMAGE_TYPE_G64) {
			return "G64";
		} else {
			// type==DISK_IMAGE_TYPE_D64
			return "D64";
		}
	}

	private boolean close(DiskImage image)
	{
	    if (image.fd == null) {
	        System.err.println(String.format("Cannot close file `%s'.",  image.filename));
	        return false;
	    }

	    try {
	    	image.fd.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

	    return true;
	}

	public void writeSector(byte[] buf, int bufPos, int track, int sector) {
		if (fd == null) {
			System.err.println("Attempt to write without disk image.");
			return;
		}
		if (readOnly) {
			System.err.println("Attempt to write to read-only disk image.");
			return;
		}

		int sectors = checkSector(this, track, sector);

		if (type==DISK_IMAGE_TYPE_D64) {
			if (sectors < 0) {
				System.err.println(String.format(
						"Track: %d, Sector: %d out of bounds.", track, sector));
				return;
			}
			long offset = sectors << 8;

			try {
				fd.seek(offset);
				fd.write(buf, bufPos, 256);
			} catch (IOException e) {
				System.err.println(String.format(
						"Error writing T:%d S:%d to disk image.", track, sector));
				return;
			}
		} else {
			// type==DISK_IMAGE_TYPE_G64
	        if (writeSector(buf, track, sector) < 0)
	            return;
		}
	}
	/* Write a sector to the GCR disk image.  */

	int writeSector(byte[] buf,
	                                int track, int sector)
	{
	    int gcr_track_start_ptr;
	    int gcr_current_track_size;

	    if (track > tracks) {
	        System.err.println(String.format("Track %d out of bounds.  Cannot write GCR sector",
	                  track));
	        return -1;
	    }
        gcr_track_start_ptr = ((track - 1) * NUM_MAX_BYTES_TRACK);
        gcr_current_track_size = drive.gcr.trackSize[track - 1];
	    if (!drive.gcr.writeSector(gcr_track_start_ptr,
	        gcr_current_track_size, buf, track, sector)) {
	        System.err.println(String.format("Could not find track %d sector %d in disk image",
	                  track, sector));
	        return -1;
	    }
	    try {
			drive.gcr.writeTrack(this, track, gcr_track_start_ptr,
					gcr_current_track_size);
		} catch (IOException e) {
	        System.err.println(String.format("Failed writing track %d to disk image.", track));
	        return -1;
		}

	    return 0;
	}


}