/*
 * drive.c - Hardware-level disk drive emulation.
 *
 * Written by
 *  Andreas Boose <viceteam@t-online.de>
 *
 * Based on old code by
 *  Daniel Sladic <sladic@eecg.toronto.edu>
 *  Ettore Perazzoli <ettore@comm2000.it>
 *  Andr Fachat <fachat@physik.tu-chemnitz.de>
 *  Teemu Rantanen <tvr@cs.hut.fi>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */
package libsidplay.components.c1541;

import static libsidplay.components.c1541.GCR.NUM_MAX_BYTES_TRACK;

import java.io.IOException;
import java.util.Arrays;

/**
 * Floppy disk drive.
 * 
 * @author Ken Hndel
 *
 */
public abstract class Floppy {

	/*
	 * Disk Drive Specs
	 * For customized disks, the values must fit beteen the NUM_ and MAX_
	 * limits. Do not change the NUM_ values, as they define the standard
	 * disk geometry.
	 */
	public static final int MIN_TRACKS_1541 = 35;
	public static final int EXT_TRACKS_1541 = 40;
	public static final int MAX_TRACKS_1541 = 42;

	/* Extended disk image handling.  */
	private static final int DRIVE_EXTEND_NEVER  =0;
	private static final int DRIVE_EXTEND_ASK    =1;
	private static final int DRIVE_EXTEND_ACCESS =2;

	/**
	 * Number of cycles before an attached disk becomes visible to the R/W head.
	 * This is mostly to make routines that auto-detect disk changes happy.
	 * 
	 */
	public static final int DRIVE_ATTACH_DELAY = (3 * 600000);

	/**
	 * Number of cycles the write protection is activated on detach.
	 */
	public static final int DRIVE_DETACH_DELAY = (3 * 200000);

	/**
	 * Number of cycles the after a disk can be inserted after a disk has been
	 * detached.
	 */
	public static final int DRIVE_ATTACH_DETACH_DELAY = (3 * 400000);

	/**
	 * Get the current drive clock.
	 * 
	 * @return the current drive clock
	 */
	protected abstract long cpuClk();

	/**
	 * Current half track on which the R/W head is positioned.
	 */
	int currentHalfTrack;

	/**
	 * Disk ID.
	 */
	byte diskID1, diskID2;

	/**
	 * Flag: does the current need to be written out to disk?
	 */
	boolean gcrDirtyTrack;

	/**
	 * GCR value being written to the disk.
	 */
	byte gcrWrite;

	/**
	 * Pointer to the start of the GCR data of this track.
	 */
	int gcrTrackStartPtr;

	/**
	 * Offset of the R/W head on the current track (bytes).
	 */
	int gcrHeadOffset;

	/**
	 * Are we in read or write mode?
	 */
	Rotation.Mode readWriteMode;

	/**
	 * Activates the byte ready line.
	 */
	int byteReadyActive;

	/**
	 * Tick when the disk image was attached.
	 */
	long attachClk;

	/**
	 * Tick when the disk image was detached.
	 */
	long detachClk;

	/**
	 * Tick when the disk image was attached, but an old image was just
	 * detached.
	 */
	long attachDetachClk;

    /**
     * Byte to read from r/w head.
     */
    byte gcrRead;

    /**
     * Is a GCR image loaded?
     */
    boolean gcrImageLoaded;

    /**
     * What extension policy?
     */
    int extendImagePolicy;

    /**
     * Flag: What parallel cable do we emulate?
     */
    int parallelCable;

    /**
     * If the user does not want to extend the disk image and `ask mode' is
    selected this flag gets cleared.
     */
    boolean askExtendDiskImage;

    /**
     * Pointer to the attached disk image.
     */
    DiskImage image;

    /**
     * The GCR image.
     */
    protected final GCR gcr = new GCR();

    public Floppy() {
    	currentHalfTrack = 2;
		gcrDirtyTrack = false;
		gcrWrite = 0x55;
		gcrTrackStartPtr = 0;
		gcrImageLoaded = false;
		extendImagePolicy = DRIVE_EXTEND_ACCESS;

		imageInitTrackSizeD64();

		/* Position the R/W head on the directory track. */
		setHalfTrack(36);
	}
    
    public void reset() {
		attachClk = 0;
		detachClk = 0;
		attachDetachClk = 0;
		readWriteMode = Rotation.Mode.READ;
    }
    
	void imageInitTrackSizeD64() {
		for (int track = 0; track < MAX_TRACKS_1541; track++) {
			gcr.trackSize[track] = RAW_TRACK_SIZE[SPEED_MAP_1541[track]];
			Arrays.fill(gcr.speedZone, NUM_MAX_BYTES_TRACK * track, NUM_MAX_BYTES_TRACK * (track+1) - 1, (byte) SPEED_MAP_1541[track]);
		}
	}

	/** Number of bytes per track size. */
	private static final int RAW_TRACK_SIZE[] = { 6250, 6666, 7142, 7692 };

	/** Number of gap bytes between sectors */
	private static final int GAPS_BETWEEN_SECTORS[] = { 9, 12, 17, 8 };

	private static final int SECTOR_MAP_D64[] = { 17, 18, 19, 21 };
	
	/** Track to speedzone map */
	private static final int SPEED_MAP_1541[] = { 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
			3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

	protected int getCurrentTrackSize() {
		return gcr.trackSize[currentHalfTrack / 2 - 1];
	}

	void setHalfTrack(int num) {
		/* Move the head to half track `num'.  */
	    if (num > 84)
	        num = 84;
	    if (num < 2)
	        num = 2;

	    int oldTrackSize = getCurrentTrackSize();
	    
	    currentHalfTrack = num;
	    gcrTrackStartPtr = (((currentHalfTrack / 2 - 1) * NUM_MAX_BYTES_TRACK));

	    /* 
	     * Reposition the head to the next track, accounting
	     * for potential change in speed zone (= track length)
	     */
	    if (getCurrentTrackSize() != 0)
	        gcrHeadOffset = gcrHeadOffset * oldTrackSize / getCurrentTrackSize();
	    else
	        gcrHeadOffset = 0;
	}

	int gapsBetweenSectors(int track) {
		return GAPS_BETWEEN_SECTORS[SPEED_MAP_1541[track - 1]];
	}
	
	public void moveHead(int step) {
		/* Increment the head position by `step' half-tracks. Valid values
		   for `step' are `+1' and `-1'.  */
	    gcrDataWriteback();
	    setHalfTrack(currentHalfTrack + step);
	}

	void gcrDataWriteback() {
	    if (image == null)
	        return;

	    if (!gcrDirtyTrack)
	        return;

		int track = currentHalfTrack / 2;
	    if (image.type == DiskImage.DISK_IMAGE_TYPE_G64) {
	    	int trackSize = gcr.trackSize[track - 1];
	        int trackStart = ((track - 1) * NUM_MAX_BYTES_TRACK);
			try {
				gcr.writeTrack(image, track, trackStart,
						trackSize);
			} catch (IOException e) {
		        System.err.println(String.format("Failed writing track %d to disk image.", track));
		        return;
			}
	    } else if (image.type == DiskImage.DISK_IMAGE_TYPE_D64) {
	        if (track > EXT_TRACKS_1541)
	            return;
		    int maxSector = imageSectorPerTrack(track);
	        if (track > image.tracks) {
	            switch (extendImagePolicy) {
	              case DRIVE_EXTEND_NEVER:
	                askExtendDiskImage = true;
	                return;
	              case DRIVE_EXTEND_ASK:
	                if (askExtendDiskImage) {
	                    if (!uiExtendImageDialog()) {
	                        askExtendDiskImage = false;
	                        return;
	                    } else {
	                        driveExtendDiskImage();
	                    }
	                } else {
	                    return;
	                }
	                break;
	              case DRIVE_EXTEND_ACCESS:
	                askExtendDiskImage = true;
	                driveExtendDiskImage();
	                break;
	            }
	        }

		    int offset;
		    byte buffer[]=new byte[GCR.SECTOR_SIZE];
		    for (int sector = 0; sector < maxSector; sector++) {
		        offset = gcr.findSectorHeader(track, sector,
		                                        gcrTrackStartPtr,
		                                        getCurrentTrackSize());
		        if (offset == -1) {
		            System.err.println(String.format("Could not find header of T:%d S:%d.",
		                      track, sector));
		        } else {
		            offset = gcr.findSectorData(offset,
		                                          gcrTrackStartPtr,
		                                          getCurrentTrackSize());
		            if (offset == -1) {
		            	System.err.println(String.format("Could not find data sync of T:%d S:%d.",
		                          track, sector));
		            } else {
		                gcrDataWriteback2(buffer, offset, track, sector);
		            }
		        }
		    }
	    }
	    gcrDirtyTrack = false;
	}

	private boolean uiExtendImageDialog() {
		// TODO ask the user to extend disk image
		return false;
	}

	private void gcrDataWriteback2(byte[] buffer, int offset, int track, int sector) {
		gcr.convertGCRToSector(buffer, offset, gcrTrackStartPtr,
				getCurrentTrackSize());
		if (buffer[0] != 0x7) {
			System.err.println(String
					.format("Could not find data block id of T:%d S:%d.",
							track, sector));
		} else {
			image.writeSector(buffer, 1, track, sector);
		}
	}
	private void driveExtendDiskImage()
	{
	    int track, sector;
	    byte buffer[] = new byte[256];

	    image.tracks = EXT_TRACKS_1541;
	    for (track = MIN_TRACKS_1541 + 1; track <= EXT_TRACKS_1541; track++) {
	        for (sector = 0;
	             sector < imageSectorPerTrack(track);
	             sector++) {
				image.writeSector(buffer, 0, track, sector);
	        }
	    }
	}

	int imageSectorPerTrack(int track) {
		return SECTOR_MAP_D64[SPEED_MAP_1541[track-1]];
	}

	byte writeprotectSense()
	{
	    /* Clear the write protection bit for the time the disk is pulled out on
	       detach.  */
	    if (detachClk != 0) {
	        if ((cpuClk()) - detachClk < Floppy.DRIVE_DETACH_DELAY)
	            return 0x0;
	        detachClk = 0;
	    }
	    /* Set the write protection bit for the minimum time until a new disk
	       can be inserted.  */
	    if (attachDetachClk != 0) {
	        if ((cpuClk()) - attachDetachClk
	            < Floppy.DRIVE_ATTACH_DETACH_DELAY)
	            return 0x10;
	        attachDetachClk = 0;
	    }
	    /* Clear the write protection bit for the time the disk is put in on
	       attach.  */
	    if (attachClk != 0) {
	        if ((cpuClk()) - attachClk < Floppy.DRIVE_ATTACH_DELAY)
	            return 0x0;
	        attachClk = 0;
	    }

	    if (gcrImageLoaded == false) {
	        /* No disk in drive, write protection is off. */
	        return 0x10;
	    } else {
	        return image != null && image.readOnly ? (byte) 0x0 : 0x10;
	    }
	}
}