/**
 * gcr.c - GCR handling.
 *
 * Written by
 *  Andreas Boose <viceteam@t-online.de>
 *  Daniel Sladic <sladic@eecg.toronto.edu>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 * @author Ken Hndel
 * 
 **/
package libsidplay.components.c1541;

import static libsidplay.components.c1541.Floppy.MAX_TRACKS_1541;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;

/**
 * 
 * Group Coded Recording.
 * 
 * @author Ken Hndel
 * 
 */
public class GCR {
	/**
	 * Byte size of a whole sector.
	 */
	public static final int SECTOR_SIZE = 260;

	/**
	 * GCR byte size of a whole sector's data.
	 */
	private static final int GCR_SECTOR_SIZE_DATA_ONLY = 325;

	/**
	 * Number of bytes in one raw track.
	 */
	public static final int NUM_MAX_BYTES_TRACK = 7928;

	/**
	 * Number of tracks we emulate.
	 */
	private static final int MAX_GCR_TRACKS = 70;

	/**
	 * Complete disk image as GCR data.
	 */
	byte[] data = new byte[MAX_GCR_TRACKS * NUM_MAX_BYTES_TRACK];

	/**
	 * Speed zone image of the disk.
	 */
	byte[] speedZone = new byte[MAX_GCR_TRACKS * NUM_MAX_BYTES_TRACK];

	/**
	 * Size of the GCR data of each track.
	 */
	int[] trackSize = new int[MAX_GCR_TRACKS];

	/**
	 * Convert a nybble to GCR code.
	 * 
	 * <PRE>
	 * Nybble Code
	 * 0000	  01010
	 * 0001	  01011
	 * 0010	  10010
	 * 0011	  10011
	 * 0100	  01110
	 * 0101	  01111
	 * 0110	  10110
	 * 0111	  10111
	 * 1000	  01001
	 * 1001	  11001
	 * 1010	  11010
	 * 1011	  11011
	 * 1100	  01101
	 * 1101	  11101
	 * 1110	  11110
	 * 1111	  10101
	 * </PRE>
	 */
	private static final byte[] GCR_CONV_DATA = {
		0x0a, 0x0b, 0x12, 0x13,
		0x0e, 0x0f, 0x16, 0x17,
		0x09, 0x19, 0x1a, 0x1b,
		0x0d, 0x1d, 0x1e, 0x15
	};

	private static final byte[] FROM_GCR_CONV_data = {
		 0,  0,  0,  0,  0,  0,  0,  0,
		 0,  8,  0,  1,  0, 12,  4,  5,
		 0,  0,  2,  3,  0, 15,  6,  7,
		 0,  9, 10,	11,  0, 13, 14,  0
	};

	/**
	 * Expected G64 image header.
	 */
	private static final byte G64_IMAGE_HEADER[] = { 0x47, 0x43, 0x52,
			0x2D, 0x31, 0x35, 0x34, 0x31 };
	/**
	 * G64 image: Hardcoded/expected values jsidplay2 works with
	 * <UL>
	 * <LI>0x1EF8 - 7928: Maximum container size for a raw GCR track
	 * <LI>0x54 - 84: Maximum container size for (half) track pointers
	 * <LI>0x00 - 0: GCR image file version number
	 * </UL>
	 */
	private static final int G64_CONTAINER_SIZES = 0x1EF85400;

	/**
	 * Convert the contents of a disk sector to GCR coded bytes.
	 * Input is a preformatted sector filled with 0x55 bytes (gap data).
	 * 
	 * @param sectorBytes
	 *            sector data to convert
	 * @param result
	 *            converted GCR data
	 * @param resultPos
	 *            position of result
	 * @param track
	 *            track, where sector is located in
	 * @param sector
	 *            sector to convert
	 * @param diskID1
	 *            first byte of the disk ID
	 * @param diskID2
	 *            second byte of the disk ID
	 * @param errorCode
	 *            error code
	 */
	public final void convertSectorToGCR(final byte[] sectorBytes,
			final byte[] result, final int resultPos, final int track,
			final int sector, final byte diskID1, final byte diskID2,
			final byte errorCode) {
		int off = resultPos;
		byte[] buf = new byte[4];
		byte headerId1;
		headerId1 = (byte) ((errorCode == 29) ? diskID1 ^ 0xff : diskID1);

		/* Sync */
		Arrays.fill(result, off, off + 5, (byte) 0xff);
		off += 5;

		buf[0] = (byte) ((errorCode == 20) ? 0xff : 0x08);
		buf[1] = (byte) (sector ^ track ^ diskID2 ^ headerId1);
		buf[2] = (byte) sector;
		buf[3] = (byte) track;

		if (errorCode == 27) {
			buf[1] ^= 0xff;
		}
		convert4BytesToGCR(buf, 0, result, off);
		off += 5;

		buf[0] = diskID2;
		buf[1] = headerId1;
		buf[2] = buf[3] = 0x0f;
		convert4BytesToGCR(buf, 0, result, off);
		off += 5;

		//Arrays.fill(result, off, off + 9, (byte) 0x55);
		off += 9;

		/* Sync */
		Arrays.fill(result, off, off + 5, (byte) 0xff);
		off += 5;

		int sectorBytesPos = 0;
		for (int i = 0; i < 65; i++) {
			convert4BytesToGCR(sectorBytes, sectorBytesPos, result, off);
			sectorBytesPos += 4;
			off += 5;
		}
	}

	/**
	 * Covert 4 bytes into the GCR coded form (5 bytes).
	 * 
	 * @param source
	 *            byte array to convert containing 4 bytes.
	 * @param sourcePos
	 *            position of source
	 * @param dest
	 *            resulting byte array
	 * @param destPos
	 *            position of dest
	 */
	private void convert4BytesToGCR(final byte[] source,
			final int sourcePos, final byte[] dest, final int destPos) {
		int idx = 0;

		int srcPos = sourcePos;
		int dstPos = destPos;
		for (int i = 2; i < 10; i += 2, srcPos++, dstPos++) {
			/* make room for the upper nybble */
			idx <<= 5;
			idx |= GCR_CONV_DATA[(source[srcPos] & 0xff) >> 4];

			/* make room for the lower nybble */
			idx <<= 5;
			idx |= GCR_CONV_DATA[(source[srcPos]) & 0x0f];

			dest[dstPos] = (byte) (idx >> i);
		}

		dest[dstPos] = (byte) idx;
	}

	/**
	 * Convert a GCR coded sector into bytes.
	 * 
	 * @param dest
	 *            resulting GCR byte array
	 * @param sectorDataPos
	 *            position of the GCR data containing the sector
	 * @param trackPos
	 *            offset in the GCR data where the track is located containing
	 *            the sector
	 * @param trackSize
	 *            the track size
	 */
	public void convertGCRToSector(byte[] dest, int sectorDataPos,
			int trackPos, int trackSize) {
		byte[] header = new byte[5];

		int bufferPos = 0;
		int offset = sectorDataPos;
		int trackEnd = trackPos + trackSize;
		for (int i = 0; i < (GCR_SECTOR_SIZE_DATA_ONLY/5); i++) {
			for (int j = 0; j < header.length; j++) {
				header[j] = data[offset++];
				if (offset >= trackEnd)
					offset = trackPos;
			}
			convertGCRTo4Bytes(header, dest, bufferPos);
			bufferPos += 4;
		}
	}

	/**
	 * Read track from image and write to GCR data.
	 * 
	 * @param image
	 *            image to read the track data from
	 * @param trackPos
	 *            offset in the GCR data where the track should be written to
	 * @param trackSize
	 *            the track size
	 * @return true - OK, false - error
	 * @throws IOException
	 *             error reading data from image
	 */
	public boolean writeTrack(DiskImage image, int track, int trackPos,
			int trackSize) throws IOException {
		int gcrTrack[] = new int[Floppy.MAX_TRACKS_1541 << 1];
		int gcrSpeed[] = new int[Floppy.MAX_TRACKS_1541 << 1];

		if (image.fd == null) {
			System.err.println("Attempt to write without disk image.");
			return false;
		}

		if (image.readOnly) {
			System.err.println("Attempt to write to read-only disk image.");
			return false;
		}

		image.fd.seek(12);
		readIntLittleEndian(image.fd, gcrTrack, image.tracks << 1);

		image.fd.seek(12 + image.tracks << 3);
		readIntLittleEndian(image.fd, gcrSpeed, image.tracks << 1);

		if (gcrTrack[(track - 1) << 1] == 0) {
			image.fd.seek(new File(image.filename).length());
			gcrTrack[(track - 1) << 1] = 0;
		}

		int offset = gcrTrack[(track - 1) << 1];

		byte[] len = new byte[2];
		len[0] = (byte) (trackSize % 256);
		len[1] = (byte) (trackSize / 256);

		image.fd.seek(offset);
		image.fd.write(len, 0, 2);

		// Clear gap between the end of the actual track and the start of the
		// next track.
		int gap = NUM_MAX_BYTES_TRACK - trackSize;
		if (gap > 0)
			Arrays.fill(data, trackPos + trackSize, trackPos + trackSize + gap,
					(byte) 0);

		image.fd.seek(offset + 2);
		image.fd.write(data, trackPos, NUM_MAX_BYTES_TRACK);

		if (speedZone != null) {
			boolean differentSpeeds = false;
			for (int i = 0; i < NUM_MAX_BYTES_TRACK; i++) {
				if ((speedZone[(track - 1) * NUM_MAX_BYTES_TRACK] != speedZone[(track - 1)
						* NUM_MAX_BYTES_TRACK + i])) {
					differentSpeeds = true;
					break;
				}
			}
			if (differentSpeeds) {
				/* This will change soon. */
				System.err
						.println("Saving different speed zones is not supported yet.");
				return false;
			}

			if (gcrSpeed[(track - 1) << 1] >= 4) {
				/* This will change soon. */
				System.err
						.println("Adding new speed zones is not supported yet.");
				return false;
			}

			offset = 12 + (image.tracks << 3) + ((track - 1) << 3);

			image.fd.seek(offset);
			writeIntLittleEndian(image.fd, gcrSpeed, (track - 1) << 1, 1);

			// TODO We do not support writing different speeds yet.
		}
		return true;
	}
	
	/**
	 * Write sector to GCR data.
	 * 
	 * @param trackPos
	 *            offset in the GCR data where the track starts
	 * @param trackSize
	 *            the track size
	 * @param writedata
	 *            sector data to write
	 * @param track
	 *            track where the sector should be written in
	 * @param sector
	 *            sector to write to
	 * @return true - OK, false - error (track/sector not found in GCR data)
	 */
	public boolean writeSector(final int trackPos, final int trackSize,
			final byte[] writedata, final int track, final int sector) {
		byte buffer[] = new byte[SECTOR_SIZE], gcrBuffer[] = new byte[GCR_SECTOR_SIZE_DATA_ONLY];

		// fill buffer with sector data to write
		buffer[0] = 0x7;
		System.arraycopy(writedata, 0, buffer, 1, 256);
		byte chksum = buffer[1];
		for (int i = 2; i < 257; i++)
			chksum ^= buffer[i];
		buffer[257] = chksum;
		buffer[258] = buffer[259] = 0;

		// convert the whole sector to GCR
		for (int i = 0, bufferPos = 0, gcrBufferPos = 0; i < gcrBuffer.length / 5; i++) {
			convert4BytesToGCR(buffer, bufferPos, gcrBuffer, gcrBufferPos);
			bufferPos += 4;
			gcrBufferPos += 5;
		}

		// Search for the sector data position in the GCR data
		int offset = findSectorHeader(track, sector, trackPos, trackSize);
		if (offset == -1)
			return false;
		offset = findSectorData(offset, trackPos, trackSize);
		if (offset == -1)
			return false;

		// store converted GCR sector data into the GCR data
		for (int i = 0; i < gcrBuffer.length; i++) {
			data[offset++] = gcrBuffer[i];
			if (offset == trackPos + trackSize)
				offset = trackPos;
		}
		return true;
	}

	/**
	 * Search for the sector header of the given track and sector in the GCR
	 * data.
	 * 
	 * @param track
	 *            track where the sector is contained in
	 * @param sector
	 *            sector to search for
	 * @param trackPos
	 *            offset in the GCR data where the track starts
	 * @param trackSize
	 *            the track size
	 * @return offset in the GCR data pointing to the sector or -1 (not found)
	 */
	public int findSectorHeader(final int track, final int sector,
			final int trackPos, final int trackSize) {
		int offset = trackPos;
		byte[] headerAsGCR = new byte[5], header = new byte[4];
		boolean wrapOver = false;
		int sync_count = 0;

		int trackEnd = trackPos + trackSize;
		while ((offset < trackEnd) && !wrapOver) {
			while (data[offset] != (byte) 0xff) {
				offset++;
				if (offset >= trackEnd)
					return -1;
			}

			while (data[offset] == (byte) 0xff) {
				offset++;
				if (offset == trackEnd) {
					offset = trackPos;
					wrapOver = true;
				}
				/* Check for killer tracks. */
				if ((++sync_count) >= trackSize)
					return -1;
			}

			for (int i = 0; i < 5; i++) {
				headerAsGCR[i] = data[offset++];
				if (offset >= trackEnd) {
					offset = trackPos;
					wrapOver = true;
				}
			}

			convertGCRTo4Bytes(headerAsGCR, header, 0);

			if (header[0] == 0x08) {
				/* FIXME: Add some sanity checks here. */
				if (header[2] == sector && header[3] == track)
					return offset;
			}
		}
		return -1;
	}

	/**
	 * Find sector GCR data starting at the sector header position.
	 * 
	 * @param sectorHeaderPos
	 *            search start position
	 * @param trackPos
	 *            offset in the GCR data where the track starts
	 * @param trackSize
	 *            the track size
	 * @return offset in the GCR data pointing to the sector data or -1 (not
	 *         found)
	 */
	public int findSectorData(final int sectorHeaderPos, final int trackPos,
			final int trackSize) {
		int trackEnd = trackPos + trackSize;
		int header = 0;
		int sectorDataPos = sectorHeaderPos;
		while (data[sectorDataPos] != (byte) 0xff) {
			sectorDataPos++;
			if (sectorDataPos >= trackEnd)
				sectorDataPos = trackPos;
			header++;
			if (header >= 500)
				return -1;
		}

		while (data[sectorDataPos] == (byte) 0xff) {
			sectorDataPos++;
			if (sectorDataPos == trackEnd)
				sectorDataPos = trackPos;
		}
		return sectorDataPos;
	}
	
	/**
	 * Convert 5-byte GCR code to 4-byte.
	 * 
	 * @param source
	 *            source buffer containing the 5-byte GCR code.
	 * @param dest
	 *            target buffer to hold the result
	 * @param destPos
	 *            off set into the target buffer
	 */
	private void convertGCRTo4Bytes(final byte[] source, final byte[] dest,
			final int destPos) {
		/* at least 24 bits for shifting into bits 16...20 */
		int idx = source[0] & 0xff;
		idx <<= 13;

		int sourcePos = 0;
		int off = destPos;
		for (int i = 5; i < 13; i += 2, off++) {
			sourcePos++;
			idx |= (source[sourcePos] & 0xff) << i;

			dest[off] = (byte) ((FROM_GCR_CONV_data[(idx >> 16) & 0x1f] & 0xff) << 4);
			idx <<= 5;

			dest[off] |= FROM_GCR_CONV_data[(idx >> 16) & 0x1f];
			idx <<= 5;
		}
	}

	/**
	 * Read image in G64 format and fill this GCR data and speed zone data.
	 * 
	 * @param image
	 *            disk image to read from
	 * @return 0 - OK, -1 means error
	 * @throws IOException
	 *             read error
	 */
	public boolean readG64Image(DiskImage image) throws IOException {
		/*
		 * Do G64 image file sanity checks, current VICE implementation does
		 * only support image file version 0 with a fixed track map container
		 * size of 84 and maximum track container size of 7928
		 */
		byte imageHeader[] = new byte[G64_IMAGE_HEADER.length];
		image.fd.seek(0);
		image.fd.readFully(imageHeader);
		if (!Arrays.equals(G64_IMAGE_HEADER, imageHeader)) {
			System.err.println("Unexpected GCR header found.");
			return false;
		}
		int containerSizes[] = new int[1];
		readIntLittleEndian(image.fd, containerSizes, 1);
		if (containerSizes[0] != G64_CONTAINER_SIZES) {
			System.err
					.println("Unexpected GCR image file constants found, jsidplay2 is unable to work with.");
			return false;
		}

		int trackCnt = image.tracks;

		// Read offsets to track data
		int trackOffsets[] = new int[MAX_TRACKS_1541 << 1];
		image.fd.seek(12);
		readIntLittleEndian(image.fd, trackOffsets, trackCnt << 1);
		// Read offsets to speed data
		int speedZoneOffsets[] = new int[MAX_TRACKS_1541 << 1];
		image.fd.seek(12 + trackCnt * 8);
		readIntLittleEndian(image.fd, speedZoneOffsets, trackCnt << 1);

		for (int track = 0; track < MAX_TRACKS_1541; track++) {
			int trackData = track * NUM_MAX_BYTES_TRACK;
			int zoneData = track * NUM_MAX_BYTES_TRACK;
			// Initialize GCR data and speed zone data
			Arrays.fill(data, trackData, trackData + NUM_MAX_BYTES_TRACK,
					(byte) 0xff);
			Arrays.fill(speedZone, zoneData, zoneData
					+ (NUM_MAX_BYTES_TRACK >> 2), (byte) 0x00);
			trackSize[track] = 6250;

			if (track <= trackCnt && trackOffsets[track << 1] != 0) {
				long offset = trackOffsets[track << 1];

				// Determine track size
				byte len[] = new byte[2];
				image.fd.seek(offset);
				image.fd.readFully(len, 0, 2);
				int trackLen = ((len[1] & 0xff) << 8) | (len[0] & 0xff);
				if (trackLen < 5000 || trackLen > 7928) {
					System.err.println(String
							.format("Track field length %d is not supported.",
									trackLen));
					return false;
				}
				trackSize[track] = trackLen;

				// Read GCR data
				image.fd.seek(offset + 2);
				image.fd.readFully(data, trackData, trackLen);

				// Determine speed zone size
				int speedZoneLen = (trackLen + 3) >> 2;

				if (speedZoneOffsets[track << 1] > 3) {
					byte speedData[] = new byte[NUM_MAX_BYTES_TRACK / 4];

					offset = speedZoneOffsets[track << 1];

					image.fd.seek(offset);
					image.fd.readFully(speedData, 0, speedZoneLen);
					for (int i = 0; i < speedZoneLen; i++) {
						speedZone[zoneData + (i << 2) + 3] = (byte) (speedData[i] & 3);
						speedZone[zoneData + (i << 2) + 2] = (byte) (((speedData[i] & 0xff) >> 2) & 3);
						speedZone[zoneData + (i << 2) + 1] = (byte) (((speedData[i] & 0xff) >> 4) & 3);
						speedZone[zoneData + (i << 2)] = (byte) (((speedData[i] & 0xff) >> 6) & 3);
					}
				} else {
					Arrays.fill(speedZone, zoneData, zoneData
							+ NUM_MAX_BYTES_TRACK,
							(byte) speedZoneOffsets[track << 1]);
				}
			}
		}
		return true;
	}

	/**
	 * Reverses endianness of 32-bit integer.
	 * 
	 * @param value
	 * @return value with endianness flipped
	 */
	private static int flip(int value) {
		return (value >> 24) & 0xff
			| (value >> 8) & 0xff00
			| (value << 8) & 0xff0000
			| (value << 24);
	}
	
	/**
	 * Reads num little-endian integers from a file (at the current position).
	 * 
	 * @param fd
	 *            file handle
	 * @param buf
	 *            target buffer containing the integer values of the file
	 * @param num
	 *            number of integers to read
	 * @throws IOException
	 *             read error occurred
	 */
	private void readIntLittleEndian(final RandomAccessFile fd,
			final int[] buf, final int num) throws IOException {
		for (int i = 0; i < num; i ++) {
			buf[i] = flip(fd.readInt());
		}
	}

	/**
	 * Writes num little-endian integers into a file (at the current position).
	 * 
	 * @param fd
	 *            file handle
	 * @param buf
	 *            source buffer containing the integer values to write into the
	 *            file
	 * @param num
	 *            number of integers to write
	 * @throws IOException
	 *             write error occurred
	 */
	private void writeIntLittleEndian(RandomAccessFile fd, int[] buf,
			int bufPos, int num) throws IOException {
		for (int i = 0; i < num; i ++) {
			fd.writeInt(flip(buf[bufPos + i]));
		}
	}
}