/*
 * iecbus.c - IEC bus handling.
 *
 * Written by
 *  Andreas Boose <viceteam@t-online.de>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */
package libsidplay.components.c1541;

import java.util.Arrays;

/**
 * IEC Bus Implementation.
 * 
 * @author Ken Hndel
 * 
 */
public class IECBus {
	/**
	 * Maximum number of floppy disk drives.
	 */
	public static final int DRIVE_NUM =4;

	/**
	 * Number of IEC devices.
	 */
	public static final int IECBUS_NUM = 16;

	protected byte drvBus[] = new byte[IECBUS_NUM];
	protected byte drvData[] = new byte[IECBUS_NUM];
	protected byte drvPort;
	protected byte cpuBus;
	protected byte cpuPort;

	/**
	 * Floppy Disk Drives connected to the bus.
	 */
	protected C1541[] drives;
	
	public IECBus() {
		Arrays.fill(drvBus, (byte) 0xff);
		Arrays.fill(drvData, (byte) 0xff);
		cpuBus = (byte) 0xff;
		cpuPort = (byte) 0xff;
		drvPort = (byte) 0x85;
	}
	
	/**
	 * Set Floppy Disk Drives.
	 * 
	 * @param drvs
	 *            Floppies
	 */
	public void setFloppies(final C1541[] drvs) {
		this.drives = drvs;
	}
	
	private void updateCpuBus(final byte data) {
		cpuBus = (byte) ((data & 0xff) << 2 & 0x80 | (data & 0xff) << 2 & 0x40 | (data & 0xff) << 1 & 0x10);
	}

	protected void updatePorts() {
		cpuPort = cpuBus;
		for (int unit = 4; unit < 8 + DRIVE_NUM; unit++) {
			cpuPort &= drvBus[unit];
		}

		drvPort = (byte) ((cpuPort & 0xff) >> 4 & 0x4 | (cpuPort & 0xff) >> 7 | (cpuBus & 0xff) << 3 & 0x80);
	}

	// triggered by CIA
	/**
	 * VICE method: iecbus_cpu_read_conf.
	 * 
	 * @return bus value
	 */
	public final byte readFromIECBus() {
		return cpuPort;
	}

	// triggered by CIA
	/**
	 * VICE method: iecbus_cpu_write_conf.
	 * 
	 * @param data
	 *            write value
	 */
	public final void writeToIECBus(final byte data) {
		byte oldCpuBus = cpuBus;
		updateCpuBus(data);

		if (((oldCpuBus ^ cpuBus) & 0x10) != 0) {
			for (int dnr = 0; dnr < drives.length; dnr++) {
				drives[dnr].getBusController().signal(VIACore.VIA_SIG_CA1,
						(cpuBus & 0x10) != 0 ? 0 : VIACore.VIA_SIG_RISE);
			}
		}
		for (int dnr = 0; dnr < drives.length; dnr++) {
			int unit = drives[dnr].getID() + 8;
			drvBus[unit] = (byte) ((((drvData[unit] & 0xff) << 3) & 0x40) | (((drvData[unit] & 0xff) << 6)
					& (((~drvData[unit] ^ cpuBus) & 0xff) << 3) & 0x80));
		}

		updatePorts();
	}

}
