/*
 * via2d.c - VIA2 emulation in the 1541, 1541II, 1571 and 2031 disk drive.
 *
 * Written by
 *  Andreas Boose <viceteam@t-online.de>
 *  Andre' Fachat <fachat@physik.tu-chemnitz.de>
 *  Daniel Sladic <sladic@eecg.toronto.edu>
 *  Ettore Perazzoli <ettore@comm2000.it>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */
package libsidplay.components.c1541;

import libsidplay.common.EventScheduler;
import libsidplay.components.c1541.Rotation.Mode;

/**
 * Handling of the Disk Controller.
 * 
 * @author Ken Hndel
 * 
 */
public final class VIA6522DC extends VIACore {
	/**
	 * Disk rotation.
	 */
	private final Rotation rot;

	/**
	 * Creates a new instance of VIA6522DC.
	 * 
	 * @param c1541
	 *            the C1541 we are attached to
	 * @param rotation
	 *            disk rotation
	 * @param ctx
	 *            event context
	 */
	public VIA6522DC(final C1541 c1541, final Rotation rotation,
			final EventScheduler ctx) {
		super(ctx, c1541, "1541Drive" + c1541.getID() + "Via2");
		this.rot = rotation;
	}

	@Override
	protected void setCa2(final int state) {
		/* drive motor on? rotate disk */
		if ((drive.byteReadyActive & 4) != 0) {
			rot.rotateDisk();
		}
		drive.byteReadyActive = this.drive.byteReadyActive & ~0x02 | state << 1;
	}

	@Override
	protected void setCb2(final int state) {
		/* drive motor on? rotate disk */
		if ((drive.byteReadyActive & 4) != 0) {
			rot.rotateDisk();
		}
		drive.readWriteMode = state != 0 ? Mode.READ : Mode.WRITE;
	}

	@Override
	protected final void storePra(final byte byt) {
		/* drive motor on? rotate disk */
		if ((drive.byteReadyActive & 4) != 0) {
			rot.rotateDisk();
		}
		this.drive.gcrWrite = byt;
	}

	@Override
	protected final void storePrb(final byte byt) {
		/* drive motor on? rotate disk */
		if ((drive.byteReadyActive & 4) != 0) {
			rot.rotateDisk();
		}

		if (((oldpb ^ byt) & 0x3) != 0 && (byt & 0x4) != 0) {
			/* Stepper motor */
			if ((oldpb & 0x3) == (byt + 1 & 0x3)) {
				this.drive.moveHead(-1);
			} else if ((oldpb & 0x3) == (byt - 1 & 0x3)) {
				this.drive.moveHead(+1);
			}
		}

		if (((oldpb ^ byt) & 0x60) != 0) {
			rot.setSpeedZone((byt & 0xff) >> 5 & 0x3);
		}
		if (((oldpb ^ byt) & 0x04) != 0) {
			drive.byteReadyActive = drive.byteReadyActive & ~0x04 | byt & 0x04;
			/* drive motor is turned on */
			if ((byt & 0x04) != 0) {
				rot.rotationBegins();
			}
		}
	}

	@Override
	protected void storeAcr(final byte value) {
	}

	@Override
	protected void storeSr(final byte value) {
	}

	@Override
	protected void storeT2l(final byte value) {
	}

	@Override
	protected final byte readPra() {
		rot.byteRead();

		/* the bus is shared between the read/write modes. The schematic says that
		 * during write mode, the read circuitry is disconnected and therefore it's
		 * likely that you will read 0xff if you change the DDR.
		 */
		byte encoderDecoderValue = drive.readWriteMode == Rotation.Mode.READ ?
				drive.gcrRead : (byte) 0xff;
		
		return (byte) (encoderDecoderValue & ~via[VIACore.VIA_DDRA] | via[VIACore.VIA_PRA] & via[VIACore.VIA_DDRA]);
	}

	@Override
	protected final byte readPrb() {
		/* drive motor on? rotate disk */
		if ((drive.byteReadyActive & 4) != 0) {
			rot.rotateDisk();
		}
		return (byte) ((rot.syncFound() | drive.writeprotectSense()) & ~via[VIACore.VIA_DDRB]
		      | via[VIACore.VIA_PRB] & via[VIACore.VIA_DDRB]);
	}
}
