/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation. For the full
 * license text, see http://www.gnu.org/licenses/gpl.html.
 */
package libsidplay.components.keyboard;




/**
 * Implements the C64's keyboard.<br>
 * <br>
 * For documentation on the C64 keyboard handling, see <a
 * href='http://www.zimmers.net/anonftp/pub/cbm/c64/programming/documents/keymatrix.txt'>http://www.zimmers.net/anonftp/pub/cbm/c64/programming/documents/keymatrix.txt</a>
 * or <a
 * href='http://www.zimmers.net/anonftp/pub/cbm/magazines/transactor/v5i5/p039.jpg'>http://www.zimmers.net/anonftp/pub/cbm/magazines/transactor/v5i5/p039.jpg</a>.
 * 
 * @author Joerg Jahnke (joergjahnke@users.sourceforge.net)
 */
public abstract class Keyboard {

	/**
	 * Keyboard matrix
	 */
	private final byte[][] rowMatrix = new byte[8][8];
	/**
	 * Reversed keyboard matrix
	 */
	private final byte[][] columnMatrix = new byte[8][8];

	/** is CBM pressed? */
	private boolean cbm;

	/** Is left shift pressed ?*/
	private boolean leftShift;

	/** Is right shift pressed? */
	private boolean rightShift;

	/**
	 * Is any key forcing shift state?
	 * > 0 when we must pretend shift is down.
	 * < 0 when we must pretend shift is not down.
	 */
	private int forcedShift = 0;

	/**
	 * Creates a new keyboard attached to a given CIA
	 */
	/**
	 * Reset the keyboard
	 */
	public void reset() {
		for (int i = 0; i < 8; ++i) {
			for (int j = 0; j < 8; ++j) {
				rowMatrix[i][j] = 0;
				columnMatrix[i][j] = 0;
			}
		}
		cbm = leftShift = rightShift = false;
		forcedShift = 0;
	}

	/**
	 * Handle a pressed key
	 * 
	 * @param key
	 *            key that was pressed
	 */
	public void keyPressed(final KeyTableEntry ktEntry) {
		final int row = ktEntry.getRow();
		final int col = ktEntry.getCol();
		rowMatrix[row][col] ++;
		columnMatrix[col][row] ++;

		if (ktEntry.hasShiftPreference()) {
			forcedShift += ktEntry.shiftDown() ? 1 : -1;
		}
	}

	/**
	 * Handle a released key
	 * 
	 * @param key
	 *            key to release
	 */
	public void keyReleased(final KeyTableEntry ktEntry) {
		if (ktEntry.hasShiftPreference()) {
			forcedShift -= ktEntry.shiftDown() ? 1 : -1;
		}

		final int row = ktEntry.getRow();
		final int col = ktEntry.getCol();
		rowMatrix[row][col] --;
		columnMatrix[col][row] --;
	}

	/**
	 * Get read adjustment for CIA 1 register PRA or PRB
	 * 
	 * @param testRegisterValue
	 *            register value from PRA or PRB
	 * @param matrix$
	 *            either activeColumnsMatrix or activeRowsMatrix
	 * @return read adjustment, to be AND connected to the normal register
	 *         output
	 */
	private byte getCIARegisterAdjustment(final int testRegisterValue, final byte[][] matrix) {
		byte result = (byte) 0xff;

		/* temporarily adjust all matrices with controls for reading */
		if (cbm) {
			keyPressed(KeyTableEntry.COMMODORE);
		}

		if (forcedShift >= (rightShift ? 0 : 1)) {
			keyPressed(KeyTableEntry.SHIFT_RIGHT);
		}

		if (forcedShift >= (leftShift ? 0 : 1)) {
			keyPressed(KeyTableEntry.SHIFT_LEFT);
		}

		// AND all active (bit is cleared) columns/rows of the matrix
		for (int i = 0; i < 8; i ++) {
			if ((testRegisterValue & 1 << i) == 0) {
				for (int j = 0; j < 8; j ++) {
					if (matrix[i][j] != 0) {
						result &= ~(1 << j);
					}
				}
			}
		}

		if (cbm) {
			keyReleased(KeyTableEntry.COMMODORE);
		}

		if (forcedShift >= (rightShift ? 0 : 1)) {
			keyReleased(KeyTableEntry.SHIFT_RIGHT);
		}

		if (forcedShift >= (leftShift ? 0 : 1)) {
			keyReleased(KeyTableEntry.SHIFT_LEFT);
		}

		return result;
	}

	/**
	 * Get read adjustment for CIA 1 register PRA
	 * 
	 * @param testRegisterValue
	 *            register value from PRA or PRB
	 * @return read adjustment, to be AND connected to the normal register
	 *         output
	 */
	public byte readColumnMatrix(final int testRegisterValue) {
		return getCIARegisterAdjustment(testRegisterValue, this.columnMatrix);
	}

	/**
	 * Get read adjustment for CIA 1 register PRB
	 * 
	 * @param testRegisterValue
	 *            register value from PRA or PRB
	 * @return read adjustment, to be AND connected to the normal register
	 *         output
	 */
	public byte readRowMatrix(final int testRegisterValue) {
		return getCIARegisterAdjustment(testRegisterValue, this.rowMatrix);
	}

	public void cbm(final boolean b) {
		cbm = b;
	}

	public void leftShift(final boolean b) {
		leftShift = b;
	}

	public void rightShift(final boolean b) {
		rightShift = b;
	}
	
	/**
	 * Restore key pressed by user
	 */
	public abstract void restore();
}
