/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation. For the full
 * license text, see http://www.gnu.org/licenses/gpl.html.
 */
package libsidplay.components.mos656x;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Arrays;

import libsidplay.common.Event;
import libsidplay.common.Event.Phase;
import libsidplay.common.EventScheduler;
import libsidplay.components.pla.Bank;

/* TODO
 * 
 * - vborder2.prg proves that whether graphics sequencer is on or off for a line is decided not from the
 *   vborder flag but something else. Perhaps there is a graphics sequencer master toggle, which is set
 *   at the start of line. At any rate, changing the vborder at middle of line must not stop graphics
 *   sequencer. (The on/off is distinct from isDisplayActive, which is graphics sequencer enabled, merely
 *   without badlines and thus reading idle data.)
 * 
 * - videomode splits indicate that videomode changes need to take effect without xscroll delay. Unfortunately
 *   fixing that will destroy the 8-pixels-at-once tables, and thus increase the costs for sequencer.
 * 
 * - sprite damaged fetch region. Placement given for sprite 0, but other sprites
 *   all follow $10 increments. Properties:
 *   - sprite pixel at $162 repeats to $169 then is cut.
 *   - no sprite is displayed if sprite is positioned between $16a .. $16e
 *   - sprite works after $16f again.
 * 
 * - sprite pixel-exact display enable. Sprites show the idle data after $164
 *   position for sprite 1+ because DMA fetch hasn't happened yet. For sprite 0,
 *   the display goes off at $163, and DMA will have happened by $16f and thus
 *   it will never show any idle data.
 */

/**
 * Implements the functionality of the C64's VIC II chip
 * 
 * For a good German documentation on the MOS 6567/6569 Videocontroller (VIC-II), see <a
 * href='http://www.minet.uni-jena.de/~andreasg/c64/vic_artikel/vic_artikel_1.htm'>http://www.minet.uni-jena.de/~andreasg/c64/vic_artikel/vic_artikel_1.htm</a> , <a
 * href='http://cbmmuseum.kuto.de/zusatz_6569_vic2.html'>http://cbmmuseum.kuto.de/zusatz_6569_vic2.html</a> or <a
 * href='http://unusedino.de/ec64/technical/misc/vic656x/vic656x-german.html'>http://unusedino.de/ec64/technical/misc/vic656x/vic656x-german.html</a>. An English version of the documentation you can
 * find at <a href='http://www.unusedino.de/ec64/technical/misc/vic656x/vic656x.html'>http://www.unusedino.de/ec64/technical/misc/vic656x/vic656x.html</a> .<br>
 * 
 * @author Joerg Jahnke (joergjahnke@users.sourceforge.net)
 * @author Antti S. Lankila (alankila@bel.fi)
 */
public abstract class MOS656X extends Bank {
	/**
	 * Chip models supported by MOS656X.
	 * 
	 * @author alankila
	 */
	public static enum Model {
		MOS6567R56A, /* OLD NTSC CHIP */
		MOS6567R8, /* NTSC */
		MOS6569R1, /* PAL */
		MOS6569R3, /* PAL */
		MOS6569R4, /* PAL */
		MOS6569R5, /* PAL */
	}

	/**
	 * Table to decode graphics byte + graphics mode directly to a color allocation.
	 * It converts a graphics byte to a 4x8 packed value using the following scheme:
	 * 
	 * <ul>
	 * <li>bits 0 .. 2 are the color value
	 * <li>bit 3 indicates the current pixel as foreground pixel
	 * </ul>
	 * With the following colors:
	 * <ul>
	 * <li>0 = bg0
	 * <li>1 = bg1
	 * <li>2 = bg2
	 * <li>3 = bg3
	 * <li>4 = color
	 * <li>5 = vmdlo
	 * <li>6 = vmdhi
	 * <li>7 = black (invalid mode)
	 * </ul>
	 */
	private static final int[][] pixelToColor = new int[10][256];
	/** Current dynamic colors */
	private final byte[] videoModeColors = new byte[8];
	/** Current sprite colors */
	private final byte[] spriteColors = new byte[11];
	/** Current border color */
	private byte borderColor;
	/** Current graphics byte to packed color array decoding table. */
	private int[] currentGraphicsMode;

	/**
	 * Determine the graphics mode after a change to VIC graphics mode registers
	 * 
	 * <ul>
	 * <li>videomode 0 = single text: bg0, color*
	 * <li>videomode 1 = multiclor text: bg0, bg1, bg2*, color*
	 * <li>videomode 2 = singlecolor bitmap: vmdlo, vmdhi*
	 * <li>videomode 3 = multicolor bitmap: bg0, vmdhi, *vmdlo, *color
	 * <li>videomode 4 = invalid singlecolor
	 * <li>videomode 5 = invalid multicolor
	 * <li>videomode 6..9 = ECM (according to video matrix top 2 bits) bg0/bg1/bg2/bg3, *color
	 * </ul>
	 */
	private void determineGraphicsMode() {
		isECM = (registers[0x11] & 0x40) != 0;
		isBMM = (registers[0x11] & 0x20) != 0;
		isMCM = (registers[0x16] & 0x10) != 0;
		if (isECM) {
			if (isBMM || isMCM) {
				/* ECM and some other bit? invalid mode */
				currentGraphicsMode = pixelToColor[isMCM ? 5 : 4];
			} else {
				currentGraphicsMode = pixelToColor[6 + (latchedVmd2 >> 6 & 3)];
			}
		} else if (isBMM) {
			/* bitmap mode, multicolor or singlecolor */
			currentGraphicsMode = pixelToColor[isMCM ? 3 : 2];
		} else {
			/* text or ECM text; multicolor is handled by drawGraphics() */
			currentGraphicsMode = pixelToColor[0];
		}
	}

	static {
		for (int i = 0; i < 256; i++) {

			for (int b = 0; b < 8; b++) {
				final int singlePixelColor = i >> b & 1; // 0 .. 1
			final int multiPixelColor = i >> (b & 6) & 3; // 0 .. 3

			pixelToColor[0][i] <<= 4;
			pixelToColor[1][i] <<= 4;
			pixelToColor[2][i] <<= 4;
			pixelToColor[3][i] <<= 4;
			pixelToColor[4][i] <<= 4;
			pixelToColor[5][i] <<= 4;
			pixelToColor[6][i] <<= 4;
			pixelToColor[7][i] <<= 4;
			pixelToColor[8][i] <<= 4;
			pixelToColor[9][i] <<= 4;

			pixelToColor[0][i] |= 12 * singlePixelColor;
			pixelToColor[1][i] |= multiPixelColor == 0 ? 0 : multiPixelColor == 1 ? 1 : multiPixelColor == 2 ? 10 : 12;
			pixelToColor[2][i] |= 5 + 9 * singlePixelColor;
			pixelToColor[3][i] |= multiPixelColor == 0 ? 0 : multiPixelColor == 1 ? 6 : multiPixelColor == 2 ? 13 : 12;
			pixelToColor[4][i] |= 7 + 8 * singlePixelColor;
			pixelToColor[5][i] |= 7 | (multiPixelColor > 1 ? 8 : 0);
			pixelToColor[6][i] |= singlePixelColor == 1 ? 12 : 0;
			pixelToColor[7][i] |= singlePixelColor == 1 ? 12 : 1;
			pixelToColor[8][i] |= singlePixelColor == 1 ? 12 : 2;
			pixelToColor[9][i] |= singlePixelColor == 1 ? 12 : 3;
			}
		}
	}

	/** contains the graphics data for the next character */
	private byte graphicsData;
	/** horizontal scrolling value */
	private byte xscroll = 0;
	private int spriteDataBuffer;

	/**
	 * Generate graphics data.
	 * 
	 * @param length
	 *            pixels to generate
	 */
	private void mergeSpritesAndGraphics(final int length) {
		if (showBorderMain) {
			for (int i = 0; i < length; i ++) {
				palEmulatedColorOutput(borderColor);
				graphicsDataBuffer >>>= 4;
			}
			spriteDataBuffer >>>= 4 * length;
		} else {
			for (int i = 0; i < length; i ++) {
				final int pc = spriteDataBuffer & 0xf;
				final byte color;
				if (pc == 0) {
					color = videoModeColors[graphicsDataBuffer & 0x7];
				} else {
					color = spriteColors[pc];
				}
				palEmulatedColorOutput(color);
				graphicsDataBuffer >>>= 4;
				spriteDataBuffer >>>= 4;
			}
		}
	}

	/** Prevailing VIC color palette for current line (odd/even) */
	private byte[] linePaletteCurrent;
	/** Table for looking up color using a packed 2x8 value. */
	private final int[] combinedLines = new int[128 * 256];
	/** VIC color palette for even rasterlines */
	private final byte[] linePaletteEven = new byte[16 * 16 * 16 * 16];
	/** VIC color palette for odd rasterlines */
	private final byte[] linePaletteOdd = new byte[16 * 16 * 16 * 16];
	/** Current 4-pixel tuple */
	private int colorFilterBuffer;
	/** Last line's color */
	private final byte[] previousLineDecodedColor = new byte[65 * 8];
	/** Index into last line */
	private int previousLineIndex = 0;

	/**
	 * Output one pixel to screen.
	 * 
	 * @param colorIndex
	 *            color of pixel to output.
	 */
	private void palEmulatedColorOutput(final byte colorIndex) {
		colorFilterBuffer <<= 4;
		colorFilterBuffer |= colorIndex;

		final byte lineColor = linePaletteCurrent[colorFilterBuffer & 0xffff];
		final byte previousLineColor = previousLineDecodedColor[previousLineIndex];
		pixels[nextPixel++] = combinedLines[(~lineColor ^ previousLineColor << 8)];
		previousLineDecodedColor[previousLineIndex ++] = lineColor;
	}

	/**
	 * Generate sprite colors for next 8 pixels.
	 * 
	 * @param sequencerData
	 *            The graphics sequencer data (used to clip sprites)
	 * @return 4x8 pixel of sprite sequencer data (0 .. 10, to be used against sprite palette)
	 */
	private int drawSprites(final int sequencerData) {
		/** Sequencer foreground pixels */
		final int sequencerForeground = sequencerData >> 3 & 0x11111111;
		/** All sprite foregrounds */
		int opaqueSpritePixels = 0;

		/** Final output */
		int mergedSprites = 0;

		Sprite prev = spriteLinkedListHead;
		Sprite current = prev.nextVisibleSprite;

		while (current != null) {
			final int spriteData = current.getNext8Pixels();

			/* find which sprite colors are foreground colors.
			 * We do this by ORing 4 neighboring bits together
			 * and selecting the 3rd bit as the signal bit. */
			int spriteForegroundMask = spriteData;
			spriteForegroundMask |= spriteForegroundMask >> 1;
			spriteForegroundMask |= spriteForegroundMask >> 2;
			spriteForegroundMask &= 0x11111111;

			/* Handle sprite-bg collision */
			if ((spriteForegroundMask & sequencerForeground) != 0) {
				if (registers[0x1f] == 0) {
					activateIRQFlag(IRQ_SPRITE_BACKGROUND_COLLISION);
				}
				registers[0x1f] |= 1 << current.index;
			}

			/* Handle sprite-sprite collision. It's easy to detect that
			 * a collision occurred, but hard to find which two sprites collided.
			 * For this purpose, we store the sprite index which contributed
			 * a pixel into the high 3 bits of a pixel slot, and keep the 0th bit
			 * as a mask that we can use to reliably identify the collision.
			 */
			if ((opaqueSpritePixels & spriteForegroundMask) != 0) {
				if (registers[0x1e] == 0) {
					activateIRQFlag(IRQ_SPRITE_SPRITE_COLLISION);
				}
				registers[0x1e] |= 1 << current.index;

				for (int pixel = 0; pixel < 32; pixel += 4) {
					/* non-transparent from us? */
					if ((spriteForegroundMask >> pixel & 1) == 0) {
						continue;
					}
					/* non-transparent from other? */
					int otherSprite = opaqueSpritePixels >> pixel & 0xf;
					if (otherSprite == 0) {
						/* no pixel set at that slot? Set ourselves */
						opaqueSpritePixels |= (1 | current.index << 1) << pixel;
					} else {
						/* Collision; register the other sprite as colliding,
						 * but keep the old value, as it doesn't matter
						 * whether it contains the other or ourselves: both
						 * collision bits are set on 0x1e once we finish. */
						registers[0x1e] |= 1 << (otherSprite >> 1);
					}
				}					
			} else {
				opaqueSpritePixels |= (1 | current.index << 1) * spriteForegroundMask;
			}	

			final int priorityMask = current.getNextPriorityMask();
			spriteForegroundMask &= ~sequencerForeground | priorityMask;
			mergedSprites ^= (spriteData ^ mergedSprites) & 0xf * spriteForegroundMask;

			/* exhausted this sprite's data? */
			if (current.consuming) {
				prev = current;
			} else {
				prev.nextVisibleSprite = current.nextVisibleSprite;
			}

			current = current.nextVisibleSprite;
		}

		return mergedSprites;
	}

	/** Property change support */
	private final PropertyChangeSupport support;

	/** First line when we check for bad lines */
	private static final int FIRST_DMA_LINE = 0x30;
	/** Last line when we check for bad lines */
	private static final int LAST_DMA_LINE = 0xf7;
	/** Display begin line */
	private int FIRST_DISPLAY_LINE;
	/** Display end line */
	private int LAST_DISPLAY_LINE;
	/** number of VIC cycles per line */
	private int CYCLES_PER_LINE;
	/** maximum number of lines to process */
	private int MAX_RASTERS;
	/** Width of screen in pixels including border */
	private int borderWidth;

	/** Credits */
	private static final String credit = "MOS656X (cycle-exact VICII) Emulation:\n"
			+ "\tCopyright (C) 2009 Joerg Jahnke <joergjahnke@users.sourceforge.net>\n"
			+ "\tCopyright (C) 2010 Antti S. Lankila <alankila@bel.fi>\n";

	// graphic modes
	/**
	 * raster IRQ flag
	 */
	private final static byte IRQ_RASTER = 1;
	/**
	 * sprite-background collision IRQ flag
	 */
	private final static byte IRQ_SPRITE_BACKGROUND_COLLISION = 2;
	/**
	 * sprite-sprite collision IRQ flag
	 */
	private final static byte IRQ_SPRITE_SPRITE_COLLISION = 4;
	/**
	 * Light-Pen IRQ flag
	 */
	private final static byte MOS656X_INTERRUPT_LP = 1 << 3;

	/** memory for chip registers */
	private final byte[] registers = new byte[0x40];
	/** video counter, a 10 bit counter */
	private int vc;
	/** video counter base, a 10 bit data register */
	private int vcBase;
	/** row counter, a 3 bit counter */
	private int rc;
	/** is the display active or idle? */
	protected boolean isDisplayActive;
	/** are bad lines enabled for this frame? */
	private boolean areBadLinesEnabled;
	/** current raster line */
	private int rasterY;
	/** the 8 sprites */
	private final Sprite[] sprites = new Sprite[8];
	/** The head of sprite linked list. The top is never rendered; it's just an anchor. */
	private final Sprite spriteLinkedListHead = new Sprite(null, -1);
	/** contains the color data of the current character */
	private final byte colorData[] = new byte[40];
	/** contains the video matrix data for the current character */
	private final byte videoMatrixData[] = new byte[40];
	/** vertical border flip-flop */
	private boolean showBorderVertical;
	/** main border flip-flop */
	private boolean showBorderMain;
	/** is the current line a bad line */
	protected boolean isBadLine;
	/** is ECM mode active? */
	private boolean isECM;
	/** is bitmap mode active */
	private boolean isBMM;
	/** is multi-color mode active? */
	private boolean isMCM;
	/** video matrix base address */
	private int videoMatrixBase;
	/** character memory base address */
	private int charMemBase;
	/** bitmap memory base address */
	private int bitmapMemBase;
	/** vertical scrolling value */
	private int yscroll;
	/** internal IRQ flags */
	private byte irqFlags;
	/** masks for the IRQ flags */
	private byte irqMask;
	/** Output RGBA screen buffer as int32 array */
	private final int[] pixels = new int[48 * 312 * 8];
	/** Index of next pixel to paint */
	private int nextPixel;
	/** Current visible line */
	private int lineCycle;
	/** Is display rendering enabled? */
	private boolean graphicsRendering;
	/** CPU's event context. */
	private final EventScheduler context;
	/** Light pen coordinates */
	private byte lpx, lpy;
	/** Has light pen IRQ been triggered in this frame already? */
	private boolean lpTriggered;
	/** System's palette */
	private final Palette palette = new Palette();
	/** Chip model of this instance. */
	private Model model;
	/** Set when new frame starts. */
	private boolean startOfFrame;

	/**
	 * Creates a new instance of VIC6569
	 * 
	 * @param c64
	 *            the C64 the VIC works for
	 */
	public MOS656X(final EventScheduler context) {
		for (int i = 0; i < sprites.length; i++) {
			sprites[i] = new Sprite(spriteLinkedListHead, i);
		}

		support = new PropertyChangeSupport(this);
		this.context = context;
		chip(Model.MOS6569R1);
	}

	public final void chip(final Model newModel) {
		model = newModel;

		switch (model) {
		// Seems to be an older NTSC chip
		case MOS6567R56A:
			MAX_RASTERS = 262;
			CYCLES_PER_LINE = 64;
			FIRST_DISPLAY_LINE = 40;
			LAST_DISPLAY_LINE = 13;
			break;

			// NTSC Chip
		case MOS6567R8:
			MAX_RASTERS = 263;
			CYCLES_PER_LINE = 65;
			FIRST_DISPLAY_LINE = 40;
			LAST_DISPLAY_LINE = 13;
			break;

			// PAL Chip
		case MOS6569R1:
		case MOS6569R3:
		case MOS6569R4:
		case MOS6569R5:
			MAX_RASTERS = 312;
			CYCLES_PER_LINE = 63;
			FIRST_DISPLAY_LINE = 15;
			LAST_DISPLAY_LINE = 300;
			break;
		}

		// XXX we should actually keep cycle start/stop values.
		// 6568 starts 4 chars before and ends 5,5 chars after,
		borderWidth = (40 + 4 + 4) * 8;
		updatePalette();
	}

	/**
	 * Get the width including the border
	 * 
	 * @return width in pixels
	 */
	public final int getBorderWidth() {
		return borderWidth;
	}

	/**
	 * Get the height including the border
	 * 
	 * @return height in pixels
	 */
	public final int getBorderHeight() {
		if (LAST_DISPLAY_LINE > FIRST_DISPLAY_LINE) {
			return LAST_DISPLAY_LINE - FIRST_DISPLAY_LINE + 1;
		} else {
			return LAST_DISPLAY_LINE - FIRST_DISPLAY_LINE + MAX_RASTERS + 1;
		}
	}

	/**
	 * Get screen as RGB data
	 * 
	 * @return C64 screen pixels as RGB data
	 */
	public final int[] getPixels() {
		return pixels;
	}

	/**
	 * Read the x-coordinate of a sprite
	 * 
	 * @param spriteNo
	 *            no of the sprite (0-7)
	 * @return x-coordinate
	 */
	private int readSpriteXCoordinate(final int spriteNo) {
		return (registers[0x00 + (spriteNo << 1)] & 0xff) + ((registers[0x10] & 1 << spriteNo) != 0 ? 256 : 0);
	}

	/**
	 * Read the RSEL flag which determines whether we display 24 or 25 lines of text
	 * 
	 * @return true if RSEL is set and we use 25 lines, otherwise false
	 */
	private boolean readRSEL() {
		return (registers[0x11] & 8) != 0;
	}

	/**
	 * Read the CSEL flag which determines whether we display 38 or 40 columns of text
	 * 
	 * @return true if CSEL is set and we use 40 columns, otherwise false
	 */
	private boolean readCSEL() {
		return (registers[0x16] & 8) != 0;
	}

	/**
	 * Read the DEN flag which tells whether the display is enabled
	 * 
	 * @return true if DEN is set, otherwise false
	 */
	private boolean readDEN() {
		return (registers[0x11] & 0x10) != 0;
	}

	/**
	 * Read the value of the raster line IRQ
	 * 
	 * @return raster line when to trigger an IRQ
	 */
	private int readRasterLineIRQ() {
		return (registers[0x12] & 0xff) + ((registers[0x11] & 0x80) << 1);
	}

	private boolean evaluateIsBadLine() {
		return areBadLinesEnabled && rasterY >= FIRST_DMA_LINE && rasterY <= LAST_DMA_LINE && (rasterY & 7) == yscroll;
	}

	/** Signal CPU interrupt if requested by VIC. */
	private void handleIrqState() {
		/* signal an IRQ unless we already signaled it */
		if ((irqFlags & irqMask & 0x0f) != 0) {
			if ((irqFlags & 0x80) == 0) {
				interrupt(true);
				irqFlags |= 0x80;
			}
		} else if ((irqFlags & 0x80) != 0) {
			interrupt(false);
			irqFlags &= 0x7f;
		}
	}

	/** Current graphics data */
	private int graphicsDataBuffer;
	/** Next 8 pixels in the graphics pipeline */
	private int graphicsDataBuffer2;
	/** Video matrix data one clock before */
	private byte latchedVmd2;
	/** Color data one clock before */
	private byte latchedColor2;
	/** xscroll value within window region text columns 1 .. 39 */
	private byte latchedXscroll;
	/** Is rasterYIRQ condition true? */
	private boolean rasterYIRQCondition;

	/**
	 * Get the video memory base address, which is determined by the inverted bits 0-1 of port A on CIA 2, plus the video matrix base address plus the character data base address plus the bitmap
	 * memory base address.
	 */
	private void determineVideoMemoryBaseAddresses() {
		videoMatrixBase = (registers[0x18] & 0xf0) << 6;
		charMemBase = (registers[0x18] & 0x0e) << 10;
		bitmapMemBase = (registers[0x18] & 0x08) << 10;
	}

	/**
	 * Set an IRQ flag and trigger an IRQ if the corresponding IRQ mask is set. The IRQ only gets activated, i.e. flag 0x80 gets set, if it was not active before.
	 */
	private void activateIRQFlag(final byte flag) {
		irqFlags |= flag;
		handleIrqState();
	}

	/**
	 * Read video matrix and color data for the next character
	 */
	private void doVideoMatrixAccess() {
		final int displayCycle = lineCycle - 24;
		videoMatrixData[displayCycle] = vicReadMemoryPHI2(videoMatrixBase | vc);
		colorData[displayCycle] = vicReadColorMemoryPHI2(vc);
	}

	private void drawSpritesAndGraphics() {
		int renderCycle = lineCycle - 26;
		if (renderCycle < 0) {
			renderCycle += CYCLES_PER_LINE;
		}

		spriteDataBuffer = drawSprites(graphicsDataBuffer);

		if (showBorderVertical || renderCycle > 40) {
			graphicsDataBuffer2 = currentGraphicsMode[0];
			mergeSpritesAndGraphics(8);
			graphicsDataBuffer = graphicsDataBuffer2;

			if (renderCycle == 39 && !readCSEL()) {
				showBorderMain = true;
			} else if (renderCycle == 40 && readCSEL()) {
				showBorderMain = true;
			}
		} else {
			/* draw screen for cycles 26 .. 65 */
			final int shift = 4 * latchedXscroll;
			if (latchedXscroll != 0) {
				mergeSpritesAndGraphics(latchedXscroll);
				/* Shift bits into left edge of graphicsdatabuffer that fall out of the
				 * right edge of the graphicsdatabuffer2. Just imagine that these two
				 * form a 64-bit long value with "graphicsdatabuffer2 graphicsdatabuffer"
				 * as the layout. */
				graphicsDataBuffer |= graphicsDataBuffer2 << 32-shift;
			}

			videoModeColors[4] = latchedColor2;
			videoModeColors[5] = (byte) (latchedVmd2 & 0xf);
			videoModeColors[6] = (byte) (latchedVmd2 >> 4 & 0xf);

			if (renderCycle < 40) {
				latchedVmd2 = isDisplayActive ? videoMatrixData[renderCycle] : 0;
				latchedColor2 = isDisplayActive ? colorData[renderCycle] : 0;
				if (isECM) {
					currentGraphicsMode = isBMM || isMCM ? pixelToColor[isMCM ? 5 : 4]
							: pixelToColor[6 + (latchedVmd2 >> 6 & 3)];
				} else if (isMCM && !isBMM) {
					currentGraphicsMode = pixelToColor[latchedColor2 >> 3];
					latchedColor2 &= 0x7;
				}
				graphicsDataBuffer2 = currentGraphicsMode[graphicsData & 0xff];
			} else {
				graphicsDataBuffer2 = currentGraphicsMode[0];
			}

			if ((renderCycle == 0 || renderCycle == 40) && readCSEL()) {
				mergeSpritesAndGraphics(8 - latchedXscroll);
				showBorderMain = renderCycle == 40;
			} else if ((renderCycle == 1 || renderCycle == 39) && !readCSEL()) {
				mergeSpritesAndGraphics(7 - latchedXscroll);
				showBorderMain = renderCycle == 39;
				mergeSpritesAndGraphics(1);
			} else {
				mergeSpritesAndGraphics(8 - latchedXscroll);
			}

			graphicsDataBuffer |= graphicsDataBuffer2 << shift;
			graphicsDataBuffer2 >>>= 32 - shift;
		}
	}

	/**
	 * In certain cases, CPU sees the stale bus data from VIC.
	 * VIC reads on every cycle, and this describes what it reads.
	 */
	public final byte getLastReadByte() {
		return graphicsData;
	}
	
	private void doPHI1Fetch() {
		switch (lineCycle) {
		case 2:
		case 3:
		case 65:
			/* idle gap accesses are cycles outside screen region */
			graphicsData = vicReadMemoryPHI1(0x3fff);
			return;

		case 4:
		case 6:
		case 8:
		case 10:
		case 12:
		case 14:
		case 16:
		case 18: {
			/* sprite data pointer byte */
			final int n = lineCycle - 4 >> 1;
			graphicsData = vicReadMemoryPHI1(videoMatrixBase | 0x03f8 | n);
			return;
		}

		case 5:
		case 7:
		case 9:
		case 11:
		case 13:
		case 15:
		case 17:
		case 19: {
			int n = (lineCycle - 5) >> 1;
			if (sprites[n].isDMA()) {
				int address = sprites[n].getCurrentByteAddress();
				graphicsData = vicReadMemoryPHI1(address);
			} else {
				graphicsData = vicReadMemoryPHI1(0x3fff);
			}
			return;
		}
			
		default: {
			/* 1, 25-64 */
			int n = lineCycle - 25;
			/* handle cycle 1 which is different on PAL and NTSC */
			if (lineCycle == 1) {
				if (CYCLES_PER_LINE == 63) {
					n = 39;
				} else {
					graphicsData = vicReadMemoryPHI1(0x3fff);
					return;
				}
			}
			
			int address = isECM ? 0x39ff : 0x3fff;
			if (isDisplayActive) {
				if (isBMM) {
					address &= bitmapMemBase | vc << 3 | rc;
				} else {
					address &= charMemBase | (videoMatrixData[n] & 0xff) << 3 | rc;
				}
				vc = vc + 1 & 0x3ff;
			}
			
			graphicsData = vicReadMemoryPHI1(address);
			return;
		}

		case 20:
		case 21:
		case 22:
		case 23:
		case 24: {
			/* dram refresh. */
			final int n = lineCycle - 20;
			final int offset = 0xff - rasterY * 5 - n & 0xff;
			graphicsData = vicReadMemoryPHI1(0x3f00 | offset);
		}
		}
	}

	/**
	 * Get memory address of sprite data.
	 * 
	 * @param n
	 *            sprite number
	 */
	private void fetchSpritePointer(final int n) {
		Sprite sprite = sprites[n];
		sprite.setPointerByte(graphicsData);
		final int x = sprites[n].getX();
		if (x >= 0x160 + 0x10 * n && x <= 0x167 + 0x10 * n) {
			sprite.event();
		}
		sprite.repeatPixels(21);

		sprite.setSpriteByte(2,
				vicReadMemoryPHI2(sprites[n].getCurrentByteAddress()));
	}

	/**
	 * Fetch 1 byte of memory starting from the sprite address.
	 * 
	 * @param n
	 *            sprite number
	 */
	private void fetchSpriteData(final int n) {
		Sprite sprite = sprites[n];
		sprite.setSpriteByte(1, graphicsData);
		sprite.setSpriteByte(0,
				vicReadMemoryPHI2(sprite.getCurrentByteAddress()));
		
		final int x = sprites[n].getX();
		if (x == 0x16f + 0x10 * n) {
			sprites[n].event();
		} else {
			handleSpriteVisibilityEvent(sprites[n]);
		}
	}

	/**
	 * Read VIC register.
	 * 
	 * @param register
	 *            Register to read.
	 */
	@Override
	public final byte read(int register) {
		register &= 0x3f;
		
		final byte value;
		switch (register) {
		// control register 1
		case 0x11: {
			value = (byte) ((registers[register] & 0x7f) | (rasterY & 0x100) >> 1);
			break;
		}

		// Raster Counter
		case 0x12:
			value = (byte) rasterY;
			break;

		case 0x13:
			value = lpx;
			break;

		case 0x14:
			value = lpy;
			break;

		case 0x19:
			// Interrupt Pending Register
			value = (byte) (irqFlags | 0x70);
			break;

		case 0x1a:
			// Interrupt Mask Register
			value = (byte) (irqMask | 0xf0);
			break;

		case 0x1e:
		case 0x1f: {
			// clear sprite collision registers after read
			final byte result = registers[register];
			registers[register] = 0;
			value = result;
			break;
		}

		// for addresses < $20 read from register directly, when < $2f set
		// bits of high nibble to 1, for >= $2f return $ff
		default:
			if (register < 0x20) {
				value = registers[register];
			} else if (register < 0x2f) {
				value = (byte) (registers[register] | 0xf0);
			} else {
				value = (byte) 0xff;
			}
		}

		setSpriteDataFromCPU(value);
		return value;
	}

	/** Display is enabled because badline condition was on for at least 1 clock */
	private final Event makeDisplayActive = new Event("Activate display due to badline") {
		@Override
		public void event() {
			isDisplayActive = true;
		}
	};
	/** AEC state was updated. */
	private final Event badLineStateChange = new Event("Update AEC signal") {
		@Override
		public void event() {
			setBA(!isBadLine);
		}
	};

	/** RasterY IRQ edge detector. */
	private final Event rasterYIRQEdgeDetector = new Event("RasterY changed") {
		@Override
		public void event() {
			final boolean oldRasterYIRQCondition = rasterYIRQCondition;
			rasterYIRQCondition = rasterY == readRasterLineIRQ();
			if (!oldRasterYIRQCondition && rasterYIRQCondition) {
				activateIRQFlag(IRQ_RASTER);
			}
		}
	};

	/* Is CIA asserting lightpen? */
	private boolean lpAsserted;
	
	/** Handle lightpen state change */
	private void lightpenEdgeDetector() {
		if (! lpAsserted) {
			return;
		}

		if (lpTriggered) {
			return;
		}	
		lpTriggered = true;
					
		/* light pen doesn't trigger on last line */
		if (rasterY == MAX_RASTERS - 1) {
			return;
		}
					
		lpx = (byte) getCurrentSpriteCycle();
		lpx ++;
		if (lpx == CYCLES_PER_LINE) {
			lpx = 0;
		}
			
		lpx <<= 2;
		lpx += context.phase() == Phase.PHI1 ? 1 : 2;
				
		lpy = (byte) (rasterY & 0xff);
		if (lineCycle == 9) {
			lpy ++;
		}
					
		activateIRQFlag(MOS656X_INTERRUPT_LP);
	}
	
	/**
	 * Write to VIC register.
	 * 
	 * @param register
	 *            Register to write to.
	 * @param data
	 *            Data byte to write.
	 */
	@Override
	public final void write(int register, final byte data) {
		register &= 0x3f;
		registers[register] = data;

		setSpriteDataFromCPU(data);
		
		switch (register) {
		// x-coordinate of a spriteID has been modified
		case 0x00:
		case 0x02:
		case 0x04:
		case 0x06:
		case 0x08:
		case 0x0a:
		case 0x0c:
		case 0x0e: {
			// determine sprite to modify
			final int n = register >> 1;
			sprites[n].setX(readSpriteXCoordinate(n));
			handleSpriteVisibilityEvent(sprites[n]);
			break;
		}

		// y-coordinate of a spriteID has been modified
		case 0x01:
		case 0x03:
		case 0x05:
		case 0x07:
		case 0x09:
		case 0x0b:
		case 0x0d:
		case 0x0f:
			sprites[register >> 1].setY(data & 0xff);
			break;

		case 0x10: {
			// bit 9 of a sprite x-coordinate has been modified
			// recalculate all sprite x-coordinates
			for (int i = 0; i < 8; i++) {
				sprites[i].setX(readSpriteXCoordinate(i));
				handleSpriteVisibilityEvent(sprites[i]);
			}
			break;
		}

		case 0x11:
			// the graphics mode might have changed plus y-scroll value
			yscroll = data & 7;
			determineGraphicsMode();

			/* display enabled at any cycle of line 48 enables badlines */
			if (rasterY == FIRST_DMA_LINE) {
				areBadLinesEnabled |= readDEN();
			}

			/* At cycle 9, it's too late for the CPU to affect VIC's decisions
			 * for that line -- no more PHI1 cycles with that rasterY value remain. */
			if (lineCycle != 9) {
				final int narrowing = readRSEL() ? 0 : 4;
				if (rasterY == FIRST_DMA_LINE + 3 + narrowing && readDEN()) {
					showBorderVertical = false;
				}
				if (rasterY == LAST_DMA_LINE + 4 - narrowing) {
					showBorderVertical = true;
				}
			}

			/* Re-evaluate badline condition */
			final boolean oldBadLine = isBadLine;
			isBadLine = evaluateIsBadLine();

			/* Schedule display activation if badlines are enabled for one clock. */
			if (!oldBadLine && isBadLine) {
				context.schedule(makeDisplayActive, 1, Phase.PHI2);
			}
			
			/* Within the display range, VIC signals AEC changes on next PHI1 */
			if ((isBadLine ^ oldBadLine) && lineCycle > 20 && lineCycle < 63) {
				context.schedule(badLineStateChange, 0, Phase.PHI1);
			}
			// $FALL-THROUGH$
			
		case 0x12:
			/* check raster Y irq condition changes at the next PHI1 */
			context.schedule(rasterYIRQEdgeDetector, 0, Phase.PHI1);
			break;

			// the sprite enable byte has changed
		case 0x15:
			for (int i = 0; i < 8; i++) {
				sprites[i].setEnabled((data & 1 << i) != 0);
			}
			break;

		case 0x16: {
			/* Blank the pixels left side of the character in buffer2. */
			graphicsDataBuffer2 ^= (graphicsDataBuffer2 ^ currentGraphicsMode[0]) & 0xffffffff << 4 * xscroll;
			xscroll = (byte) (registers[0x16] & 7);
			int renderCycle = lineCycle - 26;
			if (renderCycle < 0) {
				renderCycle += CYCLES_PER_LINE;
			}
			/* the xscroll adjustment is not seen on the last cycle. */
			if (renderCycle != 39) {
				latchedXscroll = xscroll;
			}
			determineGraphicsMode();
			break;
		}

		// the sprite y-expansion byte has changed
		case 0x17: {
			for (int i = 0; i < 8; i++) {
				final boolean expandY = (data & 1 << i) != 0;
				sprites[i].setExpandY(expandY, lineCycle == 24);
			}
			break;
		}

		// the cached video memory base addresses might have changed
		case 0x18:
			determineVideoMemoryBaseAddresses();
			break;

			// VIC Interrupt Flag Register
		case 0x19:
			irqFlags &= ~data & 0x0f | 0x80;
			handleIrqState();
			break;

			// IRQ Mask Register
		case 0x1a:
			irqMask = (byte) (data & 0x0f);
			handleIrqState();
			break;

			// the sprite priority byte has changed
		case 0x1b: {
			for (int i = 0, m = 1; i < 8; i++, m <<= 1) {
				this.sprites[i].setPriorityOverForegroundGraphics((data & m) == 0);
			}
			break;
		}

		// Sprites O-7 Multi-Color Mode Selection
		case 0x1c:
			for (int i = 0, m = 1; i < 8; ++i, m <<= 1) {
				this.sprites[i].setMulticolor((data & m) != 0);
			}
			break;

			// the sprite x-expansion byte has changed
		case 0x1d: {
			for (int i = 0, m = 1; i < 8; i++, m <<= 1) {
				this.sprites[i].setExpandX((data & m) != 0);
			}
			break;
		}

		// the border color was changed
		case 0x20:
			borderColor = (byte) (data & 0xf);
			break;

			// store color in internal registers
		case 0x21:
		case 0x22:
		case 0x23:
		case 0x24: {
			final int n = register - 0x21;
			videoModeColors[n] = (byte) (data & 0xf);
			break;
		}

		case 0x25:
			// sprite color one has been changed
			spriteColors[1] = (byte) (data & 0xf);
			registers[register] |= 0xf0;
			break;

		case 0x26:
			// sprite color three has been changed
			spriteColors[2] = (byte) (data & 0xf);
			registers[register] |= 0xf0;
			break;
		case 0x27:
		case 0x28:
		case 0x29:
		case 0x2a:
		case 0x2b:
		case 0x2c:
		case 0x2d:
		case 0x2e:
			// sprite color two has been changed, this can be done per sprite
			spriteColors[register - 0x27 + 3] = (byte) (data & 0xf);
			this.registers[register] |= 0xf0;
			break;
		}
	}

	/**
	 * If CPU reads/writes to VIC at just the cycle VIC is supposed to do a PHI2
	 * fetch for sprite data, the data fetched is set from the interaction with CPU.
	 * 
	 * @param data
	 */
	private void setSpriteDataFromCPU(byte data) {
		if (lineCycle >= 4 && lineCycle < 20) {
			Sprite damagedSprite = sprites[(lineCycle - 4) >> 1];
			damagedSprite.setSpriteByte((lineCycle & 1) == 0 ? 2 : 0, data);
		}
	}

	private int getCurrentSpriteCycle() {
		int spriteCoordinate = lineCycle - 23;
		if (spriteCoordinate < 0) {
			spriteCoordinate += CYCLES_PER_LINE;
		}
		return spriteCoordinate;
	}

	/**
	 * Schedule the rendering to begin
	 * 
	 * @param cyclesSinceFirstDisplayedCycle
	 * @param sprite
	 */
	private void handleSpriteVisibilityEvent(final Sprite sprite) {
		/* New coordinate in pipeline: cancel any pending event... */
		context.cancel(sprite);

		/* vic sprite x coordinate comparison will never trigger */
		if (sprite.getX() >> 3 >= CYCLES_PER_LINE) {
			return;
		}

		final int xpos = sprite.getX() >> 3;

			/* calculate cycles until sprite can display */
			int count = xpos - getCurrentSpriteCycle();
			/* Wrap to nearest CYCLE when comparison will match.
			 * 
			 * VICE test spritex proves that setting location very
			 * near will not work: it takes one cycle for VIC to
			 * adjust to the new sprite position.
			 */
			if (count <= 0) {
				count += CYCLES_PER_LINE;
			}
			if (count > CYCLES_PER_LINE) {
				count -= CYCLES_PER_LINE;
			}

			/* Delay sprite by 0 .. 7 pixels. */
			sprite.setDisplayStart((sprite.getX() & 7));
			context.schedule(sprite, count, Event.Phase.PHI2);
	}

	private final Event event = new Event("MOS656X") {
	@Override
	public final void event() {
		if (lineCycle == CYCLES_PER_LINE) {
			lineCycle = 0;
		}
		lineCycle++;

		if (graphicsRendering && (lineCycle >= 23 || lineCycle < 71 - CYCLES_PER_LINE)) {
			drawSpritesAndGraphics();
		} else {
			drawSprites(0);
		}

		doPHI1Fetch();
		
		switch (lineCycle) {

		case 1:
			/* PAL: graphics 39 access, NTSC: idle */
			for (final Sprite sprite : sprites) {
				if (sprite.isEnabled() && sprite.getY() == (rasterY & 0xff)) {
					sprite.beginDMA();
				}
			}

			setBA(!sprites[0].isDMA());
			break;

		case 2:
			/* idle gap access */
			for (final Sprite sprite : sprites) {
				if (sprite.isEnabled() && sprite.getY() == (rasterY & 0xff)) {
					sprite.beginDMA();
					sprite.setAllowDisplay(true);
				} else {
					sprite.setAllowDisplay(false);
				}
				sprite.expandYFlipFlop();
			}

			setBA(!sprites[0].isDMA());
			break;

		case 3:
			/* idle gap access */
			setBA(!sprites[0].isDMA() && !sprites[1].isDMA());
			break;

		case 4:
			/* sprite 0 pointer access */
			if (rc == 7) {
				vcBase = vc;
				isDisplayActive = isBadLine;
			}
			if (isDisplayActive) {
				rc = rc + 1 & 7;
			}

			for (final Sprite sprite : sprites) {
				if (sprite.isEnabled() && sprite.getY() == (rasterY & 0xff)) {
					sprite.setDisplay(true);
				}
				if (!sprite.isDMA()) {
					sprite.setDisplay(false);
				}

				sprite.initDmaAccess();
			}
			fetchSpritePointer(0);
			// already done: env.signalAEC(!sprites[0].isDMA() && !sprites[1].isDMA());
			break;

		case 5:
			/* sprite 0 data access */
			fetchSpriteData(0);
			setBA(!sprites[0].isDMA() && !sprites[1].isDMA() && !sprites[2].isDMA());
			break;

		case 6:
			/* sprite 1 pointer access */
			fetchSpritePointer(1);
			setBA(!sprites[1].isDMA() && !sprites[2].isDMA());
			break;

		case 7:
			/* sprite 1 data access */
			fetchSpriteData(1);
			setBA(!sprites[1].isDMA() && !sprites[2].isDMA() && !sprites[3].isDMA());
			break;

		case 8:
			/* sprite 2 pointer access */
			fetchSpritePointer(2);
			setBA(!sprites[2].isDMA() && !sprites[3].isDMA());
			break;

		case 9:
			/* sprite 2 data access */
			fetchSpriteData(2);
			setBA(!sprites[2].isDMA() && !sprites[3].isDMA() && !sprites[4].isDMA());
			break;

		case 10: {
			if (rasterY == MAX_RASTERS - 1) {
				vcBase = 0;
				/* last line is 1 cycle longer than it appears to be.
				 * To set rasterY = 0 at next cycle, we use a flag. */
				startOfFrame = true;
			} else {
				rasterY ++;
				rasterYIRQEdgeDetector.event();
			}
			
			// increase raster counter
			/* since raster Y just changed, we need to find out if we have entered a badline.
			 * the CPU can change the conditions any time, but since here VIC changes the raster,
			 * we *must* check badline on this cycle. In particular, linecrunch depends on canceling
			 * the bad line between the cycles 10 - 24. */
			if (rasterY == FIRST_DMA_LINE) {
				areBadLinesEnabled = readDEN();
			}


			/* mysteriously, isDisplayActive is determined already on this cycle.
			 * Normally it takes 1 clock to take effect. */
			isBadLine = evaluateIsBadLine();
			isDisplayActive |= isBadLine;

			final int narrowing = readRSEL() ? 0 : 4;
			if (rasterY == FIRST_DMA_LINE + 3 + narrowing && readDEN()) {
				showBorderVertical = false;
			}
			if (rasterY == LAST_DMA_LINE + 4 - narrowing) {
				showBorderVertical = true;
			}

			latchedXscroll = xscroll;
			colorFilterBuffer = 0;
			previousLineIndex = 0;
			if (rasterY == FIRST_DISPLAY_LINE) {
				/* current row odd? -> start with even, init, swap */
				linePaletteCurrent = (rasterY & 1) != 0 ? linePaletteEven : linePaletteOdd;
				graphicsRendering = true;
				nextPixel = 0;
				for (int i = 0; i < previousLineDecodedColor.length; i++) {
					previousLineDecodedColor[i] = linePaletteCurrent[0];
				}
			}
			linePaletteCurrent = linePaletteCurrent == linePaletteOdd ? linePaletteEven : linePaletteOdd;
			if (rasterY == LAST_DISPLAY_LINE + 1) {
				graphicsRendering = false;
				support.firePropertyChange("pixels", null, pixels);
			}

			// reset collision pointer to first pixel in line
			fetchSpritePointer(3);

			setBA(!sprites[3].isDMA() && !sprites[4].isDMA());
			break;
		}

		case 11: {
			/* sprite 3 data access */
			if (startOfFrame) {
				startOfFrame = false;
				rasterY = 0;
				rasterYIRQEdgeDetector.event();
				lpTriggered = false;
				lightpenEdgeDetector();
			}

			setBA(!sprites[3].isDMA() && !sprites[4].isDMA() && !sprites[5].isDMA());
			fetchSpriteData(3);
			break;
		}

		case 12:
			/* sprite 4 pointer access */
			fetchSpritePointer(4);
			setBA(!sprites[4].isDMA() && !sprites[5].isDMA());
			break;

		case 13:
			/* sprite 4 data access */
			fetchSpriteData(4);
			setBA(!sprites[4].isDMA() && !sprites[5].isDMA() && !sprites[6].isDMA());
			break;

		case 14:
			/* sprite 5 pointer access */
			fetchSpritePointer(5);
			setBA(!sprites[5].isDMA() && !sprites[6].isDMA());
			break;

		case 15:
			/* sprite 5 data access access */
			fetchSpriteData(5);
			setBA(!sprites[5].isDMA() && !sprites[6].isDMA() && !sprites[7].isDMA());
			break;

		case 16:
			/* sprite 6 pointer & data data access */
			fetchSpritePointer(6);
			setBA(!sprites[6].isDMA() && !sprites[7].isDMA());
			break;

		case 17:
			/* sprite 6 data access */
			fetchSpriteData(6);
			break;

		case 18:
			/* sprite 7 pointer access */
			fetchSpritePointer(7);
			setBA(!sprites[7].isDMA());
			break;

		case 19:
			/* sprite 7 data access */
			fetchSpriteData(7);
			break;

		case 20:
			/* dram refresh 0 cycle */
			setBA(true);
			break;

		case 21:
			/* dram refresh 1 */
			setBA(!isBadLine);
			break;

		case 22:
			/* dram refresh 2 */
			break;

		case 23:
			/* dram refresh 3 */
			vc = vcBase;
			if (isBadLine) {
				rc = 0;
			}
			break;

		case 24:
			/* dram refresh 4 */
			if (isBadLine) {
				doVideoMatrixAccess();
			}
			break;

		case 25:
			/* graphics memory access 0 */
			if (isBadLine) {
				doVideoMatrixAccess();
			}

			for (final Sprite sprite : sprites) {
				if (sprite.isDMA()) {
					sprite.finishDmaAccess();
				}
			}
			break;

		default:
			/* graphics memory access */
			if (isBadLine) {
				doVideoMatrixAccess();
			}
			break;

		case 64:
			/* NTSC only; graphics memory access */
			setBA(true);
			break;

		case 65:
			/* NTSC only: idle */
			break;
		}

		context.schedule(this, 1);
	}
	};
	
	public final void reset() {
		spriteLinkedListHead.nextVisibleSprite = null;
		for (final Sprite s : sprites) {
			s.consuming = false;
		}

		// clear the screen
		for (int i = 0; i < pixels.length; ++i) {
			pixels[i] = 0;
		}
		graphicsRendering = false;

		// reset all registers
		Arrays.fill(registers, (byte) 0);
		vc = 0;
		vcBase = 0;
		rc = 0;
		isDisplayActive = false;
		areBadLinesEnabled = false;
		rasterY = 0;
		graphicsData = 0;
		showBorderVertical = true;
		xscroll = 0;
		yscroll = 0;
		irqFlags = 0;
		irqMask = 0;
		nextPixel = 0;
		lpx = 0;
		lpy = 0;
		determineGraphicsMode();
		determineVideoMemoryBaseAddresses();
		lineCycle = 9; // preincremented at event
		context.schedule(event, 0, Phase.PHI1);
	}

	public static String credits() {
		return credit;
	}

	public final void addPropertyChangeListener(final PropertyChangeListener listener) {
		support.removePropertyChangeListener(listener);
		support.addPropertyChangeListener(listener);
	}

	public final Palette getPalette() {
		return palette;
	}

	public final void updatePalette() {
		palette.calculatePalette(Palette.buildPaletteVariant(model));
		System.arraycopy(palette.buildLineTable(), 0, combinedLines, 0, 128 * 256);
		System.arraycopy(palette.getEvenFiltered(), 0, linePaletteEven, 0, linePaletteEven.length);
		System.arraycopy(palette.getOddFiltered(), 0, linePaletteOdd, 0, linePaletteOdd.length);
	}

	public void triggerLightpen() {
		lpAsserted = true;
		lightpenEdgeDetector();
	}
	
	public void clearLightpen() {
		lpAsserted = false;
	}
	
	protected abstract void interrupt(boolean b);
	
	protected abstract void setBA(boolean b);

	protected abstract byte vicReadMemoryPHI1(int address);

	protected abstract byte vicReadMemoryPHI2(int address);
	
	protected abstract byte vicReadColorMemoryPHI2(int address);
}