package libsidutils;

/*
 * Copyright (c) 2002 Stefan Matthias Aust.  All Rights Reserved.
 *
 * You are granted the right to use this code in a) GPL based projects in
 * which case this code shall be also protected by the GPL, or b) in other
 * projects as long as you make all modifications or extensions to this
 * code freely available, or c) make any other special agreement with the
 * copyright holder.
 */
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import resid_builder.resid.ISIDDefs.ChipModel;
import resid_builder.resid.ISIDDefs.SamplingMethod;

/**
 * This class can read properties files in Microsoft .ini file style and provides an interface to read string, integer and boolean values. The .ini files has the following structure:
 * 
 * <pre>
 * ; a comment
 * [section]
 * key=value
 * </pre>
 * 
 * @author Stefan Matthias Aust (sma@3plus4.de)
 * @version 1
 */
public class IniReader {
	private final Map<String, Map<String, String>> sections = new LinkedHashMap<String, Map<String, String>>();

	protected Map<String, Map<String, String>> getReadSections() {
		return sections;
	}

	public IniReader(final String pathname) throws IOException {
		this(new FileReader(pathname));
	}

	public IniReader(final InputStream input) throws IOException {
		this(new InputStreamReader(input));
	}

	public IniReader(final Reader input) throws IOException {
		initialize(new BufferedReader(input));
	}

	public String[] listSections() {
		return sections.keySet().toArray(new String[] {});
	}

	private void initialize(final BufferedReader r) throws IOException {
		String section = null;
		String comment = "";
		String line;
		while ((line = r.readLine()) != null) {
			line = line.trim();
			if (line.equals("")) {
				continue;
			}

			if (line.startsWith(";")) {
				comment += line + System.getProperty("line.separator");
				continue;
			}

			/* parse section */
			if (line.startsWith("[")) {
				if (!line.endsWith("]")) {
					throw new IOException("] expected in section header");
				}
				section = line.substring(1, line.length() - 1);
				sections.put(section, new LinkedHashMap<String, String>());
				if (comment != "") {
					sections.get(section).put("_", comment);
					comment = "";
				}
				continue;
			}

			if (section == null) {
				throw new IOException("[section] header expected");
			}

			final int index = line.indexOf('=');
			if (index < 0) {
				throw new RuntimeException("key/value pair without '=' at: " + line);
			}

			final String key = line.substring(0, index).trim();
			final String value = line.substring(index + 1).trim();
			final Map<String, String> map = sections.get(section);
			map.put(key, value);
			if (comment != "") {
				map.put("_" + key, comment);
				comment = "";
			}
		}
	}

	public void save(String outputname) throws IOException {
		PrintWriter wr = new PrintWriter(new File(outputname));
		for (String section : sections.keySet()) {
			Map<String, String> unsaved = sections.get(section);
			if (unsaved.isEmpty()) {
				continue;
			}

			/* write section comment */
			if (unsaved.containsKey("_")) {
				wr.print(unsaved.get("_"));
			}
			wr.println("[" + section + "]");
			for (String key : unsaved.keySet()) {
				if (key.startsWith("_")) {
					continue;
				}
				/* write key comment */
				if (unsaved.containsKey("_" + key)) {
					wr.print(unsaved.get("_" + key));
				}
				wr.println(key + "=" + unsaved.get(key));
			}
			wr.println("");
		}
		wr.close();
	}

	public String getPropertyString(final String section, final String key, final String defaultValue) {
		final Map<String, String> map = sections.get(section);
		if (map != null) {
			final String value = map.get(key);
			if (value != null) {
				return value;
			}
		}
		return defaultValue;
	}

	public double getPropertyDouble(final String section, final String key, final double defaultValue) {
		final String s = getPropertyString(section, key, null);
		if (s != null) {
			return Double.parseDouble(s);
		}
		return defaultValue;
	}

	public int getPropertyInt(final String section, final String key, final int defaultValue) {
		final String s = getPropertyString(section, key, null);
		if (s != null && !s.equals("")) {
			return Integer.parseInt(s);
		}
		return defaultValue;
	}

	public boolean getPropertyBool(final String section, final String key, final boolean defaultValue) {
		final String s = getPropertyString(section, key, null);
		if (s != null) {
			return s.equalsIgnoreCase("true");
		}
		return defaultValue;
	}

	public int getPropertyTime(final String section, final String key, final int defaultValue) {
		final String s = getPropertyString(section, key, null);
		if (s != null) {
			return parseTime(s);
		}
		return defaultValue;
	}

	public void changeSetting(String section, String key, Object value) {
		Map<String, String> settings = sections.get(section);
		if (settings == null) {
			sections.put(section, settings = new HashMap<String, String>());
		}
		if (value == null) {
			settings.put(key, "");
		} else {
			settings.put(key, String.valueOf(value));
		}
	}

	/**
	 * Convert time from integer
	 * 
	 * @param str
	 *            time in 00:00 format
	 * @return seconds
	 */
	public static int parseTime(final String str) {
		int sep;
		int _time;

		// Check for empty string
		if (str.length() == 0) {
			return -1;
		}

		try {
			sep = str.lastIndexOf(':');
			if (sep == -1) {
				// User gave seconds
				_time = new Integer(str);
			} else {
				// Read in MM:SS format
				int val = Integer.valueOf(str.substring(0, sep));
				if (val < 0 || val > 99) {
					return -1;
				}
				_time = val * 60;
				val = Integer.valueOf(str.substring(sep + 1));
				if (val < 0 || val > 59) {
					return -1;
				}
				_time += val;
			}
		} catch (final NumberFormatException e) {
			return -1;
		}

		return _time;
	}

	private boolean dirty;

	public boolean isDirty() {
		return dirty;
	}
	public void setDirty(boolean dirty) {
		this.dirty = dirty;
	}
	
	public void makeDirty(String[] oldParam, String[] newParam) {
		if (oldParam.length!=newParam.length) {
			setDirty();
			return;
		}
		for (int i = 0; i < oldParam.length; i++) {
			if (oldParam[i] == null || !oldParam[i].equals(newParam[i])) {
				setDirty();
			}
		}
	}

	public void makeDirty(SamplingMethod oldParam, SamplingMethod newParam) {
		if (oldParam != newParam) {
			setDirty();
		}
	}

	public void makeDirty(ChipModel oldParam, ChipModel newParam) {
		if (oldParam != newParam) {
			setDirty();
		}
	}

	public void makeDirty(String oldParam, String newParam) {
		if (!oldParam.equals(newParam)) {
			setDirty();
		}
	}

	public void makeDirty(int oldParam, int newParam) {
		if (oldParam != newParam) {
			setDirty();
		}
	}

	public void makeDirty(boolean oldParam, boolean newParam) {
		if (oldParam != newParam) {
			setDirty();
		}
	}

	public void makeDirty(float oldParam, float newParam) {
		if (oldParam != newParam) {
			setDirty();
		}
	}

	private void setDirty() {
		dirty = true;
	}

}