package libsidutils;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;

public class STIL {
	private static final String COMMENT = "COMMENT: ";
	private static final String ARTIST = " ARTIST: ";
	private static final String TITLE = "  TITLE: ";
	private static final String AUTHOR = " AUTHOR: ";
	private static final String NAME = "   NAME: ";

	public class Info {
		public String name;
		public String author;
		public String title;
		public String artist;
		public String comment;

		public Info() {
		}

		@Override
		public String toString() {
			return "info";
		}
	}

	public class TuneEntry {
		public STILEntry parent;
		public String globalComment;
		public int tuneNo = -1;
		public ArrayList<Info> infos = new ArrayList<Info>();

		public TuneEntry(STILEntry entry) {
			parent = entry;
		}

		@Override
		public String toString() {
			return "" + tuneNo;
		}
	}

	public class STILEntry {
		/**
		 * comments
		 */
		String cmts;

		public String filename;
		public String globalComment;
		public ArrayList<TuneEntry> subtunes = new ArrayList<TuneEntry>();
		public ArrayList<Info> infos = new ArrayList<Info>();

		public STILEntry(String name) {
			filename = name;
		}

		@Override
		public String toString() {
			return "" + filename.substring(filename.lastIndexOf('/') + 1);
		}
	}

	ArrayList<STILEntry> entries = new ArrayList<STILEntry>();

	HashMap<String, STILEntry> fastMap = new HashMap<String, STILEntry>();

	private boolean fIsOpened;

	public STIL() {
		fIsOpened = false;
	}

	public int open(final String filename) {
		entries.clear();
		fastMap.clear();
		System.out.println("Using configfile STIL");
		try {
			FileInputStream fis = new FileInputStream(filename);
			BufferedReader r = new BufferedReader(new InputStreamReader(fis));
			String line;
			STILEntry entry = null;
			TuneEntry tuneEntry = null;
			Info lastInfo = null;
			String lastProp = null;
			boolean globalComment = true;
			StringBuffer cmts = new StringBuffer();
			while ((line = r.readLine()) != null) {
				if (line.startsWith("#")) {
					// ignore comment
					cmts.append(line + "\n");
					continue;
				}
				if (line.trim().length() == 0) {
					// ignore empty line
					continue;
				}
				if (line.startsWith("/")) {
					// relative filename
					globalComment = true;
					entry = new STILEntry(line);
					entries.add(entry);
					fastMap.put(line, entry);
					entry.cmts = cmts.toString();
					cmts.delete(0, cmts.length());
					Info info = new Info();
					lastInfo = info;
					entry.infos.add(info);
				} else if (line.startsWith("(#")) {
					if (entry == null) {
						throw new RuntimeException("Invalid format in STIL file: '(#' before '/'.");
					}
					// subtune
					globalComment = true;
					int end = line.indexOf(")");
					int tuneNo = Integer.parseInt(line.substring(2, end));
					// subtune number
					tuneEntry = new TuneEntry(entry);
					tuneEntry.tuneNo = tuneNo;
					entry.subtunes.add(tuneEntry);
					Info info = new Info();
					lastInfo = info;
					tuneEntry.infos.add(info);
				} else if (line.startsWith(COMMENT)) {
					// (global/local) comment
					lastProp = COMMENT;
					lastInfo = extract(COMMENT, line, entry, tuneEntry, lastProp, lastInfo, globalComment);
				} else if (line.startsWith(NAME)) {
					// name
					globalComment = false;
					lastProp = NAME;
					lastInfo = extract(NAME, line, entry, tuneEntry, lastProp, lastInfo, globalComment);
				} else if (line.startsWith(AUTHOR)) {
					// author
					globalComment = false;
					lastProp = AUTHOR;
					lastInfo = extract(AUTHOR, line, entry, tuneEntry, lastProp, lastInfo, globalComment);
				} else if (line.startsWith(TITLE)) {
					// title
					globalComment = false;
					lastProp = TITLE;
					lastInfo = extract(TITLE, line, entry, tuneEntry, lastProp, lastInfo, globalComment);
				} else if (line.startsWith(ARTIST)) {
					// artist
					globalComment = false;
					lastProp = ARTIST;
					lastInfo = extract(ARTIST, line, entry, tuneEntry, lastProp, lastInfo, globalComment);
				} else {
					// line belongs to the last entry
					lastInfo = extract("", line, entry, tuneEntry, lastProp, lastInfo, globalComment);
				}
			}
		} catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			return 1;
		} catch (IOException e) {
			System.err.println(e.getMessage());
			return 1;
		}
		fIsOpened = true;
		return 0;
	}

	private Info extract(String property, String line, STILEntry entry, TuneEntry tuneEntry, String lastProp, Info lastInfo, boolean globalComment) {
		if (entry.subtunes.isEmpty()) {
			// not subtune specific
			if (NAME.equals(property)) {
				if (lastInfo.name != null) {
					Info info = new Info();
					lastInfo = info;
					entry.infos.add(info);
				}
				lastInfo.name = line.substring(property.length());
			} else if (AUTHOR.equals(property)) {
				if (lastInfo.author != null) {
					Info info = new Info();
					lastInfo = info;
					entry.infos.add(info);
				}
				lastInfo.author = line.substring(property.length());
			} else if (TITLE.equals(property)) {
				if (lastInfo.title != null) {
					Info info = new Info();
					lastInfo = info;
					entry.infos.add(info);
				}
				lastInfo.title = line.substring(property.length());
			} else if (ARTIST.equals(property)) {
				if (lastInfo.artist != null) {
					Info info = new Info();
					lastInfo = info;
					entry.infos.add(info);
				}
				lastInfo.artist = line.substring(property.length());
			} else if (COMMENT.equals(property)) {
				if (globalComment) {
					entry.globalComment = line.substring(property.length());
					return lastInfo;
				}
				if (lastInfo.comment != null) {
					Info info = new Info();
					lastInfo = info;
					entry.infos.add(info);
				}
				lastInfo.comment = line.substring(property.length());
			} else {
				// line belongs to the last entry
				if (NAME.equals(lastProp)) {
					lastInfo.name += "\n" + line;
				} else if (AUTHOR.equals(lastProp)) {
					lastInfo.author += "\n" + line;
				} else if (TITLE.equals(lastProp)) {
					lastInfo.title += "\n" + line;
				} else if (ARTIST.equals(lastProp)) {
					lastInfo.artist += "\n" + line;
				} else if (COMMENT.equals(lastProp)) {
					if (globalComment) {
						entry.globalComment += "\n" + line;
						return lastInfo;
					}
					lastInfo.comment += "\n" + line;
				}
			}
		} else {
			// subtune specific
			if (NAME.equals(property)) {
				if (lastInfo.name != null) {
					Info info = new Info();
					lastInfo = info;
					tuneEntry.infos.add(info);
				}
				lastInfo.name = line.substring(property.length());
			} else if (AUTHOR.equals(property)) {
				if (lastInfo.author != null) {
					Info info = new Info();
					lastInfo = info;
					tuneEntry.infos.add(info);
				}
				lastInfo.author = line.substring(property.length());
			} else if (TITLE.equals(property)) {
				if (lastInfo.title != null) {
					Info info = new Info();
					lastInfo = info;
					tuneEntry.infos.add(info);
				}
				lastInfo.title = line.substring(property.length());
			} else if (ARTIST.equals(property)) {
				if (lastInfo.artist != null) {
					Info info = new Info();
					lastInfo = info;
					tuneEntry.infos.add(info);
				}
				lastInfo.artist = line.substring(property.length());
			} else if (COMMENT.equals(property)) {
				if (globalComment) {
					tuneEntry.globalComment = line.substring(property.length());
					return lastInfo;
				}
				if (lastInfo.comment != null) {
					Info info = new Info();
					lastInfo = info;
					tuneEntry.infos.add(info);
				}
				lastInfo.comment = line.substring(property.length());
			} else {
				// line belongs to the last entry
				if (NAME.equals(lastProp)) {
					lastInfo.name += "\n" + line;
				} else if (AUTHOR.equals(lastProp)) {
					lastInfo.author += "\n" + line;
				} else if (TITLE.equals(lastProp)) {
					lastInfo.title += "\n" + line;
				} else if (ARTIST.equals(lastProp)) {
					lastInfo.artist += "\n" + line;
				} else if (COMMENT.equals(lastProp)) {
					if (globalComment) {
						tuneEntry.globalComment += "\n" + line;
						return lastInfo;
					}
					lastInfo.comment += "\n" + line;
				}
			}
		}
		return lastInfo;
	}

	public STILEntry getSTIL(String name) {
		return fastMap.get(name);
	}

	public boolean isOpened() {
		return fIsOpened;
	}
}
