package libsidutils;

import java.io.IOException;
import java.security.AccessControlException;
import java.util.StringTokenizer;

import libsidplay.sidtune.SidTune;

public class SidDatabase {
	private static final String ERR_DATABASE_CORRUPT = "SID DATABASE ERROR: Database seems to be corrupt.";
	private static final String ERR_NO_DATABASE_LOADED = "SID DATABASE ERROR: Songlength database not loaded.";
	private static final String ERR_NO_SELECTED_SONG = "SID DATABASE ERROR: No song selected for retrieving song length.";
	private static final String ERR_UNABLE_TO_LOAD_DATABASE = "SID DATABASE ERROR: Unable to load the songlength database.";

	private IniReader database;
	private String errorString;
	private boolean fIsOpened;

	private int parseTimeStamp(final String arg) {
		/*
		 * Read in m:s format at most. Could use a system function if available.
		 */
		int seconds = 0;
		int passes = 2; // minutes, seconds
		boolean gotDigits = false;
		int argPos = 0;
		while (passes-- > 0) {
			if (Character.isDigit(arg.charAt(argPos))) {
				final String sub = arg.substring(argPos);
				int t;
				int indexOf = sub.indexOf(':');
				if (indexOf != -1) {
					t = Integer.parseInt(sub.substring(0, indexOf));
				} else {
					indexOf = sub.indexOf('(');
					if (indexOf != -1) {
						t = Integer.parseInt(sub.substring(0, indexOf));
					} else {
						indexOf = sub.indexOf(';');
						if (indexOf != -1) {
							t = Integer.parseInt(sub.substring(0, indexOf));
						} else {
							t = Integer.parseInt(sub);
						}
					}
				}
				seconds += t;
				gotDigits = true;
			}
			while (argPos < arg.length()
					&& Character.isDigit(arg.charAt(argPos))) {
				++argPos;
			}
			if (argPos < arg.length() && arg.charAt(argPos) == ':') {
				seconds *= 60;
				++argPos;
			}
		}

		// Handle -:-- time stamps and old 0:00 entries which
		// need to be rounded up by one second.
		if (!gotDigits) {
			seconds = 0;
		} else if (seconds == 0) {
			++seconds;
		}

		return seconds;
	}

	private short timesFound(final String str) {
		/*
		 * Try and determine the number of times read back. Used to check
		 * validility of times in database.
		 */
		short count = 0;
		int strPos = 0;
		while (strPos < str.length()) {
			if (str.charAt(strPos++) == ':') {
				count++;
			}
		}
		return count;
	}

	public SidDatabase() {
		database = null;
		fIsOpened = false;
	}

	public int open(final String filename) {
		close();
		try {
			database = new IniReader(filename);
		} catch (final AccessControlException e) {
			errorString = ERR_UNABLE_TO_LOAD_DATABASE;
			return -1;
		} catch (final IOException e) {
			errorString = ERR_UNABLE_TO_LOAD_DATABASE;
			return -1;
		}
		fIsOpened = true;
		return 0;
	}

	public void close() {
	}

	public int length(final SidTune tune) {
		final int song = tune.getInfo().currentSong;
		if (song == 0) {
			errorString = ERR_NO_SELECTED_SONG;
			return -1;
		}
		final String md5 = tune.getMD5Digest();
		if (md5 == null) {
			return 0;
		}
		return length(md5, song);
	}

	public int length(final String md5,
			final int song) {
		int time = 0;
		String timeStamp = null;

		if (database == null) {
			errorString = ERR_NO_DATABASE_LOADED;
			return -1;
		}

		// Read Time (and check times before hand)
		final String value = database.getPropertyString("Database", md5, null);
		// If value return is null then no entry found in database
		if (value != null) {
			final StringTokenizer st = new StringTokenizer(value, " ", false);
			for (int i = 0; i < song; i++) {
				if (!st.hasMoreTokens()) { // No time found
					errorString = ERR_DATABASE_CORRUPT;
					return -1;
				}
				timeStamp = st.nextToken();
				// Validate Time
				if (timesFound(timeStamp) != 1) {
					errorString = ERR_DATABASE_CORRUPT;
					return -1;
				}
			}

			// Parse time stamp
			time = parseTimeStamp(timeStamp);
		}

		return time;
	}

	public final String error() {
		return errorString;
	}

	public boolean isOpened() {
		return fIsOpened;
	}
}
