package netsiddev;

import java.awt.AWTException;
import java.awt.EventQueue;
import java.awt.Image;
import java.awt.MenuItem;
import java.awt.PopupMenu;
import java.awt.SystemTray;
import java.awt.Toolkit;
import java.awt.TrayIcon;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.PrintStream;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import javax.sound.sampled.LineUnavailableException;

import libsidplay.common.ISID2Types;
import resid_builder.resid.SID;
import resid_builder.resid.ISIDDefs.ChipModel;
import resid_builder.resid.ISIDDefs.SamplingMethod;
import sidplay.audio.AudioConfig;
import sidplay.audio.JavaSound;

/**
 * JSIDPlay2 SID-emulation integration protocol. It is possible to have
 * JSIDPlay2 take over the duty of the SID playback for a C64 emulator/player.
 * Every jsidplay2 instance tries to open port 6581 where they will listen to
 * connections that describe SID activity with the following protocol.
 * 
 * GENERAL OVERVIEW
 * 
 * Protocol is structured as a request-response protocol:
 * 
 * - Requests are variable length, with minimum packet size 4 bytes. There are
 *   3 fields and an amorphous data blob:
 * 
 *   - 8-bit unsigned as command.
 *   - 8-bit unsigned as SID number.
 *   - 16-bit unsigned as length of data attached to header in bytes.
 *   - Data (if any)
 * 
 * - All commands are ACKed with 1-byte response code that is OK, BUSY, ERR or
 *   READ. ERR and READ responses have further data that the client should
 *   read. The responses are written back as soon as possible, but since many
 *   commands imply some server-side processing, it can take a short while to
 *   do the processing.
 * 
 *   - OK means that the commands were accepted by server and can be discarded
 *     by client.
 *   - BUSY means that no part of the current command was accepted due to filled
 *     queue condition, and that client should wait and retry it later.
 *     1 millisecond could be a suitable delay before retry. Queue length is
 *     limited both by number of events and maximum time drift between playback
 *     clock and client clock.
 *   - ERR means that emulation has experienced a severe failure and
 *     will not recover; server will close this connection. Before closing, it
 *     will attempt to write the Java stack trace to client.
 *   - READ: successful read operation, one byte value follows that is the
 *     value read from SID.
 * 
 * Maximum packet length is 64k + header length. It is suggested that only
 * short packets are transmitted, in the order of 1k and containing no more
 * than about 1 ms worth of events. Otherwise the client-server desync brings
 * jitter that may have unpleasant consequences. At the limit it's possible to
 * simply send a fixed header that describes a single write with each packet,
 * but this is probably measurably less efficient.
 * 
 * COMMAND OVERVIEW
 * 
 * Structure of data is specific to command. Some commands require data,
 * others will not use data even if such was provided. Some commands
 * require specific lengths for the data packets. If data length is not
 * correct, results are undefined.
 * 
 * Known commands are identified by small integers, starting from 0:
 * 
 * - FLUSH (0): destroy queued data and cease audio production.
 *   - data packet is ignored, and should be 0 length
 *   - should probably be followed by RESET (SID is in unpredictable state).
 *   - always returns OK
 *
 * - WAIT_FLUSH (1): wait until buffers have drained.
 *   - data packet is ignored, and should be 0 length.
 *   - returns BUSY until write queue is flushed, otherwise OK.
 *   - should probably be followed by RESET.
 *   - audio production ceases shortly after OK.
 *
 * - MUTE (2): mute/unmute a voice
 *   - data packet should contain two 8-bit unsigned bytes:
 *     - the voice number from 0 to 2
 *     - 0 or 1 to disable/enable voice
 *     - this command bypasses buffer and takes immediate effect.
 *   - always returns OK
 *
 * - RESET (3): reset SID, setting volume to provided value.
 *   - data packet should be a 8-bit unsigned value which is written to volume
 *     register.
 *   - reset clears the MUTE status on voices and syncs oscillators
 *   - this command bypasses buffer and takes immediate effect. It should be
 *     issued only when buffer is known to be empty.
 *   - always returns OK
 *
 * - TRY_DELAY (4): inform emulation that no events have occured for a given count of cycles
 *   - data packet should be 16-bit unsigned value interpreted as delay in C64 clocks.
 *   - allows audio generation in absence of other activity.
 *   - returns BUSY if there is already enough data for playback, otherwise OK.
 *
 * - TRY_WRITE (5): try to queue a number of write-to-sid events.
 *   - data packet should be 4*N bytes long, repeating this structure:
 *     - 16-bit unsigned value interpreted as delay before the write in C64 clocks.
 *     - 8-bit unsigned SID register number from 0x00 to 0x1f.
 *     - 8-bit unsigned data value to write
 *   - returns BUSY if there is already enough data for playback, otherwise OK.
 * 
 * - READ (6): reads SID chip register.
 *   - data packet should be a 3 bytes long:
 *     - 16-bit unsigned value interpreted as delay before the read in C64 clocks.
 *     - 8-bit unsigned SID register number from 0x00 to 0x1f.
 *   - returns 2 bytes: READ and a data byte, which is the read value from SID.
 *
 * - VERSION (7): returns the version of the SID Network protocol.
 *   - data packet is ignored, and should be 0 length.
 *   - returns 2 bytes: VERSION and a data byte, which is the version of the SID Network protocol.
 * 
 * 
 * NOTES
 * 
 * The delay values do not contain the time taken to write the value to SID chip, and
 * a delay length of 0 between writes is impossible to achieve with a true C64 system,
 * although this emulator will accept it and execute several writes on the same clock.
 * 
 * At start of connection, the SID starts from RESET state with volume=0 and empty buffer.
 * 
 * Future expansion:
 * 
 * - stereo SID support
 * - select filter (chip) type remote
 * - select filter type in dialog
 * - route chips to left, right or mono.
 * - pal/ntsc select remote (not as device)
 * - implement protocol via UDP
 * - combine read and write in one data packet
 * @author Ken Hndel
 * @author Antti S. Lankila
 * @author Wilfred Bos
 */
public class NetworkSIDDevice extends Thread {
	private static AboutBox abox = null;
	
	private static final byte SID_NETWORK_PROTOCOL_VERSION = 1;
	
	private static final int DEFAULT_SID_DEVICES = 8;
	private int numberOfSIDDevices = DEFAULT_SID_DEVICES;
	
	/** Host name of the client */
	private static final String CLIENT_HOST = "localhost";

	/** Port of the client */
	private static final int CLIENT_PORT = 6581;

	/**
	 * Maximum allowed drift between jsidplay2 and client (C64 clocks).
	 * We need at least a few frames because buffer drying up is handled
	 * via stop at audio playback.
	 */
	private static final int LATENCY = 180000;

	/** Available commands to clients */
	enum Command {
		FLUSH,
		WAIT_FLUSH,
		MUTE,
		RESET,
		TRY_DELAY,
		TRY_WRITE,
		READ,
		VERSION
	}

	/** Responses generated by server. */
	enum Response {
		OK,
		BUSY,
		ERR,
		READ,
		VERSION
	}
	
	private class EventConsumerThread extends Thread {
		private long playbackClock;

		protected long getPlaybackClock() {
			return playbackClock;
		}

		/**
		 * Queue with SID writes from client. We reserve a space
		 * assuming writes come at most one every 10 cpu clocks.
		 */
		private final BlockingQueue<SIDWrite> sidCommandQueue = new ArrayBlockingQueue<SIDWrite>(LATENCY / 10);

		protected BlockingQueue<SIDWrite> getSidCommandQueue() {
			return sidCommandQueue;
		}

		/** SIDs that generate output */
		private final SID[] sid = new SID[numberOfSIDDevices];

		private final JavaSound driver = new JavaSound();
		
		protected EventConsumerThread() {
			// Read filter settings and create filter
			for (int i = 0; i < numberOfSIDDevices; i ++) {
				sid[i] = new SID();
				if ((i & 1) == 0) {
					// FilterZrx6581R4AR_2286
					sid[i].setChipModel(ChipModel.MOS6581);
					sid[i].getFilter().setCurveProperties(1300000f, 19000000f, 1.0066f, 18000);
					sid[i].getFilter().setDistortionProperties(0.5f, 3300000.0f, 1.0f);
					sid[i].set6581VoiceNonlinearity(0.96f);
				} else {
					// FilterTrurl8580R5_3691
					sid[i].setChipModel(ChipModel.MOS8580);
					sid[i].getFilter().setCurveProperties(6.55f, 20.0f, 0, 0);
					sid[i].getFilter().setDistortionProperties(0, 0, 1.0f);
				}
			}
		}
		
		@Override
		public void interrupt() {
			// make sure to close the driver to avoid problems with interrupting the thread
			if (driver != null) {
				driver.close();
			}
			
			super.interrupt();
		}		

		@Override
		public void run() {
			if (driver == null) {
				System.err.println("No sound driver available");
				return;
			}

			System.out.println("Audio starting with " + sidCommandQueue.size() + " commands queued.");

			final AudioConfig aCfg = new AudioConfig();
			/* FIXME: read audio frequency */
			aCfg.channels = 2;

			/* FIXME: NTSC/PAL, sampling mode */
			for (int i = 0; i < numberOfSIDDevices; i ++) {
				if (i == 0 || i == 1) {
					sid[i].setSamplingParameters(ISID2Types.Clock.PAL.getCpuFrequency(), SamplingMethod.RESAMPLE, aCfg.frequency, 20000);
				} else if (i == 2 || i == 3) {
					sid[i].setSamplingParameters(ISID2Types.Clock.NTSC.getCpuFrequency(), SamplingMethod.RESAMPLE, aCfg.frequency, 20000);
				} else if (i == 4 || i == 5) {
					sid[i].setSamplingParameters(ISID2Types.Clock.PAL.getCpuFrequency(), SamplingMethod.DECIMATE, aCfg.frequency, 20000);
				} else if (i == 6 || i == 7) {
					sid[i].setSamplingParameters(ISID2Types.Clock.NTSC.getCpuFrequency(), SamplingMethod.DECIMATE, aCfg.frequency, 20000);
				}
			}

			// Test devices are:
			// 0: MOS6581, PAL, Resample
			// 1: MOS8580, PAL, Resample
			// 2: MOS6581, NTSC, Resample
			// 3: MOS8580, NTSC, Resample
			// 4: MOS6581, PAL, Decimate
			// 5: MOS8580, PAL, Decimate
			// 6: MOS6581, NTSC, Decimate
			// 7: MOS8580, NTSC, Decimate
			try {
				driver.open(aCfg, null);
				try {
					final ByteBuffer output = driver.buffer();
					final int framesPerFragment = output.capacity() / 4;
					float[] audioBuffer = new float[framesPerFragment + 10000];
	
					int audioBufferPos = 0;
	
					while (!isInterrupted()) {
						/* FIXME support sid[n] and audio mixing from all involved SIDs */
						SIDWrite write = sidCommandQueue.poll();
						if (write == null) {
							Thread.yield();
							for (int i = 0; i < 50 && sidCommandQueue.peek() == null; i++) {
								Thread.sleep(1);
							} 
							
							if (sidCommandQueue.peek() == null) {
								/* buffer ran dry, cease audio production immediately. */
								driver.pause();
								System.out.println("Write queue underrun, paused...");
								
								while (sidCommandQueue.peek() == null) {
									// when this thread is interrupted during sleep, it will throw an interrupted exception
									// so no need to check isInterrupted in this loop
									Thread.sleep(20);
								}
								// always sleep to make sure the sidCommandQueue is filled enough
								// to avoid hick ups when a tune starts and to avoid another buffer underrun.
								Thread.sleep(50);
								System.out.println("Resuming with " + sidCommandQueue.size() + " commands queued.");
							} else {
								System.out.println("Write queue underrun, continued...");
								System.out.println("Resuming with " + sidCommandQueue.size() + " commands queued.");
							}
							write = sidCommandQueue.take();
						}
						
						final SID s = sid[write.chip];
						while (write.cycles > 0 && !isInterrupted()) {
							/* Calculate 10k clocks at a time.
							 * This should produce audio some < 1 kB at a time. */
							int piece = Math.min(write.cycles, 5000);
							audioBufferPos += s.clock(Math.min(piece, audioBuffer.length-audioBufferPos), audioBuffer, audioBufferPos);
							playbackClock += piece;
							write.cycles -= piece;
	
							/* generate all audio we can */
							if (audioBufferPos >= framesPerFragment) {
								output.clear();
								for (int i = 0; i < framesPerFragment; i ++) {
									int value = Math.round(audioBuffer[i] * 32767);
									if (value > 32767) {
										value = 32767;
									}
									if (value < -32768) {
										value = -32768;
									}
									output.putShort((short) value);
									output.putShort((short) value);
								}
								driver.write();
								/* move remainder of data to start of buffer */
								System.arraycopy(audioBuffer, framesPerFragment, audioBuffer, 0, audioBufferPos - framesPerFragment);
								audioBufferPos -= framesPerFragment;
							}
						}
						
						/* do the write, if this is a write command */
						if (write.reg >= 0) {
							s.write(write.reg, write.value);
						}
					}
				} catch (final InterruptedException e) {
				} finally {
					driver.close();
					System.out.println("Audio stopped.");
				}
			} catch (final LineUnavailableException e) {
				e.printStackTrace();
			}
		}

		protected void mute(final int sidNumber, final int voiceNo, final boolean mute) {
			sid[sidNumber].mute(voiceNo, mute);
		}

		protected void reset(final int sidNumber, final byte volume) {
			sid[sidNumber].reset();
			sid[sidNumber].write(0x18, volume);
			sid[sidNumber].mute(0, true);
			sid[sidNumber].mute(1, true);
			sid[sidNumber].mute(2, true);
		}
	}

	private class ConnectionHandler extends Thread {
		/** Accepted client socket */
		private final Socket client;

		/** Integration protocol data from client */
		private final DataInputStream clientIn;

		/** Response protocol data to client */
		private final DataOutputStream clientOut;

		/** Shadow SID clocked with client to read from */
		private final SID[] sidRead = new SID[numberOfSIDDevices];
		
		/** Construct a new audio player for connected client */
		protected ConnectionHandler(final Socket client) throws IOException {
			this.client = client;
			clientIn = new DataInputStream(client.getInputStream());
			clientOut = new DataOutputStream(client.getOutputStream());

			for (int i = 0; i < numberOfSIDDevices; i ++) {
				sidRead[i] = new SID();
				if ((i & 1) == 0) {
					// FilterZrx6581R4AR_2286
					sidRead[i].setChipModel(ChipModel.MOS6581);
					sidRead[i].getFilter().setCurveProperties(1300000f, 19000000f, 1.0066f, 18000);
					sidRead[i].getFilter().setDistortionProperties(0.5f, 3300000.0f, 1.0f);
					sidRead[i].set6581VoiceNonlinearity(0.96f);
				} else {
					// FilterTrurl8580R5_3691
					sidRead[i].setChipModel(ChipModel.MOS8580);
					sidRead[i].getFilter().setCurveProperties(6.55f, 20.0f, 0, 0);
					sidRead[i].getFilter().setDistortionProperties(0, 0, 1.0f);
				}
			}
		}

		@Override
		public void interrupt() {
			try {
				// make sure that DataInputStream is not blocking the thread by closing the socket
				if (client != null && !client.isClosed()) {
					client.close();
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
			super.interrupt();
		}
		
		@Override
		public void run() {
			final EventConsumerThread eventConsumerThread = new EventConsumerThread();
			eventConsumerThread.start();

			try {
				final byte[] bytes = new byte[65536];
				final byte[] rsp = new byte[2];
				final BlockingQueue<SIDWrite> sidCommandQueue = eventConsumerThread.getSidCommandQueue();

				long inputClock = 0;
				int pendingCycles = 0;
				
				// read commands until socket connection will be lost
				final Command[] commands = Command.values();
				while (!isInterrupted()) {
					// if no data is available then sleep 1 millisecond. this will increase system performance a lot.
					if (clientIn.available() <= 0) {
						// when this thread is interrupted during sleep, it will throw an interrupted exception
						// so no need to check isInterrupted in this loop.
						Thread.sleep(1);
					}
					final int commandByte = clientIn.readUnsignedByte();
					if (commandByte >= commands.length) {
						// ignore rest of buffer since command is unknown
						clientIn.skip(1);
						clientIn.readFully(bytes, 0, clientIn.readUnsignedShort());
						
						clientOut.writeByte(Response.ERR.ordinal());
						System.err.println("Invalid network SID device protocol command: " + commandByte);
						continue;
					}
					
					final Command command = commands[commandByte];
					int sidNumber = clientIn.readUnsignedByte();
					if (sidNumber >= numberOfSIDDevices) {
						sidNumber = 0;	// default to first SID
					}
					int dataLength = clientIn.readUnsignedShort();
					clientIn.readFully(bytes, 0, dataLength);

					final boolean isBufferFull = (sidCommandQueue.remainingCapacity() < dataLength) 
												 || (inputClock - eventConsumerThread.getPlaybackClock() > LATENCY);
							
					/* tell client to retry later if we have enough data. */
					if ((command == Command.TRY_WRITE || (command == Command.TRY_DELAY)) && isBufferFull) {
						clientOut.write(Response.BUSY.ordinal());
						continue;
					}

					final Response responseCode;
					byte responseData = 0;

					switch (command) {
					case FLUSH:
						sidCommandQueue.clear();
						pendingCycles = 0;
						inputClock = eventConsumerThread.getPlaybackClock();
						responseCode = Response.OK;
						break;
					case WAIT_FLUSH:
						if (sidCommandQueue.size() != 0) {
							responseCode = Response.BUSY;
						} else {
							responseCode = Response.OK;
						}
						break;
					case MUTE:
						final int voiceNo = bytes[0];
						final boolean mute = bytes[1] != 0;
						eventConsumerThread.mute(sidNumber, voiceNo, mute);
						responseCode = Response.OK;
						break;
					case RESET:
						final byte volume = bytes[0];
						eventConsumerThread.reset(sidNumber, volume);
						pendingCycles = 0;
						sidRead[sidNumber].reset();
						sidRead[sidNumber].write(0x18, volume);
						responseCode = Response.OK;
						break;
					case TRY_WRITE:
						for (int i = 0; i < dataLength; i += 4) {
							final int writeCycles = (bytes[i] & 0xff) << 8 | bytes[i+1] & 0xff;
							final byte reg = bytes[i+2];
							final byte value = bytes[i+3];
							inputClock += writeCycles + pendingCycles;
							sidCommandQueue.put(new SIDWrite(sidNumber, reg & 0x1f, value, writeCycles + pendingCycles));
							sidRead[sidNumber].clockSilent(writeCycles);
							sidRead[sidNumber].write(reg & 0x1f, value);
							pendingCycles = 0;
						}
						responseCode = Response.OK;
						break;
					case TRY_DELAY:
						// not a real write, just something to pass the time.
						final int delayCycles = (bytes[0] & 0xff) << 8 | bytes[1] & 0xff;
						inputClock += delayCycles + pendingCycles;
						sidCommandQueue.add(new SIDWrite(sidNumber, -1, (byte) 0, delayCycles + pendingCycles));
						sidRead[sidNumber].clockSilent(delayCycles);
						pendingCycles = 0;
						responseCode = Response.OK;
						break;
					case READ:
						// always read immediately after given cycles since previous writes
						// are already processed for the sidRead device
						final int readCycles = (bytes[0] & 0xff) << 8 | bytes[1] & 0xff;
						final byte register = bytes[2];
						if (isBufferFull) {
							// buffer full so add cycles to pending cycles for next write
							pendingCycles += readCycles;
						} else {
							inputClock += readCycles + pendingCycles;
							sidCommandQueue.add(new SIDWrite(sidNumber, -1, (byte) 0, readCycles + pendingCycles));
							pendingCycles = 0;
						}
						sidRead[sidNumber].clockSilent(readCycles);
						responseData = sidRead[sidNumber].read(register & 0x1f);
						responseCode = Response.READ;
						break;
					case VERSION:
						responseData = SID_NETWORK_PROTOCOL_VERSION;
						responseCode = Response.VERSION;
						break;
					default:
						responseCode = Response.ERR;
						break;
					}
					
					if ((responseCode == Response.READ) || (responseCode == Response.VERSION)) {
						// transfer in one write to ensure,
						// that the client receives a
						// two byte response
						rsp[0] = (byte) responseCode.ordinal();
						rsp[1] = responseData;
						clientOut.write(rsp);
					} else {
						clientOut.write(responseCode.ordinal());
					}
				}
			} catch (final EOFException e) {
			} catch (final Exception e) {
				/* Try to tell client that we encountered a fatal error. If
				 * it doesn't work out, that's OK, we'll close the socket
				 * anyway.
				 */
				if (clientOut != null) {
					try {
						clientOut.writeByte(Response.ERR.ordinal());
						e.printStackTrace(new PrintStream(clientOut));
					}
					catch (final IOException ioE) {
					}
				}
				e.printStackTrace();
			}
			finally {
				try {
					if (clientIn != null)
						clientIn.close();
					if (clientOut != null)
						clientOut.close();
					if (client != null && !client.isClosed())
						client.close();
				} catch (final IOException e1) {
				}
				eventConsumerThread.interrupt();
				try {
					eventConsumerThread.join();
				} catch (InterruptedException e) {
				}
			}
		}
	}

	@Override
	public void run() {
		final List<ConnectionHandler> connections = new ArrayList<ConnectionHandler>();
		try {
			/* check for new connections. */
			final ServerSocket server = new ServerSocket();

			try {
				server.setSoTimeout(1000);
				server.bind(new InetSocketAddress(CLIENT_HOST, CLIENT_PORT));

				while (!isInterrupted()) {
					final Socket client;
					try {
						client = server.accept();
					} catch (java.net.SocketTimeoutException e) {
						for (int i = connections.size() - 1; i >= 0; i--) {
							ConnectionHandler connection = connections.get(i);
							if (!connection.isAlive()) {
								// cleanup closed client connections
								connections.remove(i);
							}
						}
						continue;
					}
					
					final ConnectionHandler ch = new ConnectionHandler(client);
					connections.add(ch);
					ch.start();
				}
			}
			finally {
				for (ConnectionHandler connection : connections) {
					if (connection.isAlive()) {
						connection.interrupt();
						try {
							connection.join();
						} catch (InterruptedException e) {
						}
					}
				}
				server.close();
			}
		}
		catch (final IOException ioe) {
			ioe.printStackTrace();
		}
	}

	public NetworkSIDDevice() {
		this(DEFAULT_SID_DEVICES);
	}
	
	public NetworkSIDDevice(int numberOfSIDDevices) {
		this.numberOfSIDDevices = numberOfSIDDevices;
	}

	public static void main(final String[] args) {
		final NetworkSIDDevice sidDevice = new NetworkSIDDevice();

		Runnable runner = new Runnable() {
			public void run() {
			    createSystemTrayMenu(sidDevice);
			}
			
			private void createSystemTrayMenu(final NetworkSIDDevice sidDevice) {
				if (SystemTray.isSupported()) {
					final SystemTray tray = SystemTray.getSystemTray();

					PopupMenu popup = new PopupMenu();
					URL url = getClass().getResource("jsidplay2.gif");
					Image image = Toolkit.getDefaultToolkit().getImage(url);

					final TrayIcon trayIcon = new TrayIcon(image, "SID Network Device", popup);

					MenuItem item1 = new MenuItem("About");
					item1.addActionListener(new ActionListener() {
						synchronized public void actionPerformed(ActionEvent e) {
							if (abox == null) {
								abox = new AboutBox();
							}
							abox.show();
						}
					});

					popup.add(item1);

					MenuItem item2 = new MenuItem("Exit");
					item2.addActionListener(new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							if (abox != null) {
								abox.dispose();
							}
							tray.remove(trayIcon);
							sidDevice.interrupt();
						}
					});
					popup.add(item2);

					try {
						tray.add(trayIcon);
					} catch (AWTException e) {
						System.err.println("Can't add to tray");
					}
				} else {
					System.err.println("Tray unavailable");
				}
			}
		};
		EventQueue.invokeLater(runner);

		sidDevice.start();
		try {
			sidDevice.join();
		} catch (InterruptedException e) {
		}
	}
}
