/**
 *                           ReSid Emulation
 *                           ---------------
 *  begin                : Fri Apr 4 2001
 *  copyright            : (C) 2001 by Simon White
 *  email                : s_a_white@email.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 * @author Ken Hndel
 *
 */
package resid_builder;

import java.util.logging.Level;
import java.util.logging.Logger;

import libsidplay.common.EventScheduler;
import libsidplay.common.SIDBuilder;
import libsidplay.common.SIDEmu;
import resid_builder.resid.ISIDDefs.ChipModel;
import resid_builder.resid.ISIDDefs.SamplingMethod;
import resid_builder.resid.SID;

public class ReSID extends SIDEmu {
	private static final Logger RESID = Logger.getLogger(ReSID.class.getName());

	private static final String VERSION = "0.0.2";

	private static final int OUTPUTBUFFERSIZE = 5000; /* supports 5 ms chunk at 96 kHz */

	private final SID m_sid;

	private final String m_error;

	private final boolean m_status;

	public ReSID(final SIDBuilder builder, EventScheduler context) {
		super(builder, context);
		m_sid = new SID();
		m_status = true;
		m_error = "N/A";
		m_buffer = new float[OUTPUTBUFFERSIZE];
		m_bufferpos = 0;
		reset((byte) 0);
	}

	@Override
	public void filter(final FilterConfig filter6581, final FilterConfig filter8580) {
		if (filter6581 != null) {
			m_sid.getFilter6581().setCurveProperties(filter6581.getBaseresistance(), filter6581.getOffset(), filter6581.getSteepness(), filter6581.getMinimumfetresistance());
			m_sid.getFilter6581().setDistortionProperties(filter6581.getAttenuation(), filter6581.getNonlinearity(), filter6581.getResonanceFactor());
			m_sid.set6581VoiceNonlinearity(filter6581.getVoiceNonlinearity());
			m_sid.getFilter6581().setNonLinearity(filter6581.getVoiceNonlinearity());
		} else {
			m_sid.getFilter6581().setCurveAndDistortionDefaults();
		}

		if (filter8580 != null) {
			m_sid.getFilter8580().setCurveProperties(filter8580.getK(), filter8580.getB(), 0, 0);
			m_sid.getFilter8580().setDistortionProperties(0, 0, filter8580.getResonanceFactor());
		} else {
			m_sid.getFilter8580().setCurveAndDistortionDefaults();
		}
	}

	@Override
	public void reset(final byte volume) {
		clocksSinceLastAccess();
		m_sid.reset();
		m_sid.write(0x18, volume);
	}

	@Override
	public byte read(int addr) {
		addr &= 0x1f;
		clock();
		return m_sid.read(addr);
	}

	@Override
	public void write(int addr, final byte data) {
		addr &= 0x1f;
		super.write(addr, data);
		if (RESID.isLoggable(Level.FINE)) {
			RESID.fine(String.format("write 0x%02x=0x%02x", addr, data));
		}
		
		clock();
		m_sid.write(addr, data);
	}

	@Override
	public void clock() {
		int cycles = clocksSinceLastAccess();
		m_bufferpos += m_sid.clock(cycles, m_buffer, m_bufferpos);
	}

	@Override
	public void filter(final boolean enable) {
		m_sid.getFilter6581().enable(enable);
		m_sid.getFilter8580().enable(enable);
	}

	@Override
	public void voice(final int num, final boolean mute) {
		m_sid.mute(num, mute);
	}

	@Override
	public void sampling(final double systemclock, final float freq, final SamplingMethod method) {
		// Prevent an ArrayIndexOutOfBoundsException
		// switching the sampling method in the UI thread
		// by re-initializing the sample buffer
		m_bufferpos = 0;
		m_buffer = new float[OUTPUTBUFFERSIZE];
		m_sid.setSamplingParameters(systemclock, method, freq, 20000);
	}

	/**
	 * Set the emulated SID model
	 * 
	 * @param model
	 */
	public void model(final ChipModel model) {
		m_sid.setChipModel(model);
	}

	// Standard component functions
	public static final String credits() {
		String m_credit = "ReSID V" + VERSION + " Engine:\n";
		m_credit += "\tCopyright (C) 1999-2002 Simon White <sidplay2@yahoo.com>\n";
		m_credit += "MOS6581/8580 (SID) Emulation:\n";
		m_credit += "\tCopyright (C) 1999-2004 Dag Lem <resid@nimrod.no>\n";
		m_credit += "\tCopyright (C) 2005-2010 Antti S. Lankila <alankila@bel.fi>\n";
		return m_credit;
	}

	public final String error() {
		return m_error;
	}

	public boolean bool() {
		return m_status;
	}

	public SID sid() {
		return m_sid;
	}

}
