package sidplay;

import hardsid_builder.HardSID;
import hardsid_builder.HardSIDBuilder;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

import applet.disassembler.CPUParser;

import libsidplay.Player;
import libsidplay.common.ISID2Types.Clock;
import libsidplay.common.ISID2Types.PlayerConfig;
import libsidplay.common.SIDBuilder;
import libsidplay.common.SIDEmu;
import libsidplay.components.mos6510.IMOS6510Disassembler;
import libsidplay.sidtune.SidTune;
import libsidplay.sidtune.SidTune.Model;
import libsidplay.sidtune.SidTuneInfo;
import libsidutils.SidDatabase;
import resid_builder.FilterConfig;
import resid_builder.ReSID;
import resid_builder.ReSIDBuilder;
import resid_builder.resid.ISIDDefs.ChipModel;
import sidplay.audio.AudioConfig;
import sidplay.audio.AudioDriver;
import sidplay.audio.AudioNull;
import sidplay.audio.CompareDriver;
import sidplay.audio.JavaSound;
import sidplay.audio.LiveRecorderWav;
import sidplay.audio.WavFile;

public class ConsolePlayer {
	private static final Logger PLAYER = Logger.getLogger(ConsolePlayer.class
			.getName());

	// Previous song select timeout (3 secs)
	public static final int SID2_PREV_SONG_TIMEOUT = 4;

	/** Random source for triangular dithering */
	private static final Random RANDOM = new Random();
	/** State of TPDF. */
	private float oldRandomValue;
	
	/**
	 * Triangularly shaped noise source for audio applications.
	 * Output of this PRNG is between ]-1, 1[.
	 * 
	 * @return triangular noise sample
	 */
	private float triangularDithering() {
		float prevValue = oldRandomValue;
		oldRandomValue = RANDOM.nextFloat();
		return RANDOM.nextFloat() - prevValue;
	}

	protected static class LipshitzFIR {
		private final float[] lipshitz = new float[4];

		public float lipshitz(float input) {
			float out =
				+2.033f  * input
				-2.165f  * lipshitz[0]
			    +1.959f  * lipshitz[1]
			    -1.590f  * lipshitz[2]	
			    +0.6149f * lipshitz[3];
		
			lipshitz[3] = lipshitz[2];
			lipshitz[2] = lipshitz[1];
			lipshitz[1] = lipshitz[0];
			lipshitz[0] = input;

			return out;
		}
	}


	enum player_colour_t {
		black, red, green, yellow, blue, magenta, cyan, white
	}

	enum player_table_t {
		tableStart, tableMiddle, tableSeperator, tableEnd
	}

	public static final int playerError = 0;
	public static final int playerRunning = 1;
	public static final int playerPaused = 2;
	public static final int playerStopped = 3;
	public static final int playerRestart = 4;
	public static final int playerExit = 5;
	public static final int playerFast = 128;
	public static final int playerFastRestart = playerRestart | playerFast;
	public static final int playerFastExit = playerExit | playerFast;

	public enum SIDEMUS {
		/* Same as EMU_DEFAULT except no soundcard. Still allows wav generation */
		EMU_NONE,
		/* The following require a soundcard */
		EMU_RESID,
		/* The following should disable the soundcard */
		EMU_HARDSID
	}

	public enum OUTPUTS {
		OUT_NULL(false, new AudioNull()),
		OUT_SOUNDCARD(false, new JavaSound()),
		OUT_COMPARE(false, new CompareDriver()),
		OUT_WAV(true, new WavFile()),
		OUT_LIVE_WAV(true, new LiveRecorderWav());
		
		private final boolean fileBased;
		private final AudioDriver drv;
		
		OUTPUTS(boolean fileBased, AudioDriver drv) {
			this.fileBased = fileBased;
			this.drv = drv;
		}
		
		public boolean isFileBased() {
			return fileBased;
		}
		
		public AudioDriver getDriver() {
			return drv;
		}
	}

	/**
	 * Error and status message numbers.
	 */
	enum ERR {
		ERR_SYNTAX, ERR_NO_SIDS_DEFINED, ERR_FILE_OPEN
	}

	public Player m_engine = new Player();

	public PlayerConfig m_engCfg;

	public SidTune m_tune;

	private int m_state;

	public int state() {
		return m_state;
	}

	public String m_outfile = System.getProperty("user.home");

	public String m_filename;

	public IniConfig m_iniCfg = new IniConfig();
	public SidDatabase m_database = new SidDatabase();

	//
	// Display parameters
	//

	private short m_quietLevel;

	private IMOS6510Disassembler disassembler;
	
	boolean v1mute, v2mute, v3mute;

	protected static class m_filter_t {
		FilterConfig definition;
		boolean enabled;
	}

	private final m_filter_t m_filter = new m_filter_t();

	public static class m_driver_t {
		/** Default SID emulation */
		public SIDEMUS sid = SIDEMUS.EMU_NONE;

		/** Default output */
		public OUTPUTS output = OUTPUTS.OUT_NULL;
		/** Null audio when no output is desired. */
		public final AudioDriver nullAudio = output.getDriver();
		/** Prevailing audio configuration */
		public AudioConfig cfg = new AudioConfig();
		/** Currently active driver that sound gets written to */
		public AudioDriver selected = nullAudio;
		/** Driver we will eventually set as the selected driver */
		public AudioDriver device = nullAudio;
		/** Audio filters for noise-shaped dithering */
		public final LipshitzFIR[] fir = new LipshitzFIR[] {
				new LipshitzFIR(), new LipshitzFIR()
		};
	}

	public m_driver_t m_driver = new m_driver_t();

	public class m_timer_t { // secs
		long start;
		long current;
		long stop;
		public long length;
		boolean valid;
	}

	public m_timer_t m_timer = new m_timer_t();

	protected class m_track_t {
		/**
		 * First song number of the play-list. 0 is used, to reset the
		 * play-list start to the start tune, if a different tune is loaded.
		 */
		int first;
		/**
		 * Current song number. If first > 0 it wraps around the count of songs
		 * 0 means use start song of the tune file.
		 */
		int selected;
		/**
		 * Number of songs in the play-list.
		 */
		int songs;
		/**
		 * Loop, if the play-list is played.
		 */
		boolean loop;
		/**
		 * Always plays a single song (start song)
		 */
		boolean single;
	}

	private final m_track_t m_track = new m_track_t();

	public int getCurrentSong() {
		return m_track.selected;
	}

	public int getCurrentSongCount() {
		return m_track.songs;
	}

	public void setCurrentSingle(boolean single) {
		m_track.single = single;
	}

	protected class m_speed_t {
		short current;
		short max;
	}

	private final m_speed_t m_speed = new m_speed_t();

	private int m_verboseLevel;
	private float quantizingError;
	public SIDBuilder sidEmuFactory;

	public ConsolePlayer() {
		m_tune = null;
		m_state = playerStopped;
		m_outfile = null;
		m_quietLevel = 0;
		m_verboseLevel = 0;
		// Other defaults
		m_filename = "";
		m_filter.enabled = true;
		// m_driver.device = null;
		m_timer.start = 0;
		m_timer.length = 0; // FOREVER
		m_timer.valid = false;
		m_track.first = 0;
		m_track.selected = 0;
		m_track.loop = false;
		m_speed.current = 1;
		m_speed.max = 32;

		// Read default configuration
		m_iniCfg.read();
		m_engCfg = m_engine.getConfig();

		final IniConfig.audio_section audio = m_iniCfg.audio();
		final IniConfig.emulation_section emulation = m_iniCfg.emulation();

		// INI Configuration Settings
		m_engCfg.defaultClock = emulation.getDefaultClockSpeed();
		m_engCfg.userClock = emulation.getUserClockSpeed();
		m_engCfg.frequency = audio.getFrequency();
		m_engCfg.channels = audio.getChannels();

		m_filter.enabled = emulation.isFilter();
		m_engCfg.dualSidBase = emulation.getDualSidBase();
		m_engCfg.forceStereoTune = emulation.isForceStereoTune();

		m_track.single = m_iniCfg.sidplay2().isSingle();
		
		// Copy default setting to audio configuration
		m_driver.cfg.frequency = m_engCfg.frequency;
		m_driver.cfg.channels = m_engCfg.channels;

		if (m_iniCfg.sidplay2().getHvsc() != null) {
			m_database.open(m_iniCfg.sidplay2().getHvsc() + File.separator + "DOCUMENTS" + File.separator + "Songlengths.txt");
		}
		
		createOutput(OUTPUTS.OUT_NULL, null);
		createSidEmu(SIDEMUS.EMU_NONE, SidTune.Model.ANY);
	}

	/**
	 * Create the output object to process sound buffer
	 * 
	 * @param driver
	 * @param tuneInfo
	 * @return
	 */
	public boolean createOutput(final OUTPUTS driver, final String preferredWavBaseName) {
		// Remove old audio driver
		m_driver.selected = m_driver.nullAudio;
		if (m_driver.device != m_driver.nullAudio) {
			m_driver.device.close();
		}

		m_driver.device = driver.getDriver();

		String title = m_outfile;
		if (title == null) {
			title = "jsidplay2.wav";
		} else {
			if (new File(title).isDirectory()) {
				if (preferredWavBaseName != null) {
					title += File.separator + preferredWavBaseName;
				} else {
					title += File.separator + "jsidplay2.wav";
				}
			}
		}

		// Configure with user settings
		m_driver.cfg.frequency = m_engCfg.frequency;
		m_driver.cfg.channels = m_engCfg.channels;

		if (m_driver.device instanceof CompareDriver) {
			final IniConfig.audio_section audio = m_iniCfg.audio();
			((CompareDriver) m_driver.device).setPlayOriginal(audio.isPlayOriginal());
			((CompareDriver) m_driver.device).setMp3File(audio.getMp3File());
		}

		try { // Open the hardware
			m_driver.device.open(m_driver.cfg, title);
			if (m_driver.device != m_driver.nullAudio) {
				m_driver.nullAudio.open(m_driver.cfg, title);
			}
		}
		catch (final Exception e) {
			e.printStackTrace();
			return false;
		}

		// See what we got
		m_engCfg.frequency = m_driver.cfg.frequency;
		m_engCfg.channels = m_driver.cfg.channels;
		return true;
	}

	/**
	 * Create the sid emulation
	 * 
	 * @param emu
	 * @param tuneInfo
	 * @return
	 */
	public boolean createSidEmu(final SIDEMUS emu, final SidTune.Model sidModel) {
		sidEmuFactory = null;

		switch (emu) {
		case EMU_RESID: {
			final ReSIDBuilder rs = new ReSIDBuilder();

			final FilterConfig filter6581 = m_iniCfg.filter(ChipModel.MOS6581);
			final FilterConfig filter8580 = m_iniCfg.filter(ChipModel.MOS8580);
			rs.filter(m_filter.enabled);
			rs.filter(filter6581, filter8580);
			
			sidEmuFactory = rs;
			break;
		}
		
		case EMU_HARDSID:
			final HardSIDBuilder hs = new HardSIDBuilder();
			if (! hs.bool()) {
				displayError(hs.error());
				return false;
			}
			
			hs.setDevicesToUse(m_iniCfg, sidModel);
			if (! hs.bool()) {
				displayError(hs.error());
				return false;
			}
			
			hs.filter(m_filter.enabled);
			if (! hs.bool()) {
				displayError(hs.error());
				return false;
			}
			
			sidEmuFactory = hs;
			break;
		default:
			// Emulation Not yet handled
			// This default case results in the default
			// emulation
			break;
		}
		
		return true;
	}

	public boolean open() {
		if ((m_state & ~playerFast) == playerRestart) {
			if (m_quietLevel < 2) {
				System.err.println();
			}
			if ((m_state & playerFast) != 0) {
				m_driver.selected.reset();
			}
			m_state = playerStopped;
		}

		// Select the required song
		SidTuneInfo tuneInfo = null;
		if (m_tune != null) {
			m_track.selected = m_tune.selectSong(m_track.selected);
			if (m_track.first == 0) {
				// A different tune is opened (resetTrack was called)?
				// We mark a new play-list start
				m_track.first = m_track.selected;
			}
			tuneInfo = m_tune.getInfo();
		}
		m_engine.setTune(m_tune);

		// Get tune details
		String title = "jsidplay2.wav";
		Model sidModel = SidTune.Model.ANY;
		if (tuneInfo != null) {
			m_track.songs = tuneInfo.songs;
			String subTune = "";
			if (tuneInfo.songs > 1) {
				subTune = String.format("-%02d", tuneInfo.currentSong);
			}
			title = new File(tuneInfo.filename).getName().replace(".sid", subTune + ".wav");
			sidModel = tuneInfo.sidModel;
		}

		// Configure engine with settings
		m_engine.setConfig(m_engCfg);

		if (!createSidEmu(m_driver.sid, sidModel)) {
			return false;
		}

		if (!createOutput(m_driver.output, title)) {
			return false;
		}

		// According to the configuration, the SIDs must be updated.
		updateSidEmulation();

		// Reset SID chips, since this is not covered by the engine
		for (int chipNum = 0; chipNum < m_engine.getConfig().channels; chipNum++) {
			SIDEmu s = m_engine.getC64().getSID(chipNum);
			if (s != null) {
				s.reset((byte) 0);
			}
		}
		m_engine.reset();

		// Start the player. Do this by fast
		// forwarding to the start position
		m_driver.selected = m_driver.nullAudio;
		m_speed.current = m_speed.max;
		m_driver.device.setFastForward(m_speed.current);
		v1mute = v2mute = v3mute = false;

		// As yet we don't have a required songlength
		// so try the songlength database
		if (m_tune != null) {
			if (!m_timer.valid && m_iniCfg.sidplay2().isEnableDatabase()) {
				final int length = m_database.length(m_tune);
				if (length >= 0) {
					// length==0 means forever
					// this is used for tunes
					// of unknown length
					m_timer.length = length;
				}
			}
		} else {
			// normal reset? Disable timer
			m_timer.length = 0;
		}
		// Set up the play timer
		m_timer.stop = 0;
		m_timer.stop += m_timer.length;
		if (m_timer.valid) {
			// Length relative to start
			m_timer.stop += m_timer.start;
		} else {
			// Check to make start time dosen't exceed end
			if ((m_timer.stop & (m_timer.start >= m_timer.stop ? 1 : 0)) != 0) {
				displayError("ERROR: Start time exceeds song length!");
				return false;
			}
		}

		m_timer.current = ~0;
		m_state = playerRunning;

		// Update display
		if (tuneInfo != null) {
			menu();
		}
		return true;
	}

	public void close() {
		if (m_state == playerExit) {
			// Natural finish
			emuflush();
		} else {
			// Destroy buffers
			m_driver.selected.reset();
		}

		// Shutdown drivers, etc
		createOutput(OUTPUTS.OUT_NULL, null);
		createSidEmu(SIDEMUS.EMU_NONE, SidTune.Model.ANY);
		m_engine.close();
	}

	/**
	 * Reset track is called to play a different tune file
	 */
	public void resetTrack() {
		// 0 means set first song of play-list, next time open() is called
		m_track.first = 0;
		// 0 means use start song next time open() is called
		m_track.selected = 0;
	}

	/**
	 * Enable/disable song length database
	 * @param enableSLDb enable SLDb
	 */
	public void setSLDb(final boolean enableSLDb) {
		if (!m_timer.valid) {
			if (enableSLDb && m_tune != null) {
				final int length = m_database.length(m_tune);
				if (length > 0) {
					m_timer.length = length;
					m_timer.stop = m_timer.length;
				}
			} else {
				m_timer.length= 0;
				m_timer.stop = 0;
			}
		}
	}

	/**
	 * Flush any hardware sid fifos so all music is played and reset
	 */
	void emuflush() {
		if (sidEmuFactory instanceof HardSIDBuilder) {
			((HardSIDBuilder) sidEmuFactory).flush();
			((HardSIDBuilder) sidEmuFactory).reset();
		}
	}

	/**
	 * Out play loop to be externally called
	 */
	public boolean play() {
		/* handle switches to next song etc. */
		final int seconds = m_engine.time();
		if (seconds != m_timer.current) {
			m_timer.current = seconds;

			if (seconds == m_timer.start) {
				// Switch audio drivers.
				m_driver.selected = m_driver.device;
				normalSpeed();
				m_engine.setDebug(disassembler);
			}
			
			if (m_iniCfg.sidplay2().isEnableDatabase() && m_timer.stop != 0 && seconds >= m_timer.stop) {
				// Single song?
				if (m_track.single) {
					m_state = playerExit;
				} else {
					nextSong();
				
					// Check play-list end
					if (m_track.selected == m_track.first && !m_track.loop) {
						m_state = playerExit;
					}
				}
				if (m_state == playerExit) {
					// Natural finish
					emuflush();
				}
			}
		}

		if (m_state == playerRunning) {
			final ByteBuffer soundBuffer = m_driver.selected.buffer();
			/* elements in our audio output are always 16 bits wide */
			final float[] buffer = new float[soundBuffer.capacity() / 2];

			// Fill buffer
			int ret = m_engine.play(buffer);
			soundBuffer.clear();
			for (int i = 0; i < ret; i++) {
				float sample = buffer[i] * 32768f;
				m_driver.fir[i % m_driver.cfg.channels].lipshitz(quantizingError);
				int value = Math.round(sample + triangularDithering());
				quantizingError = value - sample;
				
				if (value > 32767) {
					value = 32767;
				}
				if (value < -32768) {
					value = -32768;
				}
				soundBuffer.putShort((short) value);
			}

			if (PLAYER.isLoggable(Level.FINE)) {
				for (int i = 0; i < soundBuffer.capacity(); i += 16) {
					final StringBuffer result = new StringBuffer();
					for (int j = 0; j < 16 && i + j < soundBuffer.capacity(); j++) {
						result.append(String.format("0x%02x ", soundBuffer.get(i+j)));
					}
					PLAYER.fine(result.toString());
				}
			}
			m_driver.selected.write();
		}
		
		return m_state == playerRunning || m_state == playerPaused;
	}

	private void decodeKeys() {
		try {
			final int key = System.in.read();
			switch (key) {
			case 'h':
				m_state = playerFastRestart;
				m_track.selected = 1;
				break;

			case 'e':
				m_state = playerFastRestart;
				m_track.selected = m_track.songs;
				break;

			case '>':
				nextSong();
				break;

			case '<':
				previousSong();
				break;

			case '.':
				fastForward();
				break;

			case ',':
				normalSpeed();
				break;

			case 'p':
				pause();
				break;

			case '1':
				v1mute = !v1mute;
				m_engine.mute(0, 0, v1mute);
				break;

			case '2':
				v2mute = !v2mute;
				m_engine.mute(0, 1, v2mute);
				break;

			case '3':
				v3mute = !v3mute;
				m_engine.mute(0, 2, v3mute);
				break;

			case '4':
				v1mute = !v1mute;
				m_engine.mute(1, 0, v1mute);
				break;

			case '5':
				v2mute = !v2mute;
				m_engine.mute(1, 1, v2mute);
				break;

			case '6':
				v3mute = !v3mute;
				m_engine.mute(1, 2, v3mute);
				break;

			case 'f':
				m_filter.enabled = !m_filter.enabled;
				sidEmuFactory.filter(m_filter.enabled);
				break;

			case 'q':
				quit();
				break;

			default:
				break;
			}
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	public void pause() {
		if (m_state == playerPaused) {
			m_state = playerRunning;
		} else {
			m_state = playerPaused;
			m_driver.selected.pause();
		}
	}

	public void nextSong() {
		m_state = playerFastRestart;
		m_track.selected++;
		if (m_track.selected > m_track.songs) {
			m_track.selected = 1;
		}
	}

	public void previousSong() {
		m_state = playerFastRestart;
		if (m_engine.time() < SID2_PREV_SONG_TIMEOUT) {
			m_track.selected--;
			if (m_track.selected < 1) {
				m_track.selected = m_track.songs;
			}
		}
	}

	public void restart() {
		m_state = playerFastRestart;
	}

	public void fastForward() {
		m_speed.current *= 2;
		if (m_speed.current > m_speed.max) {
			m_speed.current = m_speed.max;
		}
		m_driver.device.setFastForward(m_speed.current);
	}

	public void normalSpeed() {
		m_speed.current = 1;
		m_driver.device.setFastForward(1);
	}

	public void quit() {
		m_state = playerFastExit;
		m_driver.device.stop();
	}

	public void displayError(final String error) {
		System.err.println(this + ": " + error);
	}

	public void displayError(final ERR num) {
		System.err.print(this + ": ");

		if (num == ERR.ERR_SYNTAX) {
			System.err.println("command line syntax error" + "\n"
					+ "Try --help for more information.");
		} else

			if (num == ERR.ERR_NO_SIDS_DEFINED) {
				System.err.println("ERROR: No SIDs defined.");
			} else

				if (num == ERR.ERR_FILE_OPEN) {
					System.err.println("ERROR: Could not open file for binary input.");
				}
	}

	/**
	 * Convert time from integer
	 * 
	 * @param str
	 * @param time
	 * @return
	 */
	long parseTime(final String str) {
		int sep;
		long _time;

		// Check for empty string
		if (str.length() == 0) {
			return -1;
		}

		sep = str.lastIndexOf(':');
		if (sep == -1) {
			// User gave seconds
			_time = Integer.valueOf(str);
		} else {
			// Read in MM:SS format
			int val;
			val = Integer.valueOf(str.substring(0, sep));
			if (val < 0 || val > 99) {
				return -1;
			}
			_time = (long) val * 60;
			val = Integer.valueOf(str.substring(sep + 1));
			if (val < 0 || val > 59) {
				return -1;
			}
			_time += val;
		}

		return _time;
	}

	/**
	 * Parse command line arguments
	 * 
	 * @param argv
	 * @return
	 */
	public int args(final String[] argv) {
		int infile = 0;
		int i = 0;
		boolean err = false;

		if (argv.length == 0) {
			displayArgs(null);
			return -1;
		}

		// default arg options
		m_driver.output = OUTPUTS.OUT_SOUNDCARD;
		m_driver.sid = SIDEMUS.EMU_RESID;

		// parse command line arguments
		while (i < argv.length) {
			if (argv[i].charAt(0) == '-' && argv[i].length() > 1) {
				// help options
				if (argv[i].charAt(1) == 'h' || argv[i].equals("--help")) {
					displayArgs(null);
					return 0;
				} else if (argv[i].equals("--help-debug")) {
					displayDebugArgs();
					return 0;
				}

				else if (argv[i].charAt(1) == 'b') {
					final long time = parseTime(argv[i].substring(2));
					if (time == -1) {
						err = true;
					}
					m_timer.start = time;
				} else if (argv[i].equals("-fd")) {
					// Override sidTune and enable the second sid
					m_engCfg.forceStereoTune = true;
				} else if (argv[i].startsWith("-f")) {
					if (argv[i].length() == 2) {
						err = true;
					}
					m_engCfg.frequency = Integer.valueOf(argv[i].substring(2));
				}

				// New/No filter options
				else if (argv[i].startsWith("-nf")) {
					if (argv[i].length() == 3) {
						m_filter.enabled = false;
					}
				}

				// Newer sid (8580)
				else if (argv[i].startsWith("-ns")) {
					switch (argv[i].charAt(3)) {
					case '1':
						m_iniCfg.emulation().setUserSidModel(ChipModel.MOS8580);
						break;
						// No new sid so use old one (6581)
					case '0':
						m_iniCfg.emulation().setUserSidModel(ChipModel.MOS6581);
						break;
					default:
						err = true;
					}
				}

				// Track options
				else if (argv[i].startsWith("-nols")) {
					m_track.loop = true;
					m_track.single = true;
					m_track.first = Integer.valueOf(argv[i].substring(4));
				} else if (argv[i].startsWith("-ol")) {
					m_track.loop = true;
					m_track.first = Integer.valueOf(argv[i].substring(3));
				} else if (argv[i].startsWith("-os")) {
					m_track.single = true;
					m_track.first = Integer.valueOf(argv[i].substring(3));
				} else if (argv[i].startsWith("-o")) {
					// User forgot track number ?
					if (argv[i].length() == 2) {
						err = true;
					}
					m_track.first = Integer.valueOf(argv[i].substring(2));
				}

				else if (argv[i].startsWith("-q")) {
					if (argv[i].length() == 2) {
						m_quietLevel = 1;
					} else {
						m_quietLevel = Short.valueOf(argv[i].substring(2));
					}
				} else if (argv[i].equals("-s")) {
					m_engCfg.channels = 2;
				}

				else if (argv[i].startsWith("-t")) {
					final long time = parseTime(argv[i].substring(2));
					if (time == -1) {
						err = true;
					}
					m_timer.length = time;
					m_timer.valid = true;
					m_iniCfg.sidplay2().setEnableDatabase(true);
				}

				// Video/Verbose Options
				else if (argv[i].equals("-vnf")) {
					m_engCfg.userClock = Clock.NTSC;
				} else if (argv[i].equals("-vpf")) {
					m_engCfg.userClock = Clock.PAL;
				} else if (argv[i].equals("-vn")) {
					m_engCfg.defaultClock = Clock.NTSC;
				} else if (argv[i].equals("-vp")) {
					m_engCfg.defaultClock = Clock.PAL;
				} else if (argv[i].startsWith("-v")) {
					if (argv[i].length() == 2) {
						m_verboseLevel = 1;
					} else {
						m_verboseLevel = Integer.valueOf(argv[i].substring(2));
					}
					// } else if (argv[i].startsWith("--delay=")) {
					//    m_engCfg.powerOnDelay = new Integer(argv[i].substring(8));
				}

				// File format conversions
				else if (argv[i].equals("-w") || argv[i].equals("--wav")) {
					m_driver.output = OUTPUTS.OUT_WAV;
					m_outfile = argv[++ i];
				} else if (argv[i].equals("-l")) {
					m_driver.output = OUTPUTS.OUT_LIVE_WAV;
					i ++;
					m_outfile = argv[i];
				}

				// Hardware selection
				else if (argv[i].equals("--hardsid")) {
					m_driver.sid    = SIDEMUS.EMU_HARDSID;
					m_driver.output = OUTPUTS.OUT_NULL;
				}

				// These are for debug
				else if (argv[i].equals("--none")) {
					m_driver.sid = SIDEMUS.EMU_NONE;
					m_driver.output = OUTPUTS.OUT_NULL;
				} else if (argv[i].equals("--nosid")) {
					m_driver.sid = SIDEMUS.EMU_NONE;
				} else if (argv[i].equals("--cpu-debug")) {
					disassembler = CPUParser.getInstance();
				}

				else {
					err = true;
				}

			} else {
				// Reading file name
				if (infile == 0) {
					infile = i;
				} else {
					err = true;
				}
			}

			if (err) {
				displayArgs(argv[i]);
				return -1;
			}

			i++; // next index
		}

		// Can only loop if not creating audio files
		if (m_driver.output.isFileBased()) {
			m_track.loop = false;
		}

		// Check to see if we are trying to generate an audio file
		// whilst using a hardware emulation
		if (m_driver.output.isFileBased() && m_driver.sid == SIDEMUS.EMU_HARDSID) {
			displayError("ERROR: Cannot generate audio files using hardware emulations");
			return -1;
		}

		if (loadTune(argv[infile]) == -1) {
			return -1;
		}

		// Select the desired track
		// and also mark the play-list start
		m_track.first = m_tune.selectSong(m_track.first);
		m_track.selected = m_track.first;
		if (m_track.single) {
			m_track.songs = 1;
		}

		// If user provided no time then load songlength database
		// and set default lengths incase it's not found in there.
		{
			if (m_driver.output.isFileBased() && m_timer.valid && m_timer.length == 0) {
				// Time of 0 provided for wav generation
				displayError("ERROR: -t0 invalid in record mode");
				return -1;
			}
			if (! m_timer.valid) {
				m_timer.length = m_iniCfg.sidplay2().getPlayLength();
				if (m_driver.output.isFileBased()) {
					m_timer.length = m_iniCfg.sidplay2().getRecordLength();
				}
			}
		}

		m_filter.definition = m_iniCfg.filter(m_iniCfg.emulation().getUserSidModel());

		// Configure engine with settings
		m_engine.setConfig(m_engCfg);
		return 1;
	}

	/**
	 * Load a tune, this is either an absolute path name or a URL of an applet
	 * version.
	 * 
	 * @param resource
	 *            URL or filename
	 * @return 0 - OK, -1 means load error
	 */
	public int loadTune(final String resource) {
		m_filename = resource;
		if (resource == null) {
			m_tune = null;
			m_track.first = 1;
			m_track.songs = 0;
			return 0;
		}

		// Load the tune
		try {
			// load from URL (applet version)
			final URL url = new URL(resource);
			final InputStream stream = url.openConnection().getInputStream();
			m_tune = SidTune.load(stream);
			m_tune.getInfo().filename = m_filename;
			stream.close();
		} catch (final MalformedURLException e1) {
			try {
				// load from file
				m_tune = SidTune.load(m_filename);
			} catch (final Exception e) {
				e.printStackTrace();
				m_tune = null;
				return -1;
			}
		} catch (final Exception e) {
			e.printStackTrace();
			m_tune = null;
			return -1;
		}
		return 0;
	}

	void displayArgs(final String arg) {
		final PrintStream out = arg != null ? System.err : System.out;

		if (arg != null) {
			out.println("Option Error: " + arg);
		} else {
			out.println("Syntax: java -jar jsidplay2_console.jar [-<option>...] <datafile>");
		}

		out.println("Options:"
				+ "\n"
				+ " --help|-h    display this screen"
				+ "\n"
				+ " --help-debug debug help menu"
				+ "\n"
				+ " -b<num>      set start time in [m:]s format (default 0)"
				+ "\n"

				+ " -f<num>      set frequency in Hz (default: " + m_engine.getConfig().frequency + ")"
				+ ")"
				+ "\n"
				+ " -fd          force dual sid environment"
				+ "\n"
				+ " -fs          force samples to a channel (default: uses sid)"
				+ "\n"

				+ " -nf[filter]  no/new SID filter emulation"
				+ "\n"
				+ " -ns[0|1]     (no) MOS 8580 waveforms (default: from tune or cfg)"
				+ "\n"

				+ " -o<l|s>      looping and/or single track"
				+ "\n"
				+ " -o<num>      start track (default: preset)"
				+ "\n"

				// #if !defined(DISALLOW_STEREO_SOUND)
				+ " -s           stereo sid support"
				+ "\n"
				// #endif

				+ " -t<num>      set play length in [m:]s format (0 is endless)"
				+ "\n"

				+ " -<v[level]|q>       verbose (level=0,1,2) or quiet (no time display) output"
				+ "\n"
				+ " -v[p|n][f]   set VIC PAL/NTSC clock speed (default: defined by song)"
				+ "\n"
				+ "              Use 'f' to force the clock by preventing speed fixing"
				+ "\n"

				+ " -w name     create wav file"
				+ " -l name     create wav file with playback (live)");
		out.println(" --hardsid enable hardsid support\n");
		out.println("\n"
				// Changed to new homepage address
				+ "Home Page: http://jsidplay2.sourceforge.net/" + "\n");

		/* XXX: minor blemish, as we might not actually be interactive now.
		 * The REAL problem is with having args() method, which is used
		 * internally.
		 * We should just have "setXyz() method,
		 * not a cp.args(new String[] { "--xyz" });
		 */
		System.out.println("<press return>");
		try {
			System.in.read();
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	void displayDebugArgs() {
		final PrintStream out = System.out;

		out
		.println("Debug Options:"
				+ "\n"
				+ " --cpu-debug   display cpu register and assembly dumps"
				+ "\n"
				+ " --delay=<num> simulate c64 power on delay"
				+ "\n"

				+ " --wav<file>   wav file output device" + "\n"
				+ " --none        no audio output device" + "\n"
				+ " --nosid       no sid emulation" + "\n");
		System.out.println("<press return>");
		try {
			System.in.read();
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	void menu() {
		final SidTuneInfo tuneInfo = m_tune.getInfo();
		if (m_quietLevel > 1) {
			return;
		}

		final IniConfig.console_section console = m_iniCfg.console();

		System.out.println(String.format("%c%s%c", console.getTopLeft(), setfill(
				console.getHorizontal(), 54), console.getTopRight()));
		System.out.println(String.format("%c%54s%c", console.getVertical(),
				" Java SIDPLAY - Music Player & C64 SID Chip Emulator  ",
				console.getVertical()));
		System.out.println(String.format("%c%s%c", console.getJunctionLeft(),
				setfill(console.getHorizontal(), 54), console.getJunctionRight()));
		if (tuneInfo.numberOfInfoStrings > 0) {
			if (tuneInfo.numberOfInfoStrings == 3) {
				System.out.println(String.format("%c Title        : %37s %c",
						console.getVertical(), tuneInfo.infoString[0],
						console.getVertical()));
				System.out.println(String.format("%c Author       : %37s %c",
						console.getVertical(), tuneInfo.infoString[1],
						console.getVertical()));
				System.out.println(String.format("%c Released     : %37s %c",
						console.getVertical(), tuneInfo.infoString[2],
						console.getVertical()));
			} else {
				for (int i = 0; i < tuneInfo.numberOfInfoStrings; i++) {
					System.out.println(String.format(
							"%c Description  : %37s %c", console.getVertical(),
							tuneInfo.infoString[i], console.getVertical()));
				}
			}
			System.out.println(String.format("%c%s%c", console.getJunctionLeft(),
					setfill(console.getHorizontal(), 54), console.getJunctionRight()));
		}
		if (m_verboseLevel != 0) {
			System.out.println(String.format("%c File format  : %37s %c",
					console.getVertical(), tuneInfo.getClass().getSimpleName(), console.getVertical()));
			System.out.println(String.format("%c Filename(s)  : %37s %c",
					console.getVertical(), new File(tuneInfo.filename).getName(), console.getVertical()));
		}
		System.out.print(String.format("%c Playlist     : ", console.getVertical()));
		{ // This will be the format used for playlists
			int i = 1;
			if (!m_track.single) {
				i = m_track.selected;
				i -= m_track.first - 1;
				if (i < 1) {
					i += m_track.songs;
				}
			}
			System.out.println(String.format("%37s %c", i + "/" + m_track.songs
					+ " (tune " + tuneInfo.currentSong + "/" + tuneInfo.songs
					+ "[" + tuneInfo.startSong + "])"
					+ (m_track.loop ? " [LOOPING]" : ""), console.getVertical()));
		}
		if (m_verboseLevel > 0) {
			System.out.println(String.format("%c%s%c", console.getBottomLeft(),
					setfill(console.getHorizontal(), 54), console.getBottomRight()));
			System.out.println(String.format("%c Song Speed   : %37s %c",
					console.getVertical(), tuneInfo.songSpeed, console.getVertical()));
		}
		System.out.print(String.format("%c Song Length  : ", console.getVertical()));
		if (m_timer.stop != 0) {
			final String time = String.format("%02d:%02d",
					(m_timer.stop / 60 % 100), (m_timer.stop % 60));
			System.out.print(String.format("%37s %c", "" + time,
					console.getVertical()));
		} else if (m_timer.valid) {
			System.out.print(String.format("%37s %c", "FOREVER",
					console.getVertical()));
		} else {
			System.out.print(String.format("%37s %c", "UNKNOWN",
					console.getVertical()));
		}
		System.out.println();
		if (m_verboseLevel > 0) {
			System.out.println(String.format("%c%s%c", console.getBottomLeft(),
					setfill(console.getHorizontal(), 54), console.getBottomRight()));
			StringBuffer line = new StringBuffer();
			// Display PSID Driver location
			line.append("DRIVER = ");
			if (tuneInfo.determinedDriverAddr == 0) {
				line.append("NOT PRESENT");
			} else {
				line.append(String.format("$%04x", tuneInfo.determinedDriverAddr));
				line.append(String.format("-$%04x", tuneInfo.determinedDriverAddr
						+ tuneInfo.determinedDriverLength - 1));
			}
			if (tuneInfo.playAddr == 0xffff) {
				line.append(String.format(", SYS = $%04x", tuneInfo.initAddr));
			} else {
				line.append(String.format(", INIT = $%04x", tuneInfo.initAddr));
			}

			System.out.println(String.format("%c Addresses    : %37s %c",
					console.getVertical(), line.toString(), console.getVertical()));
			line = new StringBuffer();
			line.append(String.format("LOAD   = $%04x", tuneInfo.loadAddr));
			line.append(String.format("-$%04x", tuneInfo.loadAddr
					+ tuneInfo.c64dataLen - 1));
			if (tuneInfo.playAddr != 0xffff) {
				line.append(String.format(", PLAY = $%04x", tuneInfo.playAddr));
			}
			System.out.println(String.format("%c              : %37s %c",
					console.getVertical(), line.toString(), console.getVertical()));

			line = new StringBuffer();
			line.append(String.format("Filter = %s",
					(m_filter.enabled == true ? "Yes" : "No")));
			line.append(String.format(", Model = %s",
					(tuneInfo.sidModel == SidTune.Model.MOS8580 ? "8580" : "6581")));
			System.out.println(String.format("%c SID Details  : %37s %c",
					console.getVertical(), line.toString(), console.getVertical()));

			if (m_verboseLevel > 1) {
				line = new StringBuffer();
				System.out.println(String.format("%c Delay        : %37s %c",
						console.getVertical(), line.toString(), console.getVertical()));
			}
		}

		System.out.println(String.format("%c%s%c", console.getBottomLeft(), setfill(
				console.getHorizontal(), 54), console.getBottomRight()));
		System.out.println("keyboard control (press enter after command):");
		System.out.println("< > - play previous/next tune");
		System.out.println(", . - normal/faster speed");
		System.out.println("p   - pause/continue player");
		System.out.println("h e - play first/last tune");
		System.out.println("1   - mute voice 1");
		System.out.println("2   - mute voice 2");
		System.out.println("3   - mute voice 3");
		System.out.println("4   - mute voice 1 (stereo)");
		System.out.println("5   - mute voice 2 (stereo)");
		System.out.println("6   - mute voice 3 (stereo)");
		System.out.println("f   - enable/disable filter");
		System.out.println("q   - quit player");
	}

	private String setfill(final char ch, final int length) {
		final StringBuffer ret = new StringBuffer();
		for (int i = 0; i < length; i++) {
			ret.append(ch);
		}
		return ret.toString();
	}

	public static void main(final String[] args) {
		final ConsolePlayer player = new ConsolePlayer();
		if (player.args(args) < 0) {
			System.exit(1);
		}
		if (player.m_tune == null) {
			System.exit(0);
		}
		main_restart: while (true) {
			if (!player.open()) {
				player.close();
				System.exit(0);
			}
			while (true) {
				if (!player.play()) {
					break;
				}
				try {
					if (player.m_quietLevel < 2 &&
							(player.state() == playerPaused || System.in.available() != 0)) {
						player.decodeKeys();
					}
				} catch (final IOException e) {
					e.printStackTrace();
				}
			}
			if ((player.state() & ~playerFast) == playerRestart) {
				continue main_restart;
			}
			break;
		}
		player.close();
	}

	/**
	 * Replace old SIDs with new SIDs.
	 * 
	 */
	public void updateSidEmulation() {
		ChipModel chipModel = determineSIDModel();
		/* Determine stereo SID type and construct them */
		ChipModel stereoChipModel = m_iniCfg.emulation().getStereoSidModel();
		if (stereoChipModel == null) {
			// default is the same as the 1st SID chip model
			stereoChipModel = chipModel;
		}
	
		if (sidEmuFactory != null) {
			for (int i = 0; i < m_engine.getConfig().channels; i++) {
				switch (i) {
				case 1:
					// Stereo SID
					updateSIDEmu(i, stereoChipModel);
					break;
				default:
					// Mono SID
					updateSIDEmu(i, chipModel);
					break;
				}
			}
		}
	}

	/**
	 * Determine SID model to be used.
	 * The following search order is used:
	 * <OL>
	 * <LI>Use the locked SID model
	 * <LI>Use model specified in the currently played tune
	 * <LI>User default SID model
	 * </OL>
	 * @return the SID model to be used
	 */
	public ChipModel determineSIDModel() {
		/* Determine SID type and construct them */
		ChipModel chipModel = m_iniCfg.emulation().getUserSidModel();
		if (chipModel == null) {
			chipModel = m_iniCfg.emulation().getDefaultSidModel();
			if (m_tune != null) {
				final SidTuneInfo tuneInfo = m_tune.getInfo();				
				switch (tuneInfo.sidModel) {
				case MOS6581:
				case MOS8580:
					chipModel = ChipModel.valueOf(tuneInfo.sidModel.toString());
					break;
				default:
					break;
				}
			}
		}
		return chipModel;
	}

	/**
	 * Change SID model and sampling frequency. A current ReSID is re-used. If
	 * HardSID is used, a new chip is locked.
	 * 
	 * @param chipNum
	 *            chip number (0 - mono, 1 - stereo)
	 * @param model
	 *            chip model to use
	 */
	private void updateSIDEmu(final int chipNum, final ChipModel model) {
		SIDEmu s = m_engine.getC64().getSID(chipNum);
		// ReSID can be re-used, HardSID cannot be re-used
		if (s == null || s instanceof HardSID) {
			// Last HardSID must be reset!
			if (s != null && s instanceof HardSID) {
				((HardSID) s).write(0x18, (byte) 0x00);
				((HardSID) s).flush();
				s.reset();
			}
			s = sidEmuFactory
					.lock(m_engine.getC64().getEventScheduler(), model);
		} else {
			// Change current chip model
			if (s instanceof ReSID) {
				((ReSID) s).model(model);
			}
		}
		// Set sampling frequency
		s.sampling(m_engine.getC64().getClock().getCpuFrequency(), m_engine
				.getConfig().frequency, m_iniCfg.audio().getSampling());
		m_engine.getC64().setSid(chipNum, s);
	}
}
