package sidplay;

import java.nio.ByteBuffer;

import libsidplay.Player;
import libsidplay.common.ISID2Types.Clock;
import libsidplay.common.ISID2Types.PlayerConfig;
import libsidplay.common.SIDEmu;
import libsidplay.sidtune.SidTune;
import resid_builder.FilterConfig;
import resid_builder.ReSIDBuilder;
import resid_builder.resid.ISIDDefs.ChipModel;
import sidplay.audio.AudioConfig;
import sidplay.audio.JavaSound;

/**
 * This test class demonstrates the use of the ReSID engine. It has only minimum
 * dependencies to show how to integrate the player in other java environments.
 * 
 * @author Ken Hndel
 * 
 */
public class Test {

	/**
	 * Internal player engine
	 */
	private static Player engine = new Player();
	/**
	 * Audio driver to be used
	 */
	private static JavaSound device;

	/**
	 * Play a tune by filename
	 * 
	 * @param filename
	 *            the filename to load the tune
	 */
	public void playTune(final String filename) throws Exception {
		// Load tune and select first song
		final SidTune tune = SidTune.load(filename);
		tune.selectSong(1);

		// Apply the tune to the engine
		engine.setTune(tune);

		// Read filter settings and create filter
		final IniConfig iniCfg = new IniConfig();
		iniCfg.readInternal();
		final FilterConfig filter6581 = iniCfg.filter(ChipModel.MOS6581);
		final FilterConfig filter8580 = iniCfg.filter(ChipModel.MOS8580);

		// Customize player configuration
		final PlayerConfig engCfg = engine.getConfig();
		engCfg.defaultClock = Clock.PAL;
		engCfg.frequency = iniCfg.audio().getFrequency();
		engine.setConfig(engCfg);

		// Setup the SID emulation (not part of the player)
		final ReSIDBuilder rs = new ReSIDBuilder();

		// Enable/apply filter to the SID emulation
		rs.filter(true);
		rs.filter(filter6581, filter8580);

		// Create SID chip of desired model (mono tunes need exactly one)
		final SIDEmu sid = rs.lock(engine.getC64().getEventScheduler(), ChipModel.MOS6581);
		sid.sampling(engine.getC64().getClock().getCpuFrequency(), engCfg.frequency, iniCfg.audio().getSampling());

		// Apply mono SID chip to the C64, then reset
		engine.getC64().setSid(0, sid);
		engine.reset();

		// Get sound driver and open device
		device = new JavaSound();
		final AudioConfig cfg = new AudioConfig();
		cfg.frequency = engCfg.frequency;
		device.open(cfg, null);

		// Play forever
		while (true) {
			playCurrentBuffer();
		}
	}

	/**
	 * Play a buffer with sound data generated by the emulator
	 */
	private void playCurrentBuffer() {
		// Get buffer of the sound driver
		final ByteBuffer soundBuffer = device.buffer();
		soundBuffer.clear();
		final int samples = soundBuffer.capacity() / 2;

		// fill buffer with emulators output
		final float[] buffer = new float[samples];
		final long ret = engine.play(buffer);
		for (int i = 0; i < samples; i ++) {
			// convert float output to 16-bit sample data
			int value = (int) ((i < ret ? buffer[i] : 0) * 32767);
			// respect the limits
			if (value > 32767) {
				value = 32767;
			}
			if (value < -32768) {
				value = -32768;
			}
			// fill sound buffer
			soundBuffer.putShort((short) value);
		}
		// play generated sound
		device.write();
	}

	public static void main(final String[] args) throws Exception {
		if (args.length < 1) {
			System.err.println("Missing argument: <filename>");
			System.exit(-1);
		}
		final Test tst = new Test();
		tst.playTune(args[0]);
	}
}
