package sidplay.audio;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

public class JavaSound extends AudioDriver {
	private AudioFormat audioFormat;
	private SourceDataLine dataLine;

	private ByteBuffer sampleBuffer;
	
	@Override
	public void open(final AudioConfig cfg, final String name) throws LineUnavailableException {
		audioFormat = new AudioFormat(cfg.frequency, 16, cfg.channels, true, false);
		dataLine = AudioSystem.getSourceDataLine(audioFormat);
		dataLine.open(dataLine.getFormat(), cfg.bufSize * 2 * cfg.channels);

		/* Write to audio device often.
		 * We make the sample buffer size divisible by 64 to ensure that all
		 * fast forward factors can be handled. (32x speed, 2 channels)
		 */
		sampleBuffer = ByteBuffer.allocate(cfg.bufSize / 4 & 0xffffffc0);
		sampleBuffer.order(ByteOrder.LITTLE_ENDIAN);
		stopped = false;
	}

	@Override
	public void write() {
		if (! dataLine.isActive()) {
			dataLine.start();
		}

		int len;
		if (fastForward > 1) {
			sampleBuffer.rewind();
			int newLen = 0;
			int[] val = new int[audioFormat.getChannels()];
			int j = 0;
			/* for each short-formatted sample in the buffer... */
			while (sampleBuffer.position() < sampleBuffer.capacity()) {
				/* accumulate each interleaved channel into its own accumulator */
				for (int c = 0; c < audioFormat.getChannels(); c ++) {
					val[c] += sampleBuffer.getShort();
				}

				/* once enough samples have been accumulated, write one to output */
				j ++;
				if (j == fastForward) {
					j = 0;

					for (int c = 0; c < audioFormat.getChannels(); c ++) {
						sampleBuffer.putShort(newLen, (short) (val[c] / fastForward));
						newLen += 2;
					}

					/* zero accumulator */
					Arrays.fill(val, 0);
				}
			}
			len = newLen;
		} else {
			len = sampleBuffer.capacity();
		}
		while (dataLine.available() < len && !stopped) {
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		if (dataLine.available() >= len) {
			dataLine.write(sampleBuffer.array(), 0, len);
		}
	}

	@Override
	public void pause() {
		if (dataLine.isActive()) {
			dataLine.drain();
			dataLine.stop();
		}
	}

	@Override
	public void close() {
		if (dataLine == null) {
			return;
		}
		if (dataLine.isActive()) {
			dataLine.drain();
			dataLine.stop();
		}
		if (dataLine.isOpen()) {
			dataLine.close();
		}
	}

	@Override
	public void reset() {
		if (dataLine.isActive()) {
			dataLine.drain();
		}
	}

	public AudioFormat getAudioFormat() {
		return audioFormat;
	}
	
	@Override
	public ByteBuffer buffer() {
		return sampleBuffer;
	}
}
